/*
 * AudioBusManager.c 
 *
 * Copyright (c) 2008 Wind River Systems, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Lesser GNU General Public License version 2.1 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the Lesser GNU General Public License for more details.
 *
 * You should have received a copy of the Lesser GNU General Public License version 2.1
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */
/// @cond internal_documentation
/**
 * \file 
 *    Audio Bus Manager code.
 *
 * \version $Revision: 1.10 $
 * \date    $Date: 2008/07/12 00:59:37 $
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <glib.h>
#include <glib/gprintf.h>
#include <gmodule.h>

#include "PluginManager.h"
#include "AudioBusManagerPlugin.h"
#include "AudioBusManager.h"

/* Define the symbol ABM_DEBUG_INFO from either in this header file of
 * in the make file to enable all debug outputs.
 */
/*#define ABM_DEBUG_INFO*/

/* Define the symbol ABM_ERROR_LOG from either in this header file of
*in the make file to enable all error logging.
*/
#define ABM_ERROR_LOG

/** Debug print method */

#ifdef ABM_DEBUG_INFO
#define ABM_PRINT_INFO g_debug
#else
#define ABM_PRINT_INFO(fmt, arg...)
#endif

/** Debug print method */

#ifdef ABM_ERROR_LOG
#define ABM_PRINT_ERROR_LOG g_critical
#else
#define ABM_PRINT_ERROR_LOG(fmt, arg...)
#endif

/*
 * Local types and variables
 */
/** Max length of plugin file name including full path */
#define ABM_PLUGIN_MAX_FILE_NAME_LENGTH 255

/** Array of Audio Bus plugin data, indexed by plugin type*/

GArray *pABMPluginStore = NULL;

/*
 * Local functions
 */
static PluginData_t *ABM_PluginStore_GetPlugin
(
    gchar *pPluginType
);


static gboolean ABM_PluginStore_Initialize (
);


static gboolean ABM_PluginStore_Terminate (
);


static PluginData_t *ABM_PluginStore_GetPluginByIndex
(
    guint Index
);

/*--------------------------------------------------------------------------
 * Audio Bus Manager Functions
 *--------------------------------------------------------------------------
 *
 */

gboolean ABM_Initialize (
)
{
    GError *error = NULL;
    PluginData_t *pData;
    gboolean success = TRUE;
    GDir *pPluginDir = NULL;
    gchar PluginFullFileName[ABM_PLUGIN_MAX_FILE_NAME_LENGTH];
    const gchar *pFileName;
    int PluginCount = 0;

    /* Initialize plugin store */
    if (!ABM_PluginStore_Initialize ())
    {
        return FALSE;
    }
    /* scan the plugin store folder and get the file names for available plugins*/
    pPluginDir = g_dir_open (ABM_PLUGIN_STORE_FOLDER, 0, &error);

    if (!pPluginDir)
    {
        ABM_PRINT_ERROR_LOG (
            G_STRLOC "ABM Plugins folder: %s does not exist\n", ABM_PLUGIN_STORE_FOLDER);
        g_error_free (error);
        ABM_PluginStore_Terminate ();
        return FALSE;
    }
    else
    {
        do
        {
            if (PluginCount >= ABM_MAX_PLUGINS)
            {
                ABM_PRINT_INFO (
                    G_STRLOC "Max PluginCount reached\n");
                break;
            }
            pFileName = g_dir_read_name (pPluginDir);
            if (!pFileName)
            {
                ABM_PRINT_INFO ("No more Plugins to load from : %s\n", ABM_PLUGIN_STORE_FOLDER);
                break;
            }
            memset (PluginFullFileName, 0, sizeof (PluginFullFileName));
            pData = ABM_PluginStore_GetPluginByIndex (PluginCount);
            if (pData)
            {
                g_sprintf (PluginFullFileName, "%s/%s", ABM_PLUGIN_STORE_FOLDER, pFileName);
                pData->pFileName = PluginFullFileName;

                success = PM_LoadPlugin (pData, ABM_PLUGIN_INIT_FN_NAME);
                if (success)
                {
                    success = ((ABM_PluginInit_t)pData->pInitFn) (pData);
                    if (success)
                    {
                        pData->Loaded = TRUE;
                        ABM_PRINT_INFO ("Loaded Audio Bus Plugn: %s\n", pData.PluginType);
                        PluginCount++;
                    }
                    else
                    {
                        ABM_PRINT_ERROR_LOG (
                            G_STRLOC " Failed to initialise plug-in: %s\n", pData->pFileName);
                        continue;
                    }
                }
            }
            else
            {
                continue;
            }
        } while (1);
    }

    if (pPluginDir)
    {
        g_dir_close (pPluginDir);
    }

    return success;
}


gboolean ABM_Terminate (
)
{
    gboolean Result;
    Result = ABM_PluginStore_Terminate ();
    return Result;
}


ABM_Results_t ABM_Allocate
(
    ABM_AudioPluginParams_t AudioPluginParams, GHashTable *pAudioParamsTable, ABM_AudioSetupCb_t pAllocateCb
)
{
    PluginData_t *pPluginData;

    pPluginData = ABM_PluginStore_GetPlugin (AudioPluginParams.PluginType);

    if (pPluginData && pPluginData->Loaded)
    {
        PLUGIN_DATA_FUNC_TABLE->FnAllocateChannels (AudioPluginParams, pAudioParamsTable, pAllocateCb);
    }
    else
    {
        ABM_PRINT_ERROR_LOG (
            G_STRLOC "Allocation failed, unsupported Plugin Type: %s\n", AudioPluginParams.PluginType);
        return ABM_FAILURE;
    }

    return ABM_SUCCESS;
}


ABM_Results_t ABM_DeAllocate
(
    ABM_AudioPluginParams_t AudioPluginParams, GHashTable *AudioParamsTable, ABM_AudioSetupCb_t pDeAllocateCb
)
{
    PluginData_t *pPluginData;

    pPluginData = ABM_PluginStore_GetPlugin (AudioPluginParams.PluginType);

    if (pPluginData && pPluginData->Loaded)
    {
        PLUGIN_DATA_FUNC_TABLE->FnDeAllocateChannels (AudioPluginParams, AudioParamsTable, pDeAllocateCb);
    }
    else
    {
        ABM_PRINT_ERROR_LOG (
            G_STRLOC "DeAllocation failed, unsupported Plugin Type: %s\n", AudioPluginParams.PluginType);
        return ABM_FAILURE;
    }

    return ABM_SUCCESS;
}


ABM_Results_t ABM_Connect
(
    ABM_AudioPluginParams_t AudioPluginParams, GHashTable *pAudioParamsTable, ABM_AudioSetupCb_t pConnectCb
)
{
    PluginData_t *pPluginData;

    pPluginData = ABM_PluginStore_GetPlugin (AudioPluginParams.PluginType);

    if (pPluginData && pPluginData->Loaded)
    {
        PLUGIN_DATA_FUNC_TABLE->FnConnectChannels (AudioPluginParams, pAudioParamsTable, pConnectCb);
    }
    else
    {
        ABM_PRINT_ERROR_LOG (
            G_STRLOC "Connect failed, unsupported Plugin Type: %s\n", AudioPluginParams.PluginType);
        return ABM_FAILURE;
    }

    return ABM_SUCCESS;
}


ABM_Results_t ABM_Disconnect
(
    ABM_AudioPluginParams_t AudioPluginParams, GHashTable *AudioParamsTable, ABM_AudioSetupCb_t pDisconnectCb
)
{
    PluginData_t *pPluginData;

    pPluginData = ABM_PluginStore_GetPlugin (AudioPluginParams.PluginType);

    if (pPluginData && pPluginData->Loaded)
    {
        PLUGIN_DATA_FUNC_TABLE->FnDisconnectChannels (AudioPluginParams, AudioParamsTable, pDisconnectCb);
    }
    else
    {
        ABM_PRINT_ERROR_LOG (
            G_STRLOC "Disconnect failed, unsupported Plugin Type: %s\n", AudioPluginParams.PluginType);
        return ABM_FAILURE;
    }

    return ABM_SUCCESS;
}
/*--------------------------------------------------------------------------
 * Plugin Store Functions
 *--------------------------------------------------------------------------
 */
PluginData_t *ABM_PluginStore_GetPlugin
(
    gchar *pPluginType
)
{
    PluginData_t *pData;
    int i;

    for (i = 0; i < pABMPluginStore->len; ++i)
    {
        pData = &g_array_index (pABMPluginStore, PluginData_t, i);
        if (strcmp (pPluginType, pData->PluginType) == 0)
        {
            return pData;
        }
    }

    return NULL;
}


PluginData_t *ABM_PluginStore_GetPluginByIndex
(
    guint Index
)
{
    PluginData_t *pData;
    pData = &g_array_index (pABMPluginStore, PluginData_t, Index);
    return pData;
}


gboolean ABM_PluginStore_Initialize (
)
{
    PluginData_t PluginData;
    guint i;

    /*Allocate plugin structures and store them in a global array*/
    if (!pABMPluginStore)
    {
        pABMPluginStore = g_array_new (FALSE, TRUE, sizeof (PluginData_t));
        if (!pABMPluginStore)
        {
            ABM_PRINT_ERROR_LOG (
                G_STRLOC "pABMPluginArray Allocation failed\n");
            return FALSE;
        }
        memset (&PluginData, 0, sizeof (PluginData_t));
        for (i = 0; i < ABM_MAX_PLUGINS; ++i)
        {
            PluginData.Loaded = FALSE;
            PluginData.pPluginFnTable = malloc (sizeof (ABM_PluginFunctions_t));
            if (!PluginData.pPluginFnTable)
            {
                ABM_PRINT_ERROR_LOG (
                    G_STRLOC "ABM_PluginFunctions_t Allocation failed\n");
                return FALSE;
            }
            PluginData.pModule = NULL;
            PluginData.pInitFn = NULL;
            g_array_append_val (pABMPluginStore, PluginData);
        }
    }

    return TRUE;
} 

gboolean ABM_PluginStore_Terminate (
)
{
    PluginData_t *pPluginData;
    int i;

    if (pABMPluginStore)
    {
        for (i = 0; i < pABMPluginStore->len; ++i)
        {
            pPluginData = &g_array_index (pABMPluginStore, PluginData_t, i);
            if (pPluginData->Loaded)
            {
                if (((ABM_PluginFunctions_t *)pPluginData->pPluginFnTable)->FnTerminate)
                {
                    ((ABM_PluginFunctions_t *)pPluginData->pPluginFnTable)->FnTerminate ();
                }
                PM_UnloadPlugin (pPluginData);
            }
            if (pPluginData->pPluginFnTable)
            {
                free ((ABM_PluginFunctions_t *)pPluginData->pPluginFnTable);
            }
        }
        g_array_free (pABMPluginStore, TRUE);
        pABMPluginStore = NULL;
    }

    return TRUE;
}
/// @endcond
