/*
 * AudioBusManagerUtils.c 
 *
 * Copyright (c) 2008 Wind River Systems, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Lesser GNU General Public License version 2.1 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the Lesser GNU General Public License for more details.
 *
 * You should have received a copy of the Lesser GNU General Public License version 2.1
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */
/// @cond internal_documentation
/**
 * \file 
 *    Audio Bus Manager utility functions.
 *
 * \version $Revision: 1.5 $
 * \date    $Date: 2008/07/12 00:59:38 $
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <glib.h>
#include <glib/gprintf.h>
#include <gmodule.h>
#include <errno.h>

#include "AudioBusManagerPlugin.h"
#include "paclient.h"

#define ABMU_ALSA_DEVICE_STORE "/proc/asound/devices"
#define ABMU_ALSA_CARDS_STORE  "/proc/asound/cards"

#define ABMU_ALSA_DEVICE_DESC_LENGTH 64
#define ABMU_ALSA_CARDS_DESC_LENGTH 64
#define ABMU_ALSA_DEV_SINK_IDENTIFIER    "digital audio playback"
#define ABMU_ALSA_DEV_SOURCE_IDENTIFIER  "digital audio capture"

#define ABMU_PA_ALSA_SINK_MODULE_NAME   "module-alsa-sink"
#define ABMU_PA_ALSA_SOURCE_MODULE_NAME "module-alsa-source"
#define ABMU_MOST_ALSA_CARD_IDENTIFIER  "Most"
#define ABMU_IS_MOST(x) \
       ((strcmp(x, "most") == 0) ? TRUE:FALSE )

ABM_Results_t ABMU_PulseAudioModuleLoad
(
    PA_Client_Ctx_t * pCtx, int IsSink, unsigned int Device, unsigned int SubDevice, int PluginParamIndex,
    ABMU_PA_ModuleLoadHandler pModuleLoadHandler
);


int ABMU_TestStrEndswith
(
    const char *s, const char *sfx
)
{
    size_t l1, l2;

    l1 = strlen (s);
    l2 = strlen (sfx);

    return (l1 >= l2 && strcmp (s + l1 - l2, sfx) == 0);
}


ABM_Results_t ABMU_PulseAudioModuleSetup
(
    int MaxAlsaDevices, /**< Maximum Alsa devicec to be setup with 
                            pulse audio*/
    gchar *pPluginType,
    ABMU_PA_ModuleLoadHandler pModuleLoadHandler
)
{
    /* "JustOne" is a flag to indicate if only one device per device type( sin or source)
     * is to be supported
     */
    int JustOne = 0;

    FILE *pFile;
    int DeviceCount = 0, SinkDeviceCount = 0, SourceDeviceCount = 0;
    int MOSTDevice = -1;
    gboolean FoundMostCard = FALSE;
    char CardName[ABMU_ALSA_CARDS_DESC_LENGTH];
    PA_Client_Ctx_t Ctx;

    if (!(pFile = fopen (ABMU_ALSA_CARDS_STORE, "r")))
    {
        if (errno != ENOENT)
            g_printf ("open( %s) failed\n", ABMU_ALSA_CARDS_STORE);
        return ABM_FAILURE;
    }

    while (!feof (pFile))
    {
        char CardLine[ABMU_ALSA_CARDS_DESC_LENGTH];
        if (!fgets (CardLine, sizeof (CardLine), pFile))
            break;
        CardLine[strcspn (CardLine, "\r\n")] = 0;
        if (sscanf (CardLine, " %i [%*s            ]: %s ", &MOSTDevice, CardName) != 2)
            continue;
        if (strcmp (CardName, ABMU_MOST_ALSA_CARD_IDENTIFIER) == 0)
        {
            FoundMostCard = TRUE;
            g_printf ("MOST Card found,  MOSTDevice : %d, CardName: %s\n", MOSTDevice, CardName);
            break;
        }
    }

    fclose (pFile);
    pFile = NULL;

    if (ABMU_IS_MOST (pPluginType) && (!FoundMostCard))
    {
        g_printf ("Failed to find MOST Device in ( %s)\n", ABMU_ALSA_CARDS_STORE);
        return ABM_FAILURE;
    }
    /* Now load the Pulse audio modules for the Alsa devices for the given plugin type
     */
    if (PAClientCtxInit (&Ctx, "ABM_Plugin") != PA_CLIENT_OK)
    {
        g_printf ("PAClientCtxInit Fails\n");
        PAClientCtxCleanup (&Ctx);
        return ABM_FAILURE;
    }

    if (PAClientCtxConnect (&Ctx) != PA_CLIENT_OK)
    {
        g_printf ("PAClientCtxConnect Fails\n");
        return ABM_FAILURE;
    }

    if (!(pFile = fopen (ABMU_ALSA_DEVICE_STORE, "r")))
    {
        if (errno != ENOENT)
            g_printf ("open( %s) failed\n", ABMU_ALSA_DEVICE_STORE);
        PAClientCtxCleanup (&Ctx);
        return ABM_FAILURE;
    }

    while (!feof (pFile))
    {
        char DeviceLine[ABMU_ALSA_DEVICE_DESC_LENGTH];
        unsigned int Device, SubDevice;
        int IsSink;
        if (DeviceCount >= MaxAlsaDevices)
        {
            g_printf ("MAX Device count reached %d\n", DeviceCount);
            break;
        }
        if (!fgets (DeviceLine, sizeof (DeviceLine), pFile))
            break;
        DeviceLine[strcspn (DeviceLine, "\r\n")] = 0;

        /* g_printf("DeviceLine: %s\n",DeviceLine);*/

        if (ABMU_TestStrEndswith (DeviceLine, ABMU_ALSA_DEV_SINK_IDENTIFIER))
            IsSink = 1;

        else if (ABMU_TestStrEndswith (DeviceLine, ABMU_ALSA_DEV_SOURCE_IDENTIFIER))
            IsSink = 0;

        else
            continue;
        if (JustOne && IsSink && (SinkDeviceCount >= 1))
            continue;
        if (JustOne && !IsSink && (SourceDeviceCount >= 1))
            continue;
        if (sscanf (DeviceLine, " %*i: [%u- %u]: ", &Device, &SubDevice) != 2)
            continue;

        /* g_printf("Devices: %d, %d, %d\n ",MOSTDevice,Device, SubDevice);*/

        if (ABMU_IS_MOST (pPluginType) && (Device != MOSTDevice))
            continue;
        g_printf ("ALSA Device found, %s: %s\n", IsSink ? "module-alsa-sink" : "module-alsa-source", DeviceLine);
        if (ABMU_PulseAudioModuleLoad (&Ctx, IsSink, Device, SubDevice, DeviceCount, pModuleLoadHandler) != ABM_SUCCESS)
        {
            continue;
        }
        DeviceCount++;
        if (IsSink)
            SinkDeviceCount++;
        else
            SourceDeviceCount++;
    }

    fclose (pFile);

    PAClientCtxCleanup (&Ctx);
    return DeviceCount;
}


ABM_Results_t ABMU_PulseAudioModuleLoad
(
    PA_Client_Ctx_t *pCtx, int IsSink, unsigned int Device, unsigned int SubDevice, int PluginParamIndex,
    ABMU_PA_ModuleLoadHandler pModuleLoadHandler
)
{
    unsigned int PA_AlsaModuleIndex;
    unsigned int PA_AlsaDeviceIndex;
    char DeviceArgs[ABMU_ALSA_DEVICE_DESC_LENGTH];

    snprintf (DeviceArgs, sizeof (DeviceArgs), "device=hw:%u,%u", Device, SubDevice);

    /* Load the pulse audio alsa sink or source module for the given Alsa device,
     * identified by DeviceArgs
     */
    if (PAClientCtxLoadModule (pCtx,       IsSink ? ABMU_PA_ALSA_SINK_MODULE_NAME : ABMU_PA_ALSA_SOURCE_MODULE_NAME,
                               DeviceArgs, &PA_AlsaModuleIndex) != PA_CLIENT_OK)
    {
        g_printf ("PAClientCtxLoadModule Fails %s \n", DeviceArgs);
        return ABM_FAILURE;
    }

    if (PAClientGetDevFromModule (pCtx, PA_AlsaModuleIndex, IsSink ? PA_CLIENT_DEVICE_SINK : PA_CLIENT_DEVICE_SOURCE,
                                  &PA_AlsaDeviceIndex) != PA_CLIENT_OK)
    {
        g_printf ("PAClientGetDevFromModule Fails %s \n", DeviceArgs);
        return ABM_FAILURE;
    }

    pModuleLoadHandler (PluginParamIndex, PA_AlsaModuleIndex, PA_AlsaDeviceIndex, IsSink, Device, SubDevice);

    return ABM_SUCCESS;
} ABM_Results_t ABMU_PulseAudioModuleTerminate
(
    int ModuleIndex
)
{
    PA_Client_Ctx_t Ctx;

    g_printf ("%s called\n", __FUNCTION__);

    if (PAClientCtxInit (&Ctx, "ABM_Plugin") != PA_CLIENT_OK)
    {
        g_printf ("PAClientCtxInit Fails\n");
        return ABM_FAILURE;
    }

    if (PAClientCtxConnect (&Ctx) != PA_CLIENT_OK)
    {
        g_printf ("PAClientCtxConnect Fails\n");
        PAClientCtxCleanup (&Ctx);
        return ABM_FAILURE;
    }

    if (PAClientCtxUnloadModule (&Ctx, ModuleIndex) != PA_CLIENT_OK)
    {
        g_printf ("PAClientCtxUnloadModule Fails\n");
        PAClientCtxCleanup (&Ctx);
        return ABM_FAILURE;
    }

    PAClientCtxCleanup (&Ctx);
    return ABM_SUCCESS;
}
/// @endcond
