/*
 * local.c 
 *
 * Copyright (c) 2008 Wind River Systems, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Lesser GNU General Public License version 2.1 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the Lesser GNU General Public License for more details.
 *
 * You should have received a copy of the Lesser GNU General Public License version 2.1
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */
/// @cond internal_documentation
/**
 * \file
 *    Audio Manager plugin for Local Applications
 *
 *  
 *
 * \version $Revision: 1.8 $
 * \date    $Date: 2008/07/12 00:59:38 $
 */

/**************************************************************************/
/*                              Includes                                  */
/**************************************************************************/
#include <unistd.h> /* usleep() */
#include <glib.h>
#include <glib/gprintf.h>
#include <memory.h>
#include <errno.h>

#include "AudioBusManagerPlugin.h"
#include "paclient.h"

/**************************************************************************/
/*                               Defines                                  */
/**************************************************************************/
// Define the symbol LOCAL_ERROR_LOG from either in this header file of
// in the make file to enable all error logging.
#define LOCAL_ERROR_LOG
#define LOCAL_DEBUG_INFO

/** Debug print method */

#ifdef LOCAL_DEBUG_INFO
#define LOCAL_PRINT_INFO g_printf
#else
#define LOCAL_PRINT_INFO(fmt, arg...)
#endif

/** Debug print method */

#ifdef LOCAL_ERROR_LOG
#define LOCAL_PRINT_ERROR_LOG g_critical
#else
#define LOCAL_PRINT_ERROR_LOG(fmt, arg...)
#endif
#define LOCAL_AUDIO_PARAM_TEMP_BUF_SIZE 10

#define LOCAL_DEV_SINK_IDENTIFIER    "local digital audio playback"
#define LOCAL_DEV_SOURCE_IDENTIFIER  "local digital audio capture"

#define LOCAL_PA_MODULE_NAME   "module-null-sink"

#define LOCAL_AUDIO_PARAM_TABLE_KEY_CHANNEL_ID  "channel_id"

#define LOCAL_PLUGIN_VERSION 1
#define LOCAL_BUS_NAME "local"

/**************************************************************************/
/*                              Typedefs                                  */
/**************************************************************************/

typedef struct
{
    uint32_t ModuleIndex;
    uint32_t SinkIndex;
    uint32_t SourceIndex;
} Local_Channel_t;

/**************************************************************************/
/*                              Globals                                   */
/**************************************************************************/
static PA_Client_Ctx_t sPAClientCtx;
static GSList *sChannelList;

/**************************************************************************/
/*  Private Functions Prototypes-functions used internally in this file   */
/**************************************************************************/
static gboolean LOCAL_Terminate
(
    void
);


static void LOCAL_AllocateChannels
(
    ABM_AudioPluginParams_t AudioPluginParams, GHashTable *pAudioParamsTable, ABM_AudioSetupCb_t pResultCb
);


static void LOCAL_DeAllocateChannels
(
    ABM_AudioPluginParams_t AudioPluginParams, GHashTable *pAudioParamsTable, ABM_AudioSetupCb_t pResultCb
);


static void LOCAL_ConnectChannels
(
    ABM_AudioPluginParams_t AudioPluginParams, GHashTable *pAudioParamsTable, ABM_AudioSetupCb_t pResultCb
);


static void LOCAL_DisconnectChannels
(
    ABM_AudioPluginParams_t AudioPluginParams, GHashTable *pAudioParamsTable, ABM_AudioSetupCb_t pResultCb
);


gboolean ABM_PluginInit
(
    PluginData_t *pPluginData
);


static int LOCAL_PulseAudioModuleSetup (
);

/**************************************************************************/
/*                       Private Functions Bodies                         */
/**************************************************************************/

gboolean ABM_PluginInit
(
    PluginData_t *pPluginData
)
{
    PLUGIN_DATA_FUNC_TABLE->FnTerminate = LOCAL_Terminate;
    PLUGIN_DATA_FUNC_TABLE->FnAllocateChannels = LOCAL_AllocateChannels;
    PLUGIN_DATA_FUNC_TABLE->FnDeAllocateChannels = LOCAL_DeAllocateChannels;
    PLUGIN_DATA_FUNC_TABLE->FnConnectChannels = LOCAL_ConnectChannels;
    PLUGIN_DATA_FUNC_TABLE->FnDisconnectChannels = LOCAL_DisconnectChannels;

    strcpy (pPluginData->PluginType, LOCAL_BUS_NAME);

    pPluginData->Version = LOCAL_PLUGIN_VERSION;

    /* Intialize the plugin by loading PA ALSA Module for each Alsa device 
     * detected.
     */

    LOCAL_PulseAudioModuleSetup ();

    return TRUE;
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: LOCAL_Terminate                                             */
/*                                                                        */
/**************************************************************************/
/**
 *  \brief De- intializes the plugin
 *  \return TRUE - success
 **************************************************************************
*/
gboolean LOCAL_Terminate
(
    void
)
{
    GSList *tmpList = sChannelList;
    Local_Channel_t *pChannel;

    while (tmpList != NULL)
    {
        pChannel = (Local_Channel_t *)tmpList->data;
        if (pChannel)
        {
            /* remove channel */

            if (PAClientCtxUnloadModule (&sPAClientCtx, pChannel->ModuleIndex) != PA_CLIENT_OK)
            {
                LOCAL_PRINT_ERROR_LOG ("PAClientCtxUnloadModule Fails\n");
            }
            free (pChannel);
            break;
        }
        tmpList = g_slist_next (tmpList);
    }

    g_slist_free (sChannelList);
    sChannelList = NULL;
    PAClientCtxCleanup (&sPAClientCtx);
    return TRUE;
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: LOCAL_AllocateChannels                                      */
/*                                                                        */
/**************************************************************************/
/**
 *  \brief Allocate LOCAL bus channels
 *
 *  \return ABM_FAILURE - failure
 *          ABM_SUCCESS - success
 **************************************************************************
*/
void LOCAL_AllocateChannels
(
    ABM_AudioPluginParams_t AudioPluginParams, GHashTable *pAudioParamsTable, ABM_AudioSetupCb_t pResultCb
)
{
    GHashTable *pAudioParamReturnTable;
    int IsRequestedDeviceSink;
    uint32_t ModuleIndex;
    uint32_t SinkIndex;
    uint32_t SourceIndex;
    Local_Channel_t *pChannel;
    GSList *tmpList;
    gboolean FoundModule;
    gboolean success = TRUE;
    gchar TempBuffer[LOCAL_AUDIO_PARAM_TEMP_BUF_SIZE];

    pAudioParamReturnTable = g_hash_table_new (g_str_hash, g_str_equal);

    /* If the client is an Audio source, then we need a PulseAudio Sink for the application to use */

    IsRequestedDeviceSink = !(ABM_AUDIO_IS_SINK (AudioPluginParams.ClientType));

    if (PAClientCheckConnection (&sPAClientCtx) != PA_CLIENT_OK)
    {
        LOCAL_PRINT_ERROR_LOG ("Detected failed PA connection\n");
        success = FALSE;
    }

    if (success)
    {
        if (PAClientCtxLoadModule (&sPAClientCtx, LOCAL_PA_MODULE_NAME, NULL, &ModuleIndex) != PA_CLIENT_OK)
        {
            LOCAL_PRINT_ERROR_LOG ("PAClientCtxLoadModule Fails\n");
            success = FALSE;
        }
    }

    if (success)
    {
        if (PAClientGetDevFromModule (&sPAClientCtx, ModuleIndex, PA_CLIENT_DEVICE_SINK, &SinkIndex) != PA_CLIENT_OK)
        {
            LOCAL_PRINT_ERROR_LOG ("PAClientGetDevFromModule Fails\n");
            success = FALSE;
        }
    }

    if (success)
    {
        if (PAClientGetDevFromModule (&sPAClientCtx, ModuleIndex, PA_CLIENT_DEVICE_SOURCE, &SourceIndex)
                != PA_CLIENT_OK)
        {
            LOCAL_PRINT_ERROR_LOG ("PAClientGetDevFromModule Fails\n");
            success = FALSE;
        }
    }

    if (success)
    {
        FoundModule = FALSE;
        tmpList = sChannelList;
        while (tmpList != NULL)
        {
            pChannel = (Local_Channel_t *)tmpList->data;
            if (pChannel->ModuleIndex == ModuleIndex)
            {
                pChannel->SinkIndex = SinkIndex;
                pChannel->SourceIndex = SourceIndex;
                FoundModule = TRUE;
            }
            tmpList = g_slist_next (tmpList);
        }
        if (FoundModule == FALSE)
        {
            pChannel = (Local_Channel_t *)malloc (sizeof (Local_Channel_t));
            if (NULL == pChannel)
            {
                success = FALSE;
            }
            else
            {
                pChannel->ModuleIndex = ModuleIndex;
                pChannel->SinkIndex = SinkIndex;
                pChannel->SourceIndex = SourceIndex;
                sChannelList = g_slist_prepend (sChannelList, pChannel);
            }
        }
    }

    if (success)
    {
        /* Fill in the opague data block to be returned to clients.  Treat ModuleIndex as channel ID */
        g_sprintf (TempBuffer, "%d", ModuleIndex);

        g_hash_table_insert (pAudioParamReturnTable, LOCAL_AUDIO_PARAM_TABLE_KEY_CHANNEL_ID, TempBuffer);
    }

    pResultCb (AudioPluginParams, success, pAudioParamReturnTable, IsRequestedDeviceSink ? SinkIndex : SourceIndex);

    g_hash_table_unref (pAudioParamReturnTable);
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: LOCAL_DeAllocateChannels                                 */
/*                                                                        */
/**************************************************************************/
/**
 *  \brief Deallocate previously allocated channels
 *
 *  \return ABM_FAILURE - failure
 *          ABM_SUCCESS - success
 **************************************************************************
*/
void LOCAL_DeAllocateChannels
(
    ABM_AudioPluginParams_t AudioPluginParams, GHashTable *pAudioParamsTable, ABM_AudioSetupCb_t pResultCb
)
{
    uint32_t ModuleIndex;
    GSList *tmpList;
    Local_Channel_t *pChannel;
    GHashTable *pAudioParamReturnTable;
    gchar *ChannelID;
    gboolean success = TRUE;

    /* Get the channnel ID from the pAudioParamsTable */
    ChannelID = g_hash_table_lookup (pAudioParamsTable, LOCAL_AUDIO_PARAM_TABLE_KEY_CHANNEL_ID);

    if (NULL == ChannelID)
    {
        success = FALSE;
    }

    if (success)
    {
        ModuleIndex = atoi (ChannelID);

        /* remove channel from list, unload module */

        tmpList = sChannelList;
        while (tmpList != NULL)
        {
            pChannel = (Local_Channel_t *)tmpList->data;
            if (pChannel->ModuleIndex == ModuleIndex)
            {
                /* remove channel */
                sChannelList = g_slist_remove (sChannelList, pChannel);
                free (pChannel);
                break;
            }
            tmpList = g_slist_next (tmpList);
        }
        if (PAClientCtxUnloadModule (&sPAClientCtx, ModuleIndex) != PA_CLIENT_OK)
        {
            LOCAL_PRINT_ERROR_LOG ("PAClientCtxUnloadModule Fails\n");
            success = FALSE;
        }
    }

    pAudioParamReturnTable = g_hash_table_new (g_str_hash, g_str_equal);

    pResultCb (AudioPluginParams, success, pAudioParamReturnTable, ABM_INVALID_PA_INDEX);

    g_hash_table_unref (pAudioParamReturnTable);
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: sConnectChannels                                       */
/*                                                                        */
/**************************************************************************/
/**
 *  \brief Connect to Local Channel
 *
 *  \return ABM_FAILURE - failure
 *          ABM_SUCCESS - success
 **************************************************************************
*/
void LOCAL_ConnectChannels
(
    ABM_AudioPluginParams_t AudioPluginParams, GHashTable *pAudioParamsTable, ABM_AudioSetupCb_t pResultCb
)
{
    uint32_t ModuleIndex;
    uint32_t DeviceIndex = 0;
    GSList *tmpList;
    Local_Channel_t *pChannel;
    GHashTable *pAudioParamReturnTable;
    gchar *ChannelID;
    gboolean IsRequestedDeviceSink;
    gboolean FoundChannel;
    gboolean success = TRUE;

    /* Get the channnel ID from the pAudioParamsTable */
    ChannelID = g_hash_table_lookup (pAudioParamsTable, LOCAL_AUDIO_PARAM_TABLE_KEY_CHANNEL_ID);

    if (NULL == ChannelID)
    {
        success = FALSE;
    }

    if (success)
    {
        ModuleIndex = atoi (ChannelID);

        IsRequestedDeviceSink = !(ABM_AUDIO_IS_SINK (AudioPluginParams.ClientType));

        /* find device for module*/

        tmpList = sChannelList;
        FoundChannel = FALSE;
        while (tmpList != NULL)
        {
            pChannel = (Local_Channel_t *)tmpList->data;
            if (pChannel->ModuleIndex == ModuleIndex)
            {
                FoundChannel = TRUE;
                if (IsRequestedDeviceSink)
                {
                    DeviceIndex = pChannel->SinkIndex;
                }
                else
                {
                    DeviceIndex = pChannel->SourceIndex;
                }
                break;
            }
            tmpList = g_slist_next (tmpList);
        }
        if (!FoundChannel)
        {
            success = FALSE;
        }
    }

    pAudioParamReturnTable = g_hash_table_new (g_str_hash, g_str_equal);

    if (success)
    {
        g_hash_table_insert (pAudioParamReturnTable, LOCAL_AUDIO_PARAM_TABLE_KEY_CHANNEL_ID, ChannelID);
    }

    pResultCb (AudioPluginParams, success, pAudioParamReturnTable, DeviceIndex);

    g_hash_table_unref (pAudioParamReturnTable);
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: sDisconnectChannels                                    */
/*                                                                        */
/**************************************************************************/
/**
 *  \brief Disconnect from previously connected channels
 *
 *  \return ABM_FAILURE - failure
 *          ABM_SUCCESS - success
 **************************************************************************
*/
void LOCAL_DisconnectChannels
(
    ABM_AudioPluginParams_t AudioPluginParams, GHashTable *pAudioParamsTable, ABM_AudioSetupCb_t pResultCb
)
{
    /* NOP for local bus */

    GHashTable *pAudioParamReturnTable;

    pAudioParamReturnTable = g_hash_table_new (g_str_hash, g_str_equal);

    pResultCb (AudioPluginParams, TRUE, pAudioParamReturnTable, ABM_INVALID_PA_INDEX);

    g_hash_table_unref (pAudioParamReturnTable);
} 

ABM_Results_t LOCAL_PulseAudioModuleSetup (
)
{
    if (PAClientCtxInit (&sPAClientCtx, "ABM_LOCAL_Plugin") != PA_CLIENT_OK)
    {
        LOCAL_PRINT_ERROR_LOG ("PAClientCtxInit Fails\n");
        return ABM_FAILURE;
    }

    if (PAClientCtxConnect (&sPAClientCtx) != PA_CLIENT_OK)
    {
        PAClientCtxCleanup (&sPAClientCtx);
        LOCAL_PRINT_ERROR_LOG ("PAClientCtxConnect Fails\n");
        return ABM_FAILURE;
    }

    return ABM_SUCCESS;
}
/// @endcond
