/*
 * most.c 
 *
 * Copyright (c) 2008 Wind River Systems, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Lesser GNU General Public License version 2.1 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the Lesser GNU General Public License for more details.
 *
 * You should have received a copy of the Lesser GNU General Public License version 2.1
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */
/// @cond internal_documentation
/**
 * \file
 *    Audio Manager plugin for MOST
 *
 *  
 *
 * \version $Revision: 1.17 $
 * \date    $Date: 2008/07/12 00:59:38 $
 */

/* Disable VMOST dependancy: */
/*#define DEBUG_NO_VMOST 1*/

/**************************************************************************/
/*                              Includes                                  */
/**************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h> /* usleep() */
#include <glib.h>
#include <glib/gprintf.h>
#include <memory.h>
#include <errno.h>


#ifndef DEBUG_NO_VMOST
#include <vmostClientLib.h>
#endif
#include "AudioBusManagerPlugin.h"
#include "AudioBusManagerUtils.h"
#include "paclient.h"

/**************************************************************************/
/*                               Defines                                  */
/**************************************************************************/
#define MOST_PLUGIN_VERSION 1

/* *** DEBUG_NO_VMOST *** */

#ifdef DEBUG_NO_VMOST
#define VMOST_OK              0
#define VMOST_SYNC_OK         0
#define VMOST_EVENT_STATE_OFF 0

#define VMost_Msg_t void
#define VMost_Sync_Res_e int

typedef struct
{
    unsigned int SourceDelay;
    unsigned int Count;
    unsigned short Id[32];
} VMost_Sync_List_t;
#define vmostConnect(A, B, C) VMOST_OK
#define vmostRegisterToEvents(A) VMOST_OK
#define vmostDisconnect()
#define vmostSyncChannelsAllocate(A, B, C, D, E) VMOST_SYNC_OK
#define vmostSyncChannelsConnect(A, B, C, D, E) VMOST_SYNC_OK
#define vmostSyncChannelsDeallocate(A, B, C) VMOST_SYNC_OK
#define vmostSyncChannelsDisonnect(A, B, C) VMOST_SYNC_OK
#endif                        /* *** DEBUG_NO_VMOST *** */
#define MAX_ALSA_DEVICES    4 /**< Max. number of ALSA devices supported (a Linux limitation) */
#define MAX_PLUGIN_DEVICES (2 * MAX_ALSA_DEVICES)
#define ALSA_SOURCE TRUE
#define ALSA_SINK   FALSE

#define SOURCE_OFFSET 0                  /**< Offset index of source devices inside @sPluginParam array */
#define SINK_OFFSET   (MAX_ALSA_DEVICES) /**< Offset index of sink devices inside @sPluginParam array */

#define MOST_AUDIO_PARAM_TABLE_KEY_CHANNEL_COUNT      "channel_count"
#define MOST_AUDIO_PARAM_TABLE_KEY_SOURCE_DELAY       "source_delay"
#define MOST_AUDIO_PARAM_TABLE_KEY_CHANNEL_ID_FORMAT  "channel_id-%d"
#define MOST_AUDIO_PARAM_TABLE_KEY_DEVICE_HANDLE      "dev_hnd"
#define MOST_AUDIO_PARAM_TABLE_KEY_CHANNEL_ID_KEY_LENGTH  20
#define MOST_AUDIO_PARAM_TEMP_BUF_SIZE 10

#define MOST_BUS_NAME "most"

#define MOST_AUDIO_CALL_ERROR_CALLBACK()                                               \
       {                                                                               \
           GHashTable *pAudioParamTable = g_hash_table_new(g_str_hash,g_str_equal);    \
           pResultCb(AudioPluginParams,FALSE, pAudioParamTable, ABM_INVALID_PA_INDEX); \
              g_hash_table_unref(pAudioParamTable);                                    \
              return;                                                                  \
        }

/**************************************************************************/
/*                              Typedefs                                  */
/**************************************************************************/

typedef struct
{
    unsigned int SourceDelay;
    unsigned int Count;
    unsigned short Id[ABM_MAX_CHANNELS];
} ChannelInfo_t;


typedef struct
{
    ChannelInfo_t ChannelInfo;
    ABM_PulseAudioDevIndex PA_ALSADeviceIndex;
    ABM_AlsaDevIndex AlsaDevIndex;
} AudioParams_t;

/** Parameters needed to handle each ALSA<=>MOST Sync channels */
typedef struct
{
    gboolean IsAlsaTaken;
    ABM_AudioSetupCb_t FnResultCb;
    AudioParams_t AudioParam;
    ABM_AudioPluginParams_t AudioPluginParams;
    unsigned int SyncHnd;
} PluginParam;

/**************************************************************************/
/*                              Globals                                   */
/**************************************************************************/
/** array holding data for each ALSA device either sink or source */
static PluginParam sPluginParam[MAX_PLUGIN_DEVICES];

/**************************************************************************/
/*  Private Functions Prototypes-functions used internally in this file   */
/**************************************************************************/
static gboolean sTerminate
(
    void
);


static void sAllocateChannels
(
    ABM_AudioPluginParams_t AudioPluginParams, GHashTable *pAudioParamsTable, ABM_AudioSetupCb_t pResultCb
);


static void sDeAllocateChannels
(
    ABM_AudioPluginParams_t AudioPluginParams, GHashTable *pAudioParamsTable, ABM_AudioSetupCb_t pResultCb
);


static void sConnectChannels
(
    ABM_AudioPluginParams_t AudioPluginParams, GHashTable *pAudioParamsTable, ABM_AudioSetupCb_t pResultCb
);


static void sDisconnectChannels
(
    ABM_AudioPluginParams_t AudioPluginParams, GHashTable *pAudioParamsTable, ABM_AudioSetupCb_t pResultCb
);


gboolean ABM_PluginInit
(
    PluginData_t *pPluginData
);


static void sModuleLoadHandler
(
    int PluginParamIndex, int PA_ModuleIndex, int PA_DeviceIndex, int IsSink, int AlsaDevice, int AlsaSubDevice
);

/**************************************************************************/
/*                       Private Functions Bodies                         */
/**************************************************************************/

/**************************************************************************/
/*                                                                        */
/*  Function Name: sHashTableInsert                                       */
/*                                                                        */
/**************************************************************************/
/**
 *  \brief Allocates space for pKey and pValue and inserts them into pTable
 *
 *  \return none
 **************************************************************************
*/
static void sHashTableInsert
(
    GHashTable *pTable, gchar *pKey, gchar *pValue
)
{
    gchar *pTabKey;
    gchar *pTabValue;

    pTabKey = g_malloc (strlen (pKey) + 1);
    pTabValue = g_malloc (strlen (pValue) + 1);

    strcpy (pTabKey, pKey);
    strcpy (pTabValue, pValue);

    g_hash_table_insert (pTable, pTabKey, pTabValue);
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: sHashTableGetParams                                    */
/*                                                                        */
/**************************************************************************/
/**
 *  \brief Finds proper plugin params based on the value of key 
 *         @MOST_AUDIO_PARAM_TABLE_KEY_DEVICE_HANDLE in hash table
 *
 *  \return pointer to PluginParams struct or NULL if not found
 **************************************************************************
*/
static PluginParam *sHashTableGetParams
(
    GHashTable *pTable
)
{
    gchar *pValue;
    guint ParamOffs;

    pValue = g_hash_table_lookup (pTable, MOST_AUDIO_PARAM_TABLE_KEY_DEVICE_HANDLE);

    if (NULL == pValue)
    {
        g_printf ("Error: parameter %s not found in hash table\n", MOST_AUDIO_PARAM_TABLE_KEY_DEVICE_HANDLE);
        return NULL;
    }

    ParamOffs = atoi (pValue);

    /* sanity check */
    if (ParamOffs >= MAX_PLUGIN_DEVICES)
    {
        g_printf ("Error: parameter value of %s is out of range\n", MOST_AUDIO_PARAM_TABLE_KEY_DEVICE_HANDLE);
        return NULL;
    }

    return &sPluginParam[ParamOffs];
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: sMostEvent                                             */
/*                                                                        */
/**************************************************************************/
/**
 *  \brief Callback function called when a MOST ring event occurs, like
 *         MOST ring broke, reestablish, ...
 *
 *  \return none
 **************************************************************************
*/
static void sMostEvent
(
    unsigned int Event
)
{
    switch (Event)
    {
        case VMOST_EVENT_STATE_OFF:

            /* most ring went in OFF state. In this case all active sound devices
             * are lost */
            memset (&sPluginParam, 0, sizeof (sPluginParam));
            break;

        default:
            ;
    }
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: sMostReceive                                           */
/*                                                                        */
/**************************************************************************/
/**
 *  \brief Receive MOST controll messages. Its an empty function just to 
 *         satisfy vmost init function
 *
 *  \return 0
 **************************************************************************
*/
static int sMostReceive
(
    VMost_Msg_t *pRxMsg, void *pUserData
)
{
    return 0;
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: sFindNextFreeAlsaDev                                   */
/*                                                                        */
/**************************************************************************/
/**
 *  \brief Finds next ALSA device which is not taken by this plugin
 *
 *  \return none
 **************************************************************************
*/
static int sFindNextFreeAlsaDev
(
    gboolean isSource, int startOffset
)
{
    guint i;

    for (i = startOffset; i < MAX_ALSA_DEVICES; i++)
    {
        if (isSource)
        {
            if (FALSE == sPluginParam[SOURCE_OFFSET + i].IsAlsaTaken)
            {
                return i;
            }
        }
        else if (FALSE == sPluginParam[SINK_OFFSET + i].IsAlsaTaken)
        {
            return i;
        }
    }

    return -1;
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: sVmostAllocCb                                          */
/*                                                                        */
/**************************************************************************/
/**
 *  \brief Callback from vmost for alloc/connect result
 *
 *  \return none
 **************************************************************************
*/
static void sVmostAllocCb
(
    unsigned int SyncHnd, VMost_Sync_Res_e Result, VMost_Sync_List_t *pChanList, void *pUserData
)
{
    PluginParam *pPluginParam;
    ChannelInfo_t *pChInfo;
    guint i, ParamOffs;
    GHashTable *pAudioParamTable;
    gchar ChannelIdKey[MOST_AUDIO_PARAM_TABLE_KEY_CHANNEL_ID_KEY_LENGTH];
    gchar Tempbuffer[MOST_AUDIO_PARAM_TEMP_BUF_SIZE];
    ABM_PulseAudioDevIndex PAIndex;

    pAudioParamTable = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);

    ParamOffs = (guint)pUserData;
    pPluginParam = &sPluginParam[ParamOffs];
    PAIndex = pPluginParam->AudioParam.PA_ALSADeviceIndex;

    if (VMOST_SYNC_OK != Result)
    {
        sHashTableInsert (pAudioParamTable, MOST_AUDIO_PARAM_TABLE_KEY_CHANNEL_COUNT, "0");

        pPluginParam->IsAlsaTaken = FALSE;
        pPluginParam->FnResultCb (pPluginParam->AudioPluginParams, FALSE, pAudioParamTable, PAIndex);

        g_hash_table_unref (pAudioParamTable);
        return;
    }

    g_sprintf (Tempbuffer, "%d", ParamOffs);
    sHashTableInsert (pAudioParamTable, MOST_AUDIO_PARAM_TABLE_KEY_DEVICE_HANDLE, Tempbuffer);

    pPluginParam->SyncHnd = SyncHnd;
    pChInfo = &pPluginParam->AudioParam.ChannelInfo;
    pChInfo->Count = pChanList->Count;
    pChInfo->SourceDelay = pChanList->SourceDelay;

    for (i = 0; i < pChanList->Count; i++)
    {
        pChInfo->Id[i] = pChanList->Id[i];
    }
    /* Fill in the opaque data block to be retured to the clients */
    g_sprintf (Tempbuffer, "%d", pChanList->Count);
    sHashTableInsert (pAudioParamTable, MOST_AUDIO_PARAM_TABLE_KEY_CHANNEL_COUNT, Tempbuffer);

    g_sprintf (Tempbuffer, "%d", pChanList->SourceDelay);
    sHashTableInsert (pAudioParamTable, MOST_AUDIO_PARAM_TABLE_KEY_SOURCE_DELAY, Tempbuffer);

    for (i = 0; i < pChanList->Count; i++)
    {
        g_sprintf (ChannelIdKey, MOST_AUDIO_PARAM_TABLE_KEY_CHANNEL_ID_FORMAT, i);
        g_sprintf (Tempbuffer, "%d", pChanList->Id[i]);
        sHashTableInsert (pAudioParamTable, ChannelIdKey, Tempbuffer);
    }

    pPluginParam->FnResultCb (pPluginParam->AudioPluginParams, TRUE, pAudioParamTable, PAIndex);

    g_hash_table_unref (pAudioParamTable);
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: sVmostResultCb                                         */
/*                                                                        */
/**************************************************************************/
/**
 *  \brief Callback from vmost for dealloc/disconnect result
 *
 *  \return none
 **************************************************************************
*/
static void sVmostResultCb
(
    unsigned int SyncHnd, VMost_Sync_Res_e Result, void *pUserData
)
{
    PluginParam *pPluginParam;
    GHashTable *pAudioParamTable;
    ABM_PulseAudioDevIndex PAIndex;

    pPluginParam = (PluginParam *)pUserData;

    pAudioParamTable = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);

    pPluginParam->IsAlsaTaken = FALSE;
    pPluginParam->AudioParam.ChannelInfo.Count = 0;
    pPluginParam->AudioParam.ChannelInfo.SourceDelay = 0;
    PAIndex = pPluginParam->AudioParam.PA_ALSADeviceIndex;

    /* client request channel count to be present in table */
    sHashTableInsert (pAudioParamTable, MOST_AUDIO_PARAM_TABLE_KEY_CHANNEL_COUNT, "0");

    if (VMOST_SYNC_OK != Result)
    {
        pPluginParam->FnResultCb (pPluginParam->AudioPluginParams, FALSE, pAudioParamTable, PAIndex);
    }
    else
    {
        pPluginParam->FnResultCb (pPluginParam->AudioPluginParams, TRUE, pAudioParamTable, PAIndex);
    }

    g_hash_table_unref (pAudioParamTable);
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: ABM_PluginInit                                         */
/*                                                                        */
/**************************************************************************/
/**
 *  \brief Plugin entry point and plugin initialization
 *
 *  \return FALSE - failure
 *          TRUE  - success
 **************************************************************************
*/
gboolean ABM_PluginInit
(
    PluginData_t *pPluginData
)
{
    PLUGIN_DATA_FUNC_TABLE->FnTerminate = sTerminate;
    PLUGIN_DATA_FUNC_TABLE->FnAllocateChannels = sAllocateChannels;
    PLUGIN_DATA_FUNC_TABLE->FnDeAllocateChannels = sDeAllocateChannels;
    PLUGIN_DATA_FUNC_TABLE->FnConnectChannels = sConnectChannels;
    PLUGIN_DATA_FUNC_TABLE->FnDisconnectChannels = sDisconnectChannels;

    strcpy (pPluginData->PluginType, MOST_BUS_NAME);

    pPluginData->Version = MOST_PLUGIN_VERSION;

    if (VMOST_OK != vmostConnect (sMostReceive, NULL, TRUE))
    {
        g_message ("Cannot connect to vmostd");
        return FALSE;
    }
    /* simulate NET OFF event to clear taken ALSA devices flags */
    sMostEvent (VMOST_EVENT_STATE_OFF);

    if (VMOST_OK != vmostRegisterToEvents (sMostEvent))
    {
        g_message ("Cannot register Event callback in vmostd");
        vmostDisconnect ();
        return FALSE;
    }
    /* Intialize the plugin by loading PA ALSA Module for each MOST Alsa device 
     * detected.
     */
    ABMU_PulseAudioModuleSetup (MAX_PLUGIN_DEVICES, MOST_BUS_NAME, sModuleLoadHandler);

    return TRUE;
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: sTerminate                                             */
/*                                                                        */
/**************************************************************************/
/**
 *  \brief Deinits plugin
 *
 *  \return TRUE - success
 **************************************************************************
*/
static gboolean sTerminate
(
    void
)
{
    vmostDisconnect ();

    return TRUE;
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: sAllocateChannels                                      */
/*                                                                        */
/**************************************************************************/
/**
 *  \brief Allocate MOST sync channels for the source fblock
 *
 *  \return none
 **************************************************************************
*/
static void sAllocateChannels
(
    ABM_AudioPluginParams_t AudioPluginParams, GHashTable *pAudioParamsTable, ABM_AudioSetupCb_t pResultCb
)
{
    PluginParam *pPluginParam;
    int AlsaId = 0;
    gchar *pChanCount = 0;
    unsigned int Count;
    guint ParamOffs;

    /* Get parameters needed by this plugin from the GHashtable block*/
    pChanCount = g_hash_table_lookup (pAudioParamsTable, MOST_AUDIO_PARAM_TABLE_KEY_CHANNEL_COUNT);

    if (NULL == pChanCount)
    {
        MOST_AUDIO_CALL_ERROR_CALLBACK ();
    }

    Count = atoi (pChanCount);
    g_printf ("sAllocateChannels: got chan count %d \n", Count);

    if (0 == Count) /* atoi failed and returned 0*/
    {
        MOST_AUDIO_CALL_ERROR_CALLBACK ();
    }

    while (AlsaId >= 0 && AlsaId < MAX_ALSA_DEVICES)
    {
        AlsaId = sFindNextFreeAlsaDev (ALSA_SOURCE, AlsaId);
        if (AlsaId < 0)
        {
            MOST_AUDIO_CALL_ERROR_CALLBACK ();
        }
        ParamOffs = SOURCE_OFFSET + AlsaId;
        pPluginParam = &sPluginParam[ParamOffs];

        pPluginParam->IsAlsaTaken = TRUE;
        pPluginParam->FnResultCb = pResultCb;
        pPluginParam->AudioParam.ChannelInfo.Count = Count;

        /* The AudioPluginParams is stored in here simply to pass it back to the caller (ACM),
         *  when the vmostSyncChannelsAllocate returns*/
        memcpy (&pPluginParam->AudioPluginParams, &AudioPluginParams, sizeof (ABM_AudioPluginParams_t));
        pPluginParam->AudioParam.AlsaDevIndex = AlsaId;
        if (VMOST_SYNC_OK != vmostSyncChannelsAllocate ((unsigned int)AlsaId, ALSA_SOURCE, Count, sVmostAllocCb,
                                                        (void *)ParamOffs))
        {
            /* In this case the sVmostAllocCb will NOT be called */
            /* ... try with next ALSA device */
            pPluginParam->IsAlsaTaken = FALSE;
            continue;
        }
        return;
    }
    /* no ALSA device is available therefore report an error */
    MOST_AUDIO_CALL_ERROR_CALLBACK ();
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: sDeAllocateChannels                                    */
/*                                                                        */
/**************************************************************************/
/**
 *  \brief Deallocate previously allocated channels
 *
 *  \return none
 **************************************************************************
*/
static void sDeAllocateChannels
(
    ABM_AudioPluginParams_t AudioPluginParams, GHashTable *pAudioParamsTable, ABM_AudioSetupCb_t pResultCb
)
{
    PluginParam *pPluginParam;

    pPluginParam = sHashTableGetParams (pAudioParamsTable);

    if (NULL == pPluginParam || FALSE == pPluginParam->IsAlsaTaken)
    {
        MOST_AUDIO_CALL_ERROR_CALLBACK ();
    }

    pPluginParam->FnResultCb = pResultCb;

    if (VMOST_SYNC_OK != vmostSyncChannelsDeallocate (pPluginParam->SyncHnd, sVmostResultCb, pPluginParam))
    {
        /* In this case the sVmostAllocCb will NOT be called */
        pPluginParam->IsAlsaTaken = FALSE;
        MOST_AUDIO_CALL_ERROR_CALLBACK ();
    }
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: sConnectChannels                                       */
/*                                                                        */
/**************************************************************************/
/**
 *  \brief Connect MOST sync channels for the sink fblock
 *
 *  \return none
 **************************************************************************
*/
static void sConnectChannels
(
    ABM_AudioPluginParams_t AudioPluginParams, GHashTable *pAudioParamsTable, ABM_AudioSetupCb_t pResultCb
)
{
    PluginParam *pPluginParam;
    int AlsaId = 0;
    guint i;
    guint ParamOffs;
    VMost_Sync_List_t VmostSyncList;
    gchar *pChanCount;
    gchar *pChanId;
    gchar *pSourceDelay;
    gchar ChannelIdKey[MOST_AUDIO_PARAM_TABLE_KEY_CHANNEL_ID_KEY_LENGTH];

    while (AlsaId >= 0 && AlsaId < MAX_ALSA_DEVICES)
    {
        AlsaId = sFindNextFreeAlsaDev (ALSA_SINK, AlsaId);
        if (AlsaId < 0)
        {
            MOST_AUDIO_CALL_ERROR_CALLBACK ();
        }
        ParamOffs = SINK_OFFSET + AlsaId;
        pPluginParam = &sPluginParam[ParamOffs];

        pPluginParam->IsAlsaTaken = TRUE;
        pPluginParam->FnResultCb = pResultCb;
        pPluginParam->AudioParam.AlsaDevIndex = AlsaId;

        pChanCount = g_hash_table_lookup (pAudioParamsTable, MOST_AUDIO_PARAM_TABLE_KEY_CHANNEL_COUNT);
        if (NULL == pChanCount)
        {
            pPluginParam->IsAlsaTaken = FALSE;
            MOST_AUDIO_CALL_ERROR_CALLBACK ();
        }
        pSourceDelay = g_hash_table_lookup (pAudioParamsTable, MOST_AUDIO_PARAM_TABLE_KEY_SOURCE_DELAY);
        if (NULL == pSourceDelay)
        {
            pPluginParam->IsAlsaTaken = FALSE;
            MOST_AUDIO_CALL_ERROR_CALLBACK ();
        }
        VmostSyncList.SourceDelay = atoi (pSourceDelay);
        VmostSyncList.Count = atoi (pChanCount);
        for (i = 0; i < VmostSyncList.Count; i++)
        {
            g_sprintf (ChannelIdKey, MOST_AUDIO_PARAM_TABLE_KEY_CHANNEL_ID_FORMAT, i);
            pChanId = g_hash_table_lookup (pAudioParamsTable, ChannelIdKey);
            if (NULL == pChanId)
            {
                pPluginParam->IsAlsaTaken = FALSE;
                MOST_AUDIO_CALL_ERROR_CALLBACK ();
            }
            VmostSyncList.Id[i] = atoi (pChanId);
        }
        if (VMOST_SYNC_OK != vmostSyncChannelsConnect ((unsigned int)AlsaId, ALSA_SINK, &VmostSyncList, sVmostAllocCb,
                                                       (void *)ParamOffs))
        {
            /* In this case the sVmostAllocCb will NOT be called */
            /* ... try with next ALSA device */
            pPluginParam->IsAlsaTaken = FALSE;
            continue;
        }
        return;
    }
    /* no ALSA device is available therefore report an error */
    MOST_AUDIO_CALL_ERROR_CALLBACK ();
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: sDisconnectChannels                                    */
/*                                                                        */
/**************************************************************************/
/**
 *  \brief Disconnect from previously connected channels
 *
 *  \return none
 **************************************************************************
*/
static void sDisconnectChannels
(
    ABM_AudioPluginParams_t AudioPluginParams, GHashTable *pAudioParamsTable, ABM_AudioSetupCb_t pResultCb
)
{
    PluginParam *pPluginParam;

    pPluginParam = sHashTableGetParams (pAudioParamsTable);

    if (NULL == pPluginParam || FALSE == pPluginParam->IsAlsaTaken)
    {
        MOST_AUDIO_CALL_ERROR_CALLBACK ();
    }

    pPluginParam->FnResultCb = pResultCb;

    if (VMOST_SYNC_OK != vmostSyncChannelsDisonnect (pPluginParam->SyncHnd, sVmostResultCb, pPluginParam))
    {
        /* In this case the sVmostAllocCb will NOT be called */
        pPluginParam->IsAlsaTaken = FALSE;
        MOST_AUDIO_CALL_ERROR_CALLBACK ();
    }
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: sModuleLoadHandler                                     */
/*                                                                        */
/**************************************************************************/
/**
 *  \brief Callback for PulseAudio device registration
 *
 *  \return none
 **************************************************************************
*/
static void sModuleLoadHandler
(
    int PluginParamIndex, int PA_ModuleIndex, int PA_DeviceIndex, int IsSink, int AlsaDevice, int AlsaSubDevice
)
{
    int Offset = 0;

    Offset = PluginParamIndex + ((IsSink) ? SINK_OFFSET : SOURCE_OFFSET);

    sPluginParam[Offset].IsAlsaTaken = FALSE;
    sPluginParam[Offset].AudioParam.PA_ALSADeviceIndex = PA_DeviceIndex;
}
/// @endcond
