/*
 * pcm.c 
 *
 * Copyright (c) 2008 Wind River Systems, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Lesser GNU General Public License version 2.1 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the Lesser GNU General Public License for more details.
 *
 * You should have received a copy of the Lesser GNU General Public License version 2.1
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */
/// @cond internal_documentation
/**
 * \file
 *    Audio Manager plugin for PCM
 *
 *  
 *
 * \version $Revision: 1.14 $
 * \date    $Date: 2008/07/12 00:59:39 $
 */

/**************************************************************************/
/*                              Includes                                  */
/**************************************************************************/
#include <unistd.h> /* usleep() */
#include <glib.h>
#include <glib/gprintf.h>
#include <memory.h>
#include <errno.h>

#include "AudioBusManagerPlugin.h"
#include "AudioBusManagerUtils.h"
#include "paclient.h"

// Define the symbol PCM_ERROR_LOG from either in this header file of
// in the make file to enable all error logging.
#define PCM_ERROR_LOG
#define PCM_DEBUG_INFO

/** Debug print method */

#ifdef PCM_DEBUG_INFO
#define PCM_PRINT_INFO g_printf
#else
#define PCM_PRINT_INFO(fmt, arg...)
#endif

/** Debug print method */

#ifdef PCM_ERROR_LOG
#define PCM_PRINT_ERROR_LOG g_critical
#else
#define PCM_PRINT_ERROR_LOG(fmt, arg...)
#endif

/**************************************************************************/
/*                               Defines                                  */
/**************************************************************************/
#define MAX_ALSA_DEVICES 10
#define PCM_ALSA_DEVICE_DESC_LENGTH 64
#define PCM_AUDIO_PARAM_TABLE_KEY_CHANNEL_ID  "channel_id"
#define PCM_PLUGIN_VERSION    1
#define PCM_BUS_NAME "pcm"

/**************************************************************************/
/*                              Typedefs                                  */
/**************************************************************************/
/** Parameters needed to handle each ALSA channels */

typedef struct
{
    char ChannelID[PCM_ALSA_DEVICE_DESC_LENGTH];
    unsigned int PA_ALSAModuleIndex;
    unsigned int PA_ALSADeviceIndex;
    int IsSink;
    int Active;
    int IsTaken;
    int ConnectionCount;
} PluginParam_t;

/**************************************************************************/
/*                              Globals                                   */
/**************************************************************************/
/** array holding data for each ALSA device either sink or source */
static PluginParam_t sPluginParamArray[MAX_ALSA_DEVICES];

/**************************************************************************/
/*  Private Functions Prototypes-functions used internally in this file   */
/**************************************************************************/
static gboolean PCM_Terminate
(
    void
);


static void PCM_AllocateChannels
(
    ABM_AudioPluginParams_t AudioPluginParams, GHashTable *pAudioParamsTable, ABM_AudioSetupCb_t pResultCb
);


static void PCM_DeAllocateChannels
(
    ABM_AudioPluginParams_t AudioPluginParams, GHashTable *pAudioParamsTable, ABM_AudioSetupCb_t pResultCb
);


static void PCM_ConnectChannels
(
    ABM_AudioPluginParams_t AudioPluginParams, GHashTable *pAudioParamsTable, ABM_AudioSetupCb_t pResultCb
);


static void PCM_DisconnectChannels
(
    ABM_AudioPluginParams_t AudioPluginParams, GHashTable *pAudioParamsTable, ABM_AudioSetupCb_t pResultCb
);


gboolean ABM_PluginInit
(
    PluginData_t *pPluginData
);


static void PCM_ModuleLoadHandler
(
    int PluginParamIndex, int PA_ModuleIndex, int PA_DeviceIndex, int IsSink, int AlsaDevice, int AlsaSubDevice
);

/**************************************************************************/
/*                       Private Functions Bodies                         */
/**************************************************************************/

gboolean ABM_PluginInit
(
    PluginData_t *pPluginData
)
{
    int i;

    PLUGIN_DATA_FUNC_TABLE->FnTerminate = PCM_Terminate;
    PLUGIN_DATA_FUNC_TABLE->FnAllocateChannels = PCM_AllocateChannels;
    PLUGIN_DATA_FUNC_TABLE->FnDeAllocateChannels = PCM_DeAllocateChannels;
    PLUGIN_DATA_FUNC_TABLE->FnConnectChannels = PCM_ConnectChannels;
    PLUGIN_DATA_FUNC_TABLE->FnDisconnectChannels = PCM_DisconnectChannels;

    pPluginData->Version = PCM_PLUGIN_VERSION;
    strcpy (pPluginData->PluginType, PCM_BUS_NAME);

    /* Initialize the sPluginArray*/
    for (i = 0; i < MAX_ALSA_DEVICES; i++)
    {
        sPluginParamArray[i].Active = FALSE;
        sPluginParamArray[i].IsTaken = FALSE;
    }
    /* Intialize the plugin by loading PA ALSA Module for each PCM Alsa device 
     * detected.
     */
    ABMU_PulseAudioModuleSetup (MAX_ALSA_DEVICES, PCM_BUS_NAME, PCM_ModuleLoadHandler);

    return TRUE;
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: PCM_Terminate                                             */
/*                                                                        */
/**************************************************************************/
/**
 *  \brief De- intializes the plugin
 *  \return TRUE - success
 **************************************************************************
*/
gboolean PCM_Terminate
(
    void
)
{
    int i;

    for (i = 0; i < MAX_ALSA_DEVICES; i++)
    {
        if (TRUE == sPluginParamArray[i].Active)
        {
            ABMU_PulseAudioModuleTerminate (sPluginParamArray[i].PA_ALSAModuleIndex);
        }
    }

    return TRUE;
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: PCM_AllocateChannels                                      */
/*                                                                        */
/**************************************************************************/
/**
 *  \brief Allocate MPCM Alsa channels
 *
 *  \return ABM_FAILURE - failure
 *          ABM_SUCCESS - success
 **************************************************************************
*/
void PCM_AllocateChannels
(
    ABM_AudioPluginParams_t AudioPluginParams, GHashTable *pAudioParamsTable, ABM_AudioSetupCb_t pResultCb
)
{
    int i = 0;
    gboolean success = FALSE;
    GHashTable *pAudioParamReturnTable;
    int IsRequestedDeviceSink;
    gchar *ChannelID;
    ABM_PulseAudioDevIndex PA_Index = ABM_INVALID_PA_INDEX;

    pAudioParamReturnTable = g_hash_table_new (g_str_hash, g_str_equal);

    IsRequestedDeviceSink = !(ABM_AUDIO_IS_SINK (AudioPluginParams.ClientType));

    /* Get the channnel ID from the pAudioParamsTable */
    ChannelID = g_hash_table_lookup (pAudioParamsTable, PCM_AUDIO_PARAM_TABLE_KEY_CHANNEL_ID);

    if (NULL == ChannelID)
    {
        success = FALSE;
    }

    for (i = 0; i < MAX_ALSA_DEVICES; i++)
    {
        g_printf ("sPluginParamArray: Active:%d, IsSink: %d, channelId: %s, IsTaken: %d\n", sPluginParamArray[i].Active, sPluginParamArray[i].IsSink, sPluginParamArray[i].ChannelID, sPluginParamArray[i].IsTaken);

        /* Locate the pa device index for the given channelID and client type*/
        if ((TRUE == sPluginParamArray[i].Active) && (IsRequestedDeviceSink == sPluginParamArray[i].IsSink) && (sPluginParamArray[i].IsTaken == FALSE) && (strcmp (sPluginParamArray[i].ChannelID, ChannelID) == 0))
        {
            success = TRUE;
            sPluginParamArray[i].IsTaken = TRUE;
            PA_Index = sPluginParamArray[i].PA_ALSADeviceIndex;
            g_hash_table_insert (pAudioParamReturnTable, PCM_AUDIO_PARAM_TABLE_KEY_CHANNEL_ID, sPluginParamArray[i].ChannelID);
            break;
        }
    }

    pResultCb (AudioPluginParams, success, pAudioParamReturnTable, PA_Index);
    g_hash_table_unref (pAudioParamReturnTable);
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: PCM_DeAllocateChannels                                 */
/*                                                                        */
/**************************************************************************/
/**
 *  \brief Deallocate previously allocated channels
 *
 *  \return ABM_FAILURE - failure
 *          ABM_SUCCESS - success
 **************************************************************************
*/
void PCM_DeAllocateChannels
(
    ABM_AudioPluginParams_t AudioPluginParams, GHashTable *pAudioParamsTable, ABM_AudioSetupCb_t pResultCb
)
{
    gchar *ChannelID;
    int i;
    int IsRequestedDeviceSink;
    gboolean success = FALSE;
    GHashTable *pAudioParamReturnTable;
    ABM_PulseAudioDevIndex PA_Index = ABM_INVALID_PA_INDEX;

    IsRequestedDeviceSink = !(ABM_AUDIO_IS_SINK (AudioPluginParams.ClientType));
    pAudioParamReturnTable = g_hash_table_new (g_str_hash, g_str_equal);

    /* Get the channnel ID from the pAudioParamsTable */
    ChannelID = g_hash_table_lookup (pAudioParamsTable, PCM_AUDIO_PARAM_TABLE_KEY_CHANNEL_ID);

    if (NULL == ChannelID)
    {
        success = FALSE;
    }
    else
    {
        for (i = 0; i < MAX_ALSA_DEVICES; i++)
        {
            /* Locate the pa device index for the given channelID and client type*/
            if ((TRUE == sPluginParamArray[i].Active) && (IsRequestedDeviceSink == sPluginParamArray[i].IsSink) && (strcmp (sPluginParamArray[i].ChannelID, ChannelID) == 0) && (sPluginParamArray[i].IsTaken == TRUE))
            {
                if (!sPluginParamArray[i].ConnectionCount)
                {
                    success = TRUE;
                    sPluginParamArray[i].IsTaken = FALSE;
                    PA_Index = sPluginParamArray[i].PA_ALSADeviceIndex;
                }
                else
                {
                    PCM_PRINT_ERROR_LOG ("%s -Dealloc failed, Channel: %s has %d active connections\n", __FUNCTION__, sPluginParamArray[i].ChannelID, sPluginParamArray[i].ConnectionCount);
                    success = FALSE;
                    PA_Index = sPluginParamArray[i].PA_ALSADeviceIndex;
                }
                break;
            }
        }
    }

    pResultCb (AudioPluginParams, success, pAudioParamReturnTable, PA_Index);
    g_hash_table_unref (pAudioParamReturnTable);
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: sConnectChannels                                       */
/*                                                                        */
/**************************************************************************/
/**
 *  \brief Connect MOST sync channels for the sink fblock
 *
 *  \return ABM_FAILURE - failure
 *          ABM_SUCCESS - success
 **************************************************************************
*/
void PCM_ConnectChannels
(
    ABM_AudioPluginParams_t AudioPluginParams, GHashTable *pAudioParamsTable, ABM_AudioSetupCb_t pResultCb
)
{
    gchar *ChannelID;
    int i;
    int IsRequestedDeviceSink;
    gboolean success = FALSE;
    GHashTable *pAudioParamReturnTable;
    ABM_PulseAudioDevIndex PA_Index = ABM_INVALID_PA_INDEX;

    pAudioParamReturnTable = g_hash_table_new (g_str_hash, g_str_equal);

    IsRequestedDeviceSink = !(ABM_AUDIO_IS_SINK (AudioPluginParams.ClientType));

    /* Get the channnel ID from the pAudioParamsTable */
    ChannelID = g_hash_table_lookup (pAudioParamsTable, PCM_AUDIO_PARAM_TABLE_KEY_CHANNEL_ID);

    if (NULL == ChannelID)
    {
        success = FALSE;
    }
    else
    {
        /* Locate the pa device index for the given channelID and client type*/
        for (i = 0; i < MAX_ALSA_DEVICES; i++)
        {
            /*g_debug("sPluginParamArray[i].ChannelID: %s,sPluginParamArray[i].IsSink : %d ChannelID: %s\n",
                    sPluginParamArray[i].ChannelID, sPluginParamArray[i].IsSink, ChannelID);
            */
            if ((TRUE == sPluginParamArray[i].Active) && (IsRequestedDeviceSink == sPluginParamArray[i].IsSink) && (strcmp (sPluginParamArray[i].ChannelID, ChannelID) == 0) && (sPluginParamArray[i].IsTaken == TRUE))
            {
                success = TRUE;
                PA_Index = sPluginParamArray[i].PA_ALSADeviceIndex;
                sPluginParamArray[i].ConnectionCount++;
                g_hash_table_insert (pAudioParamReturnTable, PCM_AUDIO_PARAM_TABLE_KEY_CHANNEL_ID, ChannelID);
                break;
            }
        }
    }

    pResultCb (AudioPluginParams, success, pAudioParamReturnTable, PA_Index);
    g_hash_table_unref (pAudioParamReturnTable);
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: sDisconnectChannels                                    */
/*                                                                        */
/**************************************************************************/
/**
 *  \brief Disconnect from previously connected channels
 *
 *  \return ABM_FAILURE - failure
 *          ABM_SUCCESS - success
 **************************************************************************
*/
void PCM_DisconnectChannels
(
    ABM_AudioPluginParams_t AudioPluginParams, GHashTable *pAudioParamsTable, ABM_AudioSetupCb_t pResultCb
)
{
    gchar *ChannelID;
    int i;
    int IsRequestedDeviceSink;
    gboolean success = FALSE;
    GHashTable *pAudioParamReturnTable;
    ABM_PulseAudioDevIndex PA_Index = ABM_INVALID_PA_INDEX;

    pAudioParamReturnTable = g_hash_table_new (g_str_hash, g_str_equal);

    IsRequestedDeviceSink = !(ABM_AUDIO_IS_SINK (AudioPluginParams.ClientType));

    /* Get the channnel ID from the pAudioParamsTable */
    ChannelID = g_hash_table_lookup (pAudioParamsTable, PCM_AUDIO_PARAM_TABLE_KEY_CHANNEL_ID);

    if (NULL == ChannelID)
    {
        success = FALSE;
    }
    else
    {
        for (i = 0; i < MAX_ALSA_DEVICES; i++)
        {
            /* Locate the pa device index for the given channelID and client type*/
            if ((TRUE == sPluginParamArray[i].Active) && (IsRequestedDeviceSink == sPluginParamArray[i].IsSink) && (strcmp (sPluginParamArray[i].ChannelID, ChannelID) == 0) && (sPluginParamArray[i].IsTaken == TRUE))
            {
                success = TRUE;
                PA_Index = sPluginParamArray[i].PA_ALSADeviceIndex;
                sPluginParamArray[i].ConnectionCount--;
                break;
            }
        }
    }

    pResultCb (AudioPluginParams, success, pAudioParamReturnTable, PA_Index);

    g_hash_table_unref (pAudioParamReturnTable);
} 

void PCM_ModuleLoadHandler
(
    int PluginParamIndex, int PA_ModuleIndex, int PA_DeviceIndex, int IsSink, int AlsaDevice, int AlsaSubDevice
)
{
    sprintf (sPluginParamArray[PluginParamIndex].ChannelID, "hw:%u,%u", AlsaDevice, AlsaSubDevice);
    sPluginParamArray[PluginParamIndex].PA_ALSADeviceIndex = PA_DeviceIndex;
    sPluginParamArray[PluginParamIndex].PA_ALSAModuleIndex = PA_ModuleIndex;
    sPluginParamArray[PluginParamIndex].Active = TRUE;
    sPluginParamArray[PluginParamIndex].IsSink = IsSink;
    sPluginParamArray[PluginParamIndex].IsTaken = FALSE;
    sPluginParamArray[PluginParamIndex].ConnectionCount = 0;
}
/// @endcond
