/*
 * main.c 
 *
 * Copyright (c) 2008 Wind River Systems, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Lesser GNU General Public License version 2.1 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the Lesser GNU General Public License for more details.
 *
 * You should have received a copy of the Lesser GNU General Public License version 2.1
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */
/// @cond internal_documentation
/**
 * \file
 *    Audio manager test application for MOST plugin
 *
 *  
 *
 */

/**************************************************************************/
/*                              Includes                                  */
/**************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <memory.h>
#include <fcntl.h>
#include <string.h>

#include <glib.h>
#include <glib-object.h>
#include <glib/gprintf.h>

#include "AudioBusManagerPlugin.h"
#include "PluginData.h"
#include "PluginManager.h"

/**************************************************************************/
/*                               Defines                                  */
/**************************************************************************/
#define MAX_KEY_BUFF 100
#define PLUGIN_RESPOND_TIMEOUT 3
#define PACKAGE_STRING "abmtest-0.3"

#define AUDIO_PARAM_TABLE_KEY_CHANNEL_COUNT "channel_count"
#define AUDIO_PARAM_TABLE_KEY_SOURCE_DELAY  "source_delay"
#define AUDIO_PARAM_TABLE_KEY_CHANNEL_ID_FORMAT  "channel_id-%d"
#define MOST_AUDIO_PARAM_TABLE_KEY_DEVICE_HANDLE "dev_hnd"
#define AUDIO_PARAM_TABLE_KEY_CHANNEL_ID_KEY_LENGTH  20
#define AUDIO_PARAM_TEMP_BUF_SIZE 256

/* the following defines should match the same ones in most.c: */
#define MAX_ALSA_DEVICES 4                 /**< Max. number of ALSA devices supported (a Linux limitation) */
#define SOURCE_OFFSET    0                 /**< Offset index of source devices inside @sPluginParam array  */
#define SINK_OFFSET     (MAX_ALSA_DEVICES) /**< Offset index of sink devices inside @sPluginParam array    */
/**************************************************************************/
/*  Private Functions Prototypes-functions used internally in this file   */
/**************************************************************************/

/**************************************************************************/
/*                               Defines                                  */
/**************************************************************************/

/**************************************************************************/
/*                                Types                                   */
/**************************************************************************/

/**************************************************************************/
/*                              Globals                                   */
/**************************************************************************/

static char *gPluginName = "/usr/lib/audiomanagement/abm/plugins/libabm-most.so";


static gboolean sPrintVersion = FALSE;
static volatile gboolean gExit;
static volatile gboolean gGotVmostReply;


static GOptionEntry entries[] =
{
{ "plugin", 'p', 0, G_OPTION_ARG_STRING, &gPluginName, "Name of the plugin file", "<filename>" },
{ "version", 0, 0, G_OPTION_ARG_NONE, &sPrintVersion, "Print version", NULL},
{ NULL },
};

/**************************************************************************/
/*                       Private Functions Bodies                         */
/**************************************************************************/

/**************************************************************************/
/*                                                                        */
/*  Function Name: sVmostCb                                               */
/*                                                                        */
/**************************************************************************/
/**
 *  
 *  \return None
 **************************************************************************
*/
static void sVmostCb
(
    ABM_AudioPluginParams_t AudioPluginParams, gboolean success, GHashTable *pAudioParamTable,
    ABM_PulseAudioDevIndex PA_Index
)
{
    int i;
    gchar *pData;
    guint ChanCount, AlsaDev;
    gchar Tempbuffer[AUDIO_PARAM_TEMP_BUF_SIZE];

    g_printf ("*** Got allocation/connection callback:\n");

    if (success)
    {
        pData = g_hash_table_lookup (pAudioParamTable, AUDIO_PARAM_TABLE_KEY_CHANNEL_COUNT);
        if (NULL == pData)
        {
            g_printf ("Error: parameter %s not found in hash table\n", AUDIO_PARAM_TABLE_KEY_CHANNEL_COUNT);
            gGotVmostReply = TRUE;
            return;
        }
        g_printf ("\tChanCount=%s\n", pData);
        ChanCount = atoi (pData);

        pData = g_hash_table_lookup (pAudioParamTable, MOST_AUDIO_PARAM_TABLE_KEY_DEVICE_HANDLE);
        if (NULL == pData)
        {
            g_printf ("Error: parameter %s not found in hash table\n", MOST_AUDIO_PARAM_TABLE_KEY_DEVICE_HANDLE);
            gGotVmostReply = TRUE;
            return;
        }
        AlsaDev = atoi (pData) % MAX_ALSA_DEVICES;
        g_printf ("\tALSA device=%u\n", AlsaDev);
        for (i = 0; i < ChanCount; i++)
        {
            sprintf (Tempbuffer, AUDIO_PARAM_TABLE_KEY_CHANNEL_ID_FORMAT, i);
            pData = g_hash_table_lookup (pAudioParamTable, Tempbuffer);
            g_printf ("\tChan[%u]=%s\n", i, pData);
        }
        g_printf ("\tPulseDevIndex=%u\n", PA_Index);
        g_printf ("*** SUCCESS\n");
    }
    else
    {
        g_printf ("*** FAILURE\n");
    }

    gGotVmostReply = TRUE;
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: sVmostDeallocCb                                        */
/*                                                                        */
/**************************************************************************/
/**
 *  
 *  \return None
 **************************************************************************
*/
static void sVmostDeallocCb
(
    ABM_AudioPluginParams_t AudioPluginParams, gboolean success, GHashTable *pAudioParamTable,
    ABM_PulseAudioDevIndex PA_Index
)
{
    int i;
    gchar *pData;
    guint ChanCount, AlsaDev;
    gchar Tempbuffer[AUDIO_PARAM_TEMP_BUF_SIZE];

    g_printf ("*** Got deallocation/disconnection callback:\n");

    if (success)
    {
        pData = g_hash_table_lookup (pAudioParamTable, AUDIO_PARAM_TABLE_KEY_CHANNEL_COUNT);
        if (NULL == pData)
        {
            g_printf ("Error: parameter %s not found in hash table\n", AUDIO_PARAM_TABLE_KEY_CHANNEL_COUNT);
            gGotVmostReply = TRUE;
            return;
        }
        g_printf ("\tChanCount=%s\n", pData);
        ChanCount = atoi (pData);

        AlsaDev = atoi (pData) % MAX_ALSA_DEVICES;
        g_printf ("\tALSA device=%u\n", AlsaDev);
        for (i = 0; i < ChanCount; i++)
        {
            sprintf (Tempbuffer, AUDIO_PARAM_TABLE_KEY_CHANNEL_ID_FORMAT, i);
            pData = g_hash_table_lookup (pAudioParamTable, Tempbuffer);
            g_printf ("\tChan[%u]=%s\n", i, pData);
        }
        g_printf ("\tPulseDevIndex=%u\n", PA_Index);
        g_printf ("*** SUCCESS\n");
    }
    else
    {
        g_printf ("*** FAILURE\n");
    }

    gGotVmostReply = TRUE;
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: setNonblocking                                         */
/*                                                                        */
/**************************************************************************/
/**
 *  \return None
 **************************************************************************
*/
int setNonblocking
(
    int fd
)
{
    int flags;

    /* If they have O_NONBLOCK, use the Posix way to do it */


    #if defined(O_NONBLOCK)

    /* Fixme: O_NONBLOCK is defined but broken on SunOS 4.1.x and AIX 3.2.5. */
    if (-1 == (flags = fcntl (fd, F_GETFL, 0)))
        flags = 0;

    return fcntl (fd, F_SETFL, flags | O_NONBLOCK);
    #else

    /* Otherwise, use the old way of doing it */
    flags = 1;
    return ioctl (fd, FIOBIO, &flags);
    #endif

}
/**************************************************************************/
/*                                                                        */
/*  Function Name: sSignalCatch                                           */
/*                                                                        */
/**************************************************************************/
/**
 *  \return None
 **************************************************************************
*/
static void sSignalCatch
(
    int sig
)
{
    g_printf ("Received signal %d. Initiating shutdown procedure...\n", sig);
    gExit = TRUE;
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: sShutdown                                              */
/*                                                                        */
/**************************************************************************/
/**
 *  Application shutdown procedure
 *
 *  \return None
 **************************************************************************
*/
static void sShutdown
(
    PluginData_t *pPluginData
)
{
    PLUGIN_DATA_FUNC_TABLE->FnTerminate ();
    free ((ABM_PluginFunctions_t *)pPluginData->pPluginFnTable);
    PM_UnloadPlugin (pPluginData);
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: sProcessStdIn                                          */
/*                                                                        */
/**************************************************************************/
/**
 *  Read char from STDIN. If '\n' is detected, process command line 
 *
 *  \return FALSE - no action
 *          TRUE  - end of line detected. pKeyBuff is valid.
 **************************************************************************
*/
static gboolean sProcessStdIn
(
    char * *pKeyBuff
)
{
    static char KeyBuffer[MAX_KEY_BUFF];
    static int KeyLen;

    if (read (STDIN_FILENO, &KeyBuffer[KeyLen], 1) <= 0)
    {
        return FALSE;
    }

    if ('\n' == KeyBuffer[KeyLen])
    {
        KeyBuffer[KeyLen] = 0;
        *pKeyBuff = KeyBuffer;
        KeyLen = 0;
        return TRUE;
    }
    else
    {
        KeyLen++;
    }

    return FALSE;
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: sUsage                                                   */
/*                                                                        */
/**************************************************************************/
/**
 *  Print usage
 *
 *  \return None
 **************************************************************************
*/
static void sUsage
(
    void
)
{
    g_printf ("*** ABM test program for plugins ***\n");
    g_printf ("Supported commands:\n");
    g_printf ("\thelp - Print this menu\n");
    g_printf ("\talloc <Channnel count> - Allocate <Channnel count> MOST sync channels.\n");
    g_printf ("\tdealloc <Alsa ID> - Deallocate previously allocated sync channels on <Alsa ID>\n");
    g_printf ("\tconnect <SrcDelay> <Chan 0> <Chan 1> ... <Chan N> - Connect to N sync channels\n");
    g_printf ("\tdisconnect <Alsa ID> - Disconnect from previously connected sync channels on <Alsa ID>\n");
    g_printf ("Command:\n");
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: sAlloc                                                 */
/*                                                                        */
/**************************************************************************/
/**
 *  Allocate MOST channels
 *
 *  \return None
 **************************************************************************
*/
static void sAlloc
(
    int argc,     /**< number of arguments including application name */
    char * *argv, /**< array of string arguments */
    PluginData_t *pPluginData
)
{
    int MostChanCount;
    GTimer *gtimer;
    GHashTable *pAudioParamTable;
    ABM_AudioPluginParams_t AudioPluginParams;

    /* The  AudioPluginParams is populated with daata relevant to clients only
       This data is not suppoesed to be used by the most plugin.
     */
    strcpy (AudioPluginParams.DeviceID, "MP3Player");
    strcpy (AudioPluginParams.ClientType, "sink");
    strcpy (AudioPluginParams.PluginType, "PluginTest");

    if (argc < 2)
    {
        g_printf ("Error: missing <Channnel count> argument !\n");
        return;
    }

    pAudioParamTable = g_hash_table_new (g_str_hash, g_str_equal);

    MostChanCount = atoi (argv[1]);
    g_printf ("*** Allocating %u MOST sync channels...\n", MostChanCount);
    g_hash_table_insert (pAudioParamTable, AUDIO_PARAM_TABLE_KEY_CHANNEL_COUNT, argv[1]);

    gGotVmostReply = FALSE;
    PLUGIN_DATA_FUNC_TABLE->FnAllocateChannels (AudioPluginParams, pAudioParamTable, sVmostCb);

    g_hash_table_unref (pAudioParamTable);

    gtimer = g_timer_new ();

    while (FALSE == gGotVmostReply)
    {
        usleep (10000);
        if (g_timer_elapsed (gtimer, NULL) > (gdouble)PLUGIN_RESPOND_TIMEOUT)
        {
            g_timer_destroy (gtimer);
            g_printf ("Error: Plugin not responded in %u secs !\n", PLUGIN_RESPOND_TIMEOUT);
            return;
        }
    }
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: sDeAlloc                                               */
/*                                                                        */
/**************************************************************************/
/**
 *  Deallocate previously allocated MOST channels
 *
 *  \return None
 **************************************************************************
*/
static void sDeAlloc
(
    int argc,     /**< number of arguments including application name */
    char * *argv, /**< array of string arguments */
    PluginData_t *pPluginData
)
{
    int AlsaId;
    GTimer *gtimer;
    GHashTable *pAudioParamTable;
    ABM_AudioPluginParams_t AudioPluginParams;
    char Tempbuffer[64];

    strcpy (AudioPluginParams.DeviceID, "MP3Player");
    strcpy (AudioPluginParams.ClientType, "sink");
    strcpy (AudioPluginParams.PluginType, "PluginTest");

    if (argc < 2)
    {
        g_printf ("Error: missing <Alsa ID> argument !\n");
        return;
    }

    AlsaId = atoi (argv[1]);
    g_printf ("*** Deallocating MOST sync channels on Alsa device %u...\n", AlsaId);

    pAudioParamTable = g_hash_table_new (g_str_hash, g_str_equal);

    g_sprintf (Tempbuffer, "%d", AlsaId + SOURCE_OFFSET);
    g_hash_table_insert (pAudioParamTable, MOST_AUDIO_PARAM_TABLE_KEY_DEVICE_HANDLE, Tempbuffer);

    gGotVmostReply = FALSE;
    PLUGIN_DATA_FUNC_TABLE->FnDeAllocateChannels (AudioPluginParams, pAudioParamTable, sVmostDeallocCb);

    g_hash_table_unref (pAudioParamTable);

    gtimer = g_timer_new ();

    while (FALSE == gGotVmostReply)
    {
        usleep (10000);
        if (g_timer_elapsed (gtimer, NULL) > (gdouble)PLUGIN_RESPOND_TIMEOUT)
        {
            g_timer_destroy (gtimer);
            g_printf ("Error: Plugin not responded in %u secs !\n", PLUGIN_RESPOND_TIMEOUT);
            return;
        }
    }
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: sConnect                                               */
/*                                                                        */
/**************************************************************************/
/**
 *  Connect to sync MOST channels
 *
 *  \return None
 **************************************************************************
*/
static void sConnect
(
    int argc,     /**< number of arguments including application name */
    char * *argv, /**< array of string arguments */
    PluginData_t *pPluginData
)
{
    int ArgIndx, TempBufCnt;
    GTimer *gtimer;
    GHashTable *pAudioParamTable;
    gchar Tempbuffer[32][AUDIO_PARAM_TEMP_BUF_SIZE];
    ABM_AudioPluginParams_t AudioPluginParams;
    int ChannelCount = 0;

    /* The  AudioPluginParams is populated with daata relevant to clients only
       This data is not suppoesed to be used by the most plugin.
     */
    strcpy (AudioPluginParams.DeviceID, "MP3Player");
    strcpy (AudioPluginParams.ClientType, "source");
    strcpy (AudioPluginParams.PluginType, "PluginTest");

    TempBufCnt = 0;

    if (argc < 3)
    {
        g_printf ("Error: At least <SrcDelay> <Chan 0> have to be defined !\n");
        return;
    }

    pAudioParamTable = g_hash_table_new (g_str_hash, g_str_equal);

    g_hash_table_insert (pAudioParamTable, AUDIO_PARAM_TABLE_KEY_SOURCE_DELAY, argv[1]);
    ArgIndx = 2;

    ChannelCount = 0;

    while (ArgIndx < argc)
    {
        sprintf (Tempbuffer[TempBufCnt], AUDIO_PARAM_TABLE_KEY_CHANNEL_ID_FORMAT, ChannelCount);
        g_hash_table_insert (pAudioParamTable, Tempbuffer[TempBufCnt++], argv[ArgIndx]);
        ArgIndx++;
        ChannelCount++;
    }

    sprintf (Tempbuffer[TempBufCnt], "%d", ChannelCount);
    g_hash_table_insert (pAudioParamTable, AUDIO_PARAM_TABLE_KEY_CHANNEL_COUNT, Tempbuffer[TempBufCnt++]);

    g_printf ("*** Connecting to %u MOST sync channels...\n", ChannelCount);

    gGotVmostReply = FALSE;

    PLUGIN_DATA_FUNC_TABLE->FnConnectChannels (AudioPluginParams, pAudioParamTable, sVmostCb);

    g_hash_table_unref (pAudioParamTable);

    gtimer = g_timer_new ();

    while (FALSE == gGotVmostReply)
    {
        usleep (10000);
        if (g_timer_elapsed (gtimer, NULL) > (gdouble)PLUGIN_RESPOND_TIMEOUT)
        {
            g_timer_destroy (gtimer);
            g_printf ("Error: Plugin not responded in %u secs !\n", PLUGIN_RESPOND_TIMEOUT);
            return;
        }
    }
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: sDisconnect                                            */
/*                                                                        */
/**************************************************************************/
/**
 *  Disconnect from previously connceted MOST channels
 *
 *  \return None
 **************************************************************************
*/
static void sDisconnect
(
    int argc,     /**< number of arguments including application name */
    char * *argv, /**< array of string arguments */
    PluginData_t *pPluginData
)
{
    int AlsaId;
    GTimer *gtimer;
    GHashTable *pAudioParamTable;
    char Tempbuffer[64];

    ABM_AudioPluginParams_t AudioPluginParams;

    strcpy (AudioPluginParams.DeviceID, "MP3Player");
    strcpy (AudioPluginParams.ClientType, "source");
    strcpy (AudioPluginParams.PluginType, "PluginTest");

    if (argc < 2)
    {
        g_printf ("Error: missing <Alsa ID> argument !\n");
        return;
    }

    AlsaId = atoi (argv[1]);
    g_printf ("*** Disconnecting MOST sync channels on Alsa device %u...\n", AlsaId);

    pAudioParamTable = g_hash_table_new (g_str_hash, g_str_equal);

    g_sprintf (Tempbuffer, "%d", AlsaId + SINK_OFFSET);
    g_hash_table_insert (pAudioParamTable, MOST_AUDIO_PARAM_TABLE_KEY_DEVICE_HANDLE, Tempbuffer);

    gGotVmostReply = FALSE;
    PLUGIN_DATA_FUNC_TABLE->FnDisconnectChannels (AudioPluginParams, pAudioParamTable, sVmostDeallocCb);

    g_hash_table_unref (pAudioParamTable);

    gtimer = g_timer_new ();

    while (FALSE == gGotVmostReply)
    {
        usleep (10000);
        if (g_timer_elapsed (gtimer, NULL) > (gdouble)PLUGIN_RESPOND_TIMEOUT)
        {
            g_timer_destroy (gtimer);
            g_printf ("Error: Plugin not responded in %u secs !\n", PLUGIN_RESPOND_TIMEOUT);
            return;
        }
    }
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: main                                                   */
/*                                                                        */
/**************************************************************************/
/**
 *  Main entry
 *
 *  \return None
 **************************************************************************
*/
int main
(
    int argc,    /**< number of arguments including application name */
    char * *argv /**< array of string arguments */
)
{
    GOptionContext *context;
    GError *pGError = NULL;
    PluginData_t PluginData;
    char *pCmdLine;
    gint pCmdLineArgc;
    gchar * *pCmdLineArgv;

    g_type_init ();

    if (!g_thread_supported ())
    {
        g_thread_init (NULL);
    }

    {
    GLogLevelFlags fatal_mask;

    fatal_mask = g_log_set_always_fatal (G_LOG_FATAL_MASK);
    fatal_mask |= G_LOG_LEVEL_WARNING | G_LOG_LEVEL_CRITICAL;
    g_log_set_always_fatal (fatal_mask);
    }

    context = g_option_context_new ("");
    g_option_context_add_main_entries (context, entries, NULL);
    g_option_context_set_summary (context, "Audio manager test application");
    g_option_context_set_description (context, "This is a description");

    if (!g_option_context_parse (context, &argc, &argv, &pGError))
    {
        g_printf ("%s\n", pGError->message);
        exit (1);
    }

    if (sPrintVersion)
    {
        g_printf ("Package: '%s'\n", PACKAGE_STRING);
        exit (1);
    }
    /* catch Ctrl-C and Hangup signals for clean shutdown */
    (void)signal (SIGINT, sSignalCatch);
    (void)signal (SIGHUP, sSignalCatch);
    (void)signal (SIGTERM, sSignalCatch);

    /* Load and init plugin */
    memset (&PluginData, 0, sizeof (PluginData));

    g_printf ("Loading interface plugin %s\n", gPluginName);

    PluginData.pFileName = gPluginName;

    if (FALSE == PM_LoadPlugin (&PluginData, ABM_PLUGIN_INIT_FN_NAME))
    {
        g_printf ("Error loading plugin %s\n", gPluginName);
        exit (1);
    }

    PluginData.pPluginFnTable = malloc (sizeof (ABM_PluginFunctions_t));

    if (!PluginData.pPluginFnTable)
    {
        g_printf ("Error: ABM_PluginFunctions_t Allocation failed\n");
        exit (1);
    }

    if (NULL == PluginData.pInitFn)
    {
        g_printf ("Error: Plugin's init function is NULL\n");
        exit (1);
    }

    if (FALSE == ((ABM_PluginInit_t)PluginData.pInitFn) (&PluginData))
    {
        g_printf (" Failed to initialise plug-in: %s\n", PluginData.pFileName);
        exit (1);
    }

    setNonblocking (STDIN_FILENO);

    sUsage ();
    g_printf ("Press <Crtl-C> to exit.\n");
    gExit = FALSE;

    while (FALSE == gExit)
    {
        if (TRUE == sProcessStdIn (&pCmdLine))
        {
            /* '/n' char detected. Go parsing command: */
            if (FALSE == g_shell_parse_argv (pCmdLine, &pCmdLineArgc, &pCmdLineArgv, &pGError))
            {
                g_printf ("Error parsing '%s'. Err code: %d, %s\n", pCmdLine, (*pGError).code, (*pGError).message);
                g_clear_error (&pGError);
                continue;
            }
            if (0 == strcasecmp ("alloc", pCmdLineArgv[0]))
            {
                sAlloc (pCmdLineArgc, pCmdLineArgv, &PluginData);
            }
            else if (0 == strcasecmp ("dealloc", pCmdLineArgv[0]))
            {
                sDeAlloc (pCmdLineArgc, pCmdLineArgv, &PluginData);
            }
            else if (0 == strcasecmp ("connect", pCmdLineArgv[0]))
            {
                sConnect (pCmdLineArgc, pCmdLineArgv, &PluginData);
            }
            else if (0 == strcasecmp ("disconnect", pCmdLineArgv[0]))
            {
                sDisconnect (pCmdLineArgc, pCmdLineArgv, &PluginData);
            }
            else if (0 == strcasecmp ("help", pCmdLineArgv[0]))
            {
                sUsage ();
            }
            else
            {
                g_printf ("Error: Unknown command!\n");
                sUsage ();
            }
        }
        usleep (10000);
    }

    sShutdown (&PluginData);

    exit (0);
}
/// @endcond
