/*
 * AudioConnectionManager.c 
 *
 * Copyright (c) 2008 Wind River Systems, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Lesser GNU General Public License version 2.1 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the Lesser GNU General Public License for more details.
 *
 * You should have received a copy of the Lesser GNU General Public License version 2.1
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */
/// @cond internal_documentation
/**
 * \file 
 * This file contains the implementation of the DBus objects and message handler functions
 * for the com.windriver.audio.ConnectionManager DBus interface    
 *
 * \version $Revision: 1.18 $
 * \date    $Date: 2008/07/12 01:40:53 $
 */

/**************************************************************************/
/*                              Includes                                  */
/**************************************************************************/
#include "acmd.h"
#include "AudioBusManagerPlugin.h"
#include "AudioBusManager.h"
#include "AudioConnectionManager.h"
#include "acm-glue.h"
#include "parouting.h"
#include <alloca.h>

/**************************************************************************/
/*                               Defines                                  */
/**************************************************************************/

G_DEFINE_TYPE
(
    AudioConnectionManager, acm, G_TYPE_OBJECT
);

/**************************************************************************/
/*                              Externs                                   */
/**************************************************************************/
extern DBusGConnection *dbus_connection;

/* Find the length of STRING, but scan at most MAXLEN characters.
   If no '\0' terminator is found in that many characters, return MAXLEN.  */
extern size_t strnlen (__const char *__string, size_t __maxlen)
__THROW __attribute_pure__ __nonnull ((1));

/**************************************************************************/
/*                              Globals                                   */
/**************************************************************************/

/**************************************************************************/
/*                          Local Variables                               */
/**************************************************************************/

/**
* The PA Routing Context used to connect the Audio Connection Manager to Pulse Audio.
*/
static PA_Routing_Ctx_t *pPACtx = NULL;

/**
 * Dbus connection to the system bus used to contact the device control modules.
 */
static DBusGConnection *pDbusConnection = NULL;

/**************************************************************************/
/*  Private Functions Prototypes-functions used internally in this file   */
/**************************************************************************/

/**************************************************************************/
/*                       Private Functions Bodies                         */
/**************************************************************************/

/**************************************************************************/
/*                                                                        */
/*  Function Name: acm_convertStringToClientType                          */
/*                                                                        */
/**************************************************************************/
/**
 *  This function translates an ABM plugin type string ("sink", "source") 
 * to the corresponding enumeration value.  If an invalid string is passed in,
 * the returned enum value will be ABM_AUDIO_INVALID_TYPE to indicate an error.               
 *
 *  \return the enumeration value for the ABM plugin type.                 
 *                                                     
 **************************************************************************
 */
gboolean acm_plugin_param_init
(
    gchar *deviceID,                             /**< [In] Device identifier that is same as the device ID registerd with ACM*/
    gchar *busName,                              /**< [in] The name of the bus to allocate a channel on.  */
    gchar *clientType,                           /**< [In] client type to indicate if the allocation is for source or for a sink */
    DBusGMethodInvocation *context,              /**< [in] the dbus method invocation context. */
    ABM_AudioPluginParams_t *pAudioPluginParams, /**< [Out] Parameter structure to be filled in*/
    GError * *pError                             /**< [Out] erro value to be returned to dbus*/
)
{
    gboolean success = TRUE;
    gint deviceIdLength = sizeof (deviceID);
    gint busNameLength = sizeof (busName);
    gint clientTypeLength = sizeof (clientType);

    if ((!deviceIdLength) || (deviceIdLength >= ABM_MAX_DEVICE_ID_LENGTH))
    {
        success = FALSE;
        *pError = getGError (ACM_ERROR_INVALID_DEV_ID);
    }

    if ((!busNameLength) || (busNameLength >= ABM_MAX_PLUGIN_TYPE_NAME_LENGTH))
    {
        success = FALSE;
        *pError = getGError (ACM_ERROR_UNKNOWN_BUS);
    }

    if ((!clientTypeLength) || (clientTypeLength >= ABM_MAX_CLIENT_TYPE_NAME_LENGTH))
    {
        success = FALSE;
        *pError = getGError (ACM_ERROR_UNKNOWN_CLIENT_TYPE);
    }

    if (success)
    {
        strcpy (pAudioPluginParams->PluginType, busName);
        strcpy (pAudioPluginParams->DeviceID, deviceID);
        strcpy (pAudioPluginParams->ClientType, clientType);
        pAudioPluginParams->pDbusContext = context;
    }

    return success;
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: acm_establishedPAConnection                            */
/*                                                                        */
/**************************************************************************/
/**
 * This function ensures our connection to PA is up and attempts to bring
 * up the connection if it isn't.
 *
 *  \return TRUE to indicate an established connection to PulseAudio
 *                                                     
 **************************************************************************
 */
/**
 */
static gboolean acm_establishedPAConnection
(
    void
)
{
    gboolean success = TRUE;
    gboolean need_connect = FALSE;

    if (NULL != pPACtx)
    {
        if (PA_ROUTING_OK != PARoutingModuleCheckPAConnection (pPACtx))
        {
            PARoutingPADisconnect (pPACtx);
            need_connect = TRUE;
        }
    }
    else
    {
        pPACtx = PARoutingModuleNew (ACM_DEVICE_CONTROL_ENVIRONMENT_KEYWORD);
        if (NULL == pPACtx)
        {
            g_message ("AudioConnectionManager unable to connect to Pulse Audio!");
            success = FALSE;
        }
        else
        {
            need_connect = TRUE;
        }
    }

    if ((NULL != pPACtx) && (TRUE == need_connect))
    {
        if (PA_ROUTING_OK != PARoutingPAConnect (pPACtx))
        {
            success = FALSE;
            g_message ("AudioConnectionManager unable to connect to Pulse Audio!");
        }
        else
        {
            /* We just reconnected so we should apply the routes we have stored */
            PARoutingModuleApplyRoutes (pPACtx);
        }
    }

    return success;
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: acm_allocate_callback                                  */
/*                                                                        */
/**************************************************************************/
/**
 *  Callback function used to complete the Allocate() DBus method call.  This 
 * function should be called by an ABM plugin after it has performed the channel
 * allocation on the specified bus.
 * 
 * \see acm_allocate
 *                                                     
 **************************************************************************
 */
static void acm_allocate_callback
(
    ABM_AudioPluginParams_t AudioPluginParams, /**< [in] Audio parametrs used to 
                                               initiate the allocation*/
    gboolean success, /**< [in] TRUE if the allocation was a
                                success, FALSE otherwise.*/
    GHashTable *returnValue, /**< [in] A set of key/value pairs that
                                       should be returned to the 
                                       DBus client that called the
                                       Allocate() method.*/
    ABM_PulseAudioDevIndex PA_Index /**< [In] Pulse Audio parameters from allocation */
)
{
    GError *error = NULL;

    if (!success)
    {
        error = getGError (ACM_ERROR_UNABLE_TO_ALLOCATE);
    }

    g_debug (
        "acm_allocate_callback called : DeviceID : %s, ClientType: %s,"
            "PluginType: %s, PA_ALSA index: %d,\n", AudioPluginParams.DeviceID,
        AudioPluginParams.ClientType,
        AudioPluginParams.PluginType,
        PA_Index);

    ACM_DBUS_RETURN (success, AudioPluginParams.pDbusContext, returnValue, error);
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: acm_deallocate_callback                                */
/*                                                                        */
/**************************************************************************/
/**
 *  Callback function used to complete the DeAllocate() DBus method call.  This 
 * function should be called by an ABM plugin after it has performed the channel
 * deallocation on the specified bus.
 * 
 * \see acm_deallocate
 *                                                     
 **************************************************************************
 */
static void acm_deallocate_callback
(
    ABM_AudioPluginParams_t AudioPluginParams, /**< [in] Audio parametrs used to 
                                               initiate the deallocation*/
    gboolean success, /**< [in] TRUE if the deallocation was a
                                success, FALSE otherwise.*/
    GHashTable *returnValue, /**< [in] A set of key/value pairs that
                                       should be returned to the 
                                       DBus client that called the
                                       DeAllocate() method.*/
    ABM_PulseAudioDevIndex PA_Index /**< [In] Pulse Audio parameters from deallocation */
)
{
    GError *error = NULL;

    if (!success)
    {
        error = getGError (ACM_ERROR_UNABLE_TO_DEALLOCATE);
    }

    g_debug (
        "acm_deallocate_callback called : DeviceID : %s, ClientType: %s,"
            "PluginType: %s, PA_ALSA index: %d,\n", AudioPluginParams.DeviceID,
        AudioPluginParams.ClientType,
        AudioPluginParams.PluginType,
        PA_Index);

    ACM_DBUS_RETURN (success, AudioPluginParams.pDbusContext, returnValue, error);
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: acm_connect_callback                                  */
/*                                                                        */
/**************************************************************************/
/**
 *  Callback function used to complete the Allocate() DBus method call.  This 
 * function should be called by an ABM plugin after it has performed the channel
 * allocation on the specified bus.
 * 
 * \see acm_allocate
 *                                                     
 **************************************************************************
 */
static void acm_connect_callback
(
    ABM_AudioPluginParams_t AudioPluginParams, /**< [in] Audio parametrs used to 
                                               initiate the connection*/
    gboolean success, /**< [in] TRUE if the connect operation
                               was a success, FALSE otherwise.*/
    GHashTable *returnValue, /**< [in] A set of key/value pairs that
                                       should be returned to the 
                                       DBus client that called the
                                       Allocate() method.*/
    ABM_PulseAudioDevIndex PA_Index /**< [In] Pulse Audio parameters from connect 
                                          operation */
)
{
    GError *error = NULL;

    if (!success)
    {
        error = getGError (ACM_ERROR_UNABLE_TO_ALLOCATE);
    }

    g_debug (
        "acm_connect_callback called : DeviceID : %s, ClientType: %s,"
            "PluginType: %s, PA_ALSA index: %d,\n", AudioPluginParams.DeviceID,
        AudioPluginParams.ClientType,
        AudioPluginParams.PluginType,
        PA_Index);

    /* source -> output_routing */
    /* sink -> input_routing */
    if (ABM_AUDIO_IS_LOCAL (AudioPluginParams.ClientType))
    {
        acm_establishedPAConnection ();
        PARoutingModuleSetAndApplyRoute (
            pPACtx,
            (ABM_AUDIO_IS_SINK (AudioPluginParams.ClientType) ? PA_ROUTE_INPUT : PA_ROUTE_OUTPUT),
            AudioPluginParams.DeviceID,
            PA_Index);
    }

    ACM_DBUS_RETURN (success, AudioPluginParams.pDbusContext, returnValue, error);
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: acm_disconnect_callback                                */
/*                                                                        */
/**************************************************************************/
/**
 * Callback function used to complete the Disconnect() DBus method call.  This 
 * function should be called by an ABM plugin after it has performed the channel
 * disconnect on the specified bus.
 * 
 * \see acm_disconnect
 *                                                     
 **************************************************************************
 */
static void acm_disconnect_callback
(
    ABM_AudioPluginParams_t AudioPluginParams, /**< [in] Audio parameters used to 
                                                    disconnect*/
    gboolean success, /**< [in] TRUE if the disconnect operation
                                was a success, FALSE otherwise.*/
    GHashTable *returnValue, /**< [in] A set of key/value pairs that
                                       should be returned to the 
                                       DBus client that called the
                                       Disconnect() method.*/
    ABM_PulseAudioDevIndex PA_Index /**< [In] Pulse Audio parameters from disconnect 
                                              operation */
)
{
    GError *error = NULL;

    if (!success)
    {
        error = getGError (ACM_ERROR_UNABLE_TO_ALLOCATE);
    }

    g_debug (
        "acm_disconnect_callback called : DeviceID : %s, ClientType: %s,"
            "PluginType: %s, PA index: %d,\n", AudioPluginParams.DeviceID,
        AudioPluginParams.ClientType,
        AudioPluginParams.PluginType,
        PA_Index);

    ACM_DBUS_RETURN (success, AudioPluginParams.pDbusContext, returnValue, error);
}
/**************************************************************************/
/*                       Public Functions Bodies                         */
/**************************************************************************/
/***********************************************************************
 * Object Control functions for the managed classes.
 *
 * These function must be in the same file as where the GType is defined,
 * so these functions are just wrappers for the "real" init functions
 * contained in external files.
 ***********************************************************************/

/**************************************************************************/
/*                                                                        */
/*  Function Name: acm_class_init                                */
/*                                                                        */
/**************************************************************************/
/**
 *  Class initialization function.               
 *                 
 *                                                     
 **************************************************************************
 */
/* This is used by the g_object functionality */
void acm_class_init
(
    AudioConnectionManagerClass *klass /**< [in] The class object to init.*/
)
{
    /* Define all signals for the interface. */
    klass->ACMSignalList[ACM_HELLO_CALLED] = g_signal_new (
                                                 "hello_called",                        G_OBJECT_CLASS_TYPE (klass),
                                                 (G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION), 0,
                                                 NULL,                                  NULL,
                                                 g_cclosure_marshal_VOID__STRING,       G_TYPE_NONE,
                                                 1,                                     G_TYPE_STRING);
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: acm_init                                               */
/*                                                                        */
/**************************************************************************/
/**
 *  Instance initialization function.                                
 *                                                     
 **************************************************************************
 */
void acm_init
(
    AudioConnectionManager *server /**< [in] The object instance to init.*/
)
{
    if (!server->initialized)
    {
        server->connection = dbus_connection;

        dbus_g_object_type_install_info (acm_get_type (), &dbus_glib_acm_object_info);

        dbus_g_connection_register_g_object (server->connection, ACM_BASE_OBJECT_PATH, G_OBJECT (server));

        /* Attempt to connect to PulseAudio.  Note that the server can still
        start if a connection can't be established; the server will attempt to
        reconnect as needed */
        acm_establishedPAConnection ();

        server->initialized = TRUE;
    }
    else
    {
        g_critical ("AudioConnectionManager already created!");
    }
}


void acm_finalize
(
    AudioConnectionManager *server /**< [in] the object instance to destroy.*/
)
{
    dbus_connection_unregister_object_path (
        dbus_g_connection_get_connection (server->connection), ACM_BASE_OBJECT_PATH);

    if (NULL != pPACtx)
    {
        if (PA_ROUTING_OK == PARoutingModuleCheckPAConnection (pPACtx))
        {
            PARoutingPADisconnect (pPACtx);
        }
        PARoutingModuleFree (pPACtx);
        pPACtx = NULL;
    }

    g_type_free_instance ((GTypeInstance *)server);
}
/**************************************************************************/
/*           ConnectionManager DBus message handler functions             */
/**************************************************************************/

gboolean acm_allocate
(
    AudioConnectionManager *server, /**< [in] the object instance that the method was called on. */
    gchar *deviceID,                /**< [In] Device identifier that is same as the device ID registerd with ACM*/
    gchar *busName,                 /**< [in] The name of the bus to allocate a channel on.  */
    gchar *clientType,              /**< [In] client type to indicate if the allocation is for source or for a sink */
    GHashTable *busData,            /**< [in] Bus-specific data to be processed by the ABM plugin.  */
    DBusGMethodInvocation *context  /**< [in] the method invocation context. */
)
{
    gboolean success = TRUE;
    GError *error = NULL;
    ABM_Results_t result;
    ABM_AudioPluginParams_t AudioPluginParams;

    success = acm_plugin_param_init (deviceID, busName, clientType, context, &AudioPluginParams, &error);

    if (success)
    {
        result = ABM_Allocate (AudioPluginParams, busData, acm_allocate_callback);
    }
    else
    {
        dbus_g_method_return_error (context, error);
    }

    return success;
}


gboolean acm_de_allocate
(
    AudioConnectionManager *server, /**< [in] the object instance that the method was called on. */
    gchar *deviceID,                /**< [In] Device identifier that is same as the device ID registerd with ACM*/
    gchar *busName,                 /**< [in] The name of the bus to allocate a channel on.  */
    gchar *clientType,              /**< [In] client type to indicate if the allocation is for source or for a sink */
    GHashTable *busData,            /**< [in] Bus-specific data to be processed by the ABM plugin.  */
    DBusGMethodInvocation *context  /**< [in] the method invocation context. */
)
{
    gboolean success = TRUE;
    GError *error = NULL;
    ABM_AudioPluginParams_t AudioPluginParams;
    ABM_Results_t result;

    success = acm_plugin_param_init (deviceID, busName, clientType, context, &AudioPluginParams, &error);

    if (success)
    {
        result = ABM_DeAllocate (AudioPluginParams, busData, acm_deallocate_callback);
    }
    else
    {
        dbus_g_method_return_error (context, error);
    }

    return success;
}


gboolean acm_connect
(
    AudioConnectionManager *server, gchar *deviceID, gchar *busName, gchar *clientType, GHashTable *busData,
    DBusGMethodInvocation *context
)
{
    gboolean success = TRUE;
    GError *error = NULL;
    ABM_Results_t result;
    ABM_AudioPluginParams_t AudioPluginParams;

    success = acm_plugin_param_init (deviceID, busName, clientType, context, &AudioPluginParams, &error);

    if (success)
    {
        if (ABM_AUDIO_IS_LOCAL (AudioPluginParams.ClientType))
        {
            result = ABM_Connect (AudioPluginParams, busData, acm_connect_callback);
        }
        else
        {
            success = FALSE;
            error = getGError (ACM_ERROR_NONLOCAL_DEVICE);
            dbus_g_method_return_error (context, error);
        }
    }
    else
    {
        dbus_g_method_return_error (context, error);
    }

    return success;
}


gboolean acm_disconnect
(
    AudioConnectionManager *server, /**< [in] the object instance that the method was called on. */
    gchar *deviceID,                /**< [In] Device identifier that is same as the device ID registerd with ACM*/
    gchar *busName,                 /**< [in] The name of the bus to disconnect a channel on.  */
    gchar *clientType,              /**< [In] client type to indicate if the dis connection is for source or for a sink */
    GHashTable *busData,            /**< [in] Bus-specific data to be processed by the ABM plugin.  */
    DBusGMethodInvocation *context  /**< [in] the method invocation context. */
)
{
    gboolean success = TRUE;
    GError *error = NULL;
    ABM_AudioPluginParams_t AudioPluginParams;
    ABM_Results_t result;

    success = acm_plugin_param_init (deviceID, busName, clientType, context, &AudioPluginParams, &error);

    if (success)
    {
        if (ABM_AUDIO_IS_LOCAL (AudioPluginParams.ClientType))
        {
            acm_establishedPAConnection ();
            PARoutingModuleRemoveRoute (
                pPACtx, (ABM_AUDIO_IS_SINK (AudioPluginParams.ClientType) ? PA_ROUTE_INPUT : PA_ROUTE_OUTPUT),
                AudioPluginParams.DeviceID);
            result = ABM_Disconnect (AudioPluginParams, busData, acm_disconnect_callback);
        }
        else
        {
            success = FALSE;
            error = getGError (ACM_ERROR_NONLOCAL_DEVICE);
            dbus_g_method_return_error (context, error);
        }
    }
    else
    {
        dbus_g_method_return_error (context, error);
    }

    return success;
}
/**************************************************************************/
/*           ConnectionControl DBus message handler functions             */
/**************************************************************************/
static DBusGProxy *acm_ConnectToDevice
(
    gchar *deviceID /**< [in] The ID of the audio device to connect to. */
)
{
    GError *error;
    error = NULL;
    DBusGProxy *deviceProxy = NULL;
    gchar *pathName = NULL;
    int i, deviceIDLen;

    if (NULL == pDbusConnection)
    {
        pDbusConnection = dbus_g_bus_get (DBUS_BUS_SYSTEM, &error);
        if (pDbusConnection == NULL)
        {
            g_printerr ("Failed to open connection to bus: %s\n", error->message);
            g_error_free (error);
            return NULL;
        }
    }

    deviceIDLen = strnlen (deviceID, 256);
    pathName = alloca (deviceIDLen) + 2; /* +2 for the leading / and the trailing \0 */
    /* make a copy of the deviceID into pathName, substituting "/" for "." as we go */
    i = 0;
    pathName[0] = '/';
    pathName[deviceIDLen + 1] = '\0';

    while (i < deviceIDLen)
    {
        if ('\0' == deviceID[i])
        {
            pathName[i + 1] = '\0';
            break;
        }
        else if ('.' == deviceID[i])
        {
            pathName[i + 1] = '/';
        }
        else
        {
            pathName[i + 1] = deviceID[i];
        }
        i++;
    }
    /* Create a proxy object for the specified device */
    deviceProxy = dbus_g_proxy_new_for_name_owner (pDbusConnection,                          deviceID, pathName,
                                                   ACM_DEVICE_CONTROL_MODULE_DBUS_INTERFACE, &error);

    if (NULL == deviceProxy)
    {
        g_printerr ("Failed to create dbus proxy to %s(%s): %s\n", deviceID, pathName, error->message);
        g_error_free (error);
        return NULL;
    }

    return deviceProxy;
}


static GError *acm_ConnectionControlOperation
(
    gchar *operation, /**< [in] The operation to perform. */
    gchar *deviceID   /**< [in] The ID of the audio device to connect to. */
)
{
    gboolean success = TRUE;
    GError *error = NULL;
    DBusGProxy *deviceProxy;

    deviceProxy = acm_ConnectToDevice (deviceID);

    if (NULL == deviceProxy)
    {
        success = FALSE;
    }

    if (success)
    {
        success = dbus_g_proxy_call (deviceProxy, operation, &error, G_TYPE_INVALID, G_TYPE_INVALID);
    }

    if (!success)
    {
        if (NULL != error)
        {
            g_printerr ("Failed to %s device %s: %s\n", operation, deviceID, error->message);
            g_error_free (error);
        }
        error = getGError (ACM_ERROR_UNABLE_TO_SEND_CONTROL_MESSAGE_TO_DEVICE);
    }

    if (deviceProxy)
    {
        g_object_unref (deviceProxy);
    }

    return error;
}


gboolean acm_mute
(
    AudioConnectionManager *server, /**< [in] the object instance that the method was called on. */
    gchar *deviceID,                /**< [in] The ID of the audio device to mute. */
    DBusGMethodInvocation *context  /**< [in] the method invocation context. */
)
{
    gboolean success = TRUE;
    GError *error = NULL;

    error = acm_ConnectionControlOperation ("Mute", deviceID);
    success = (error ? FALSE : TRUE);
    ACM_DBUS_VOID_RETURN (success, context, error);
    return success;
}


gboolean acm_un_mute
(
    AudioConnectionManager *server, /**< [in] the object instance that the method was called on. */
    gchar *deviceID,                /**< [in] The ID of the audio device to unmute. */
    DBusGMethodInvocation *context  /**< [in] the method invocation context. */
)
{
    gboolean success = TRUE;
    GError *error = NULL;

    error = acm_ConnectionControlOperation ("UnMute", deviceID);
    success = (error ? FALSE : TRUE);
    ACM_DBUS_VOID_RETURN (success, context, error);
    return success;
}


gboolean acm_start
(
    AudioConnectionManager *server, /**< [in] the object instance that the method was called on. */
    gchar *deviceID,                /**< [in] The ID of the audio device to start. */
    DBusGMethodInvocation *context  /**< [in] the method invocation context. */
)
{
    gboolean success = TRUE;
    GError *error = NULL;

    error = acm_ConnectionControlOperation ("Start", deviceID);
    success = (error ? FALSE : TRUE);
    ACM_DBUS_VOID_RETURN (success, context, error);
    return success;
}


gboolean acm_stop
(
    AudioConnectionManager *server, /**< [in] the object instance that the method was called on. */
    gchar *deviceID,                /**< [in] The ID of the audio device to stop. */
    DBusGMethodInvocation *context  /**< [in] the method invocation context. */
)
{
    gboolean success = TRUE;
    GError *error = NULL;

    error = acm_ConnectionControlOperation ("Stop", deviceID);
    success = (error ? FALSE : TRUE);
    ACM_DBUS_VOID_RETURN (success, context, error);
    return success;
} gboolean acm_pause
(
    AudioConnectionManager *server, /**< [in] the object instance that the method was called on. */
    gchar *deviceID,                /**< [in] The ID of the audio device to pause. */
    DBusGMethodInvocation *context  /**< [in] the method invocation context. */
)
{
    gboolean success = TRUE;
    GError *error = NULL;

    error = acm_ConnectionControlOperation ("Pause", deviceID);
    success = (error ? FALSE : TRUE);
    ACM_DBUS_VOID_RETURN (success, context, error);
    return success;
}

/// @endcond
