/*
 * AudioConnectionManager.h 
 *
 * Copyright (c) 2008 Wind River Systems, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Lesser GNU General Public License version 2.1 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the Lesser GNU General Public License for more details.
 *
 * You should have received a copy of the Lesser GNU General Public License version 2.1
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */
/// @cond internal_documentation
/**
 * \file 
 * This file contains the declaration of the DBus objects and message handler functions
 * for the com.windriver.audio.ConnectionManager DBus interface    
 *
 * \version $Revision: 1.10 $
 * \date    $Date: 2008/07/12 01:27:17 $
 */

#ifndef AUDIOCONNECTIONMANAGER_H_
#define AUDIOCONNECTIONMANAGER_H_

/**************************************************************************/
/*                              Includes                                  */
/**************************************************************************/
#include <glib.h>
#include <dbus/dbus-glib-bindings.h>
#include <dbus/dbus-glib-lowlevel.h>
#include <dbus/dbus.h>
#include "custom_marshall.h"

/**************************************************************************/
/*                               Defines                                  */
/**************************************************************************/

/**************************************************************************/
/*                              Typedefs                                  */
/**************************************************************************/
/** List of signals for the SMS interface. */

enum ACMSignals
{
    ACM_HELLO_CALLED,
    ACM_LAST_SIGNAL
} ;

/** Structure representing the AudioConnectionManager DBus object. */
typedef struct
{
    GObject parent;              /**< The object's parent object. */
    gchar *objectPath;           /**< The DBus object path for this object.*/
    DBusGConnection *connection; /**< A pointer to the shared DBus connection. */
    gboolean initialized;        /**< TRUE if this object has been initialized successfully.*/
} AudioConnectionManager;

/** Structure representing the class for the PhoneDevice DBus object. */
typedef struct
{
    GObjectClass parent; /**< The class's parent object. */
    guint ACMSignalList[ACM_LAST_SIGNAL]; /**< An array of the registered signal IDs 
                                               for the ACM interface. */
} AudioConnectionManagerClass;

/**************************************************************************/
/*                      Functions Declarations                            */
/**************************************************************************/

/**************************************************************************/
/*                                                                        */
/*  Function Name: acm_finalize                                  */
/*                                                                        */
/**************************************************************************/
/**
 *  Instance destructor function.               
 *                 
 **************************************************************************
 */
void acm_finalize
(
    AudioConnectionManager *server /**< [in] the object instance to destroy.*/
);

/**************************************************************************/
/*           ConnectionManager DBus message handler functions             */
/**************************************************************************/

/**************************************************************************/
/*                                                                        */
/*  Function Name: acm_allocate                                           */
/*                                                                        */
/**************************************************************************/
/**
 *  Allocates an audio channel on the specified bus.              
 *
 *  \return TRUE if successful.                 
 *                                                     
 **************************************************************************
 */
gboolean acm_allocate
(
    AudioConnectionManager *server, /**< [in] the object instance that the method was called on. */
    gchar *deviceID,                /**< [In] Device identifier that is same as the device ID registerd with ACM*/
    gchar *busName,                 /**< [in] The name of the bus to allocate a channel on.  */
    gchar *clientType,              /**< [In] client type to indicate if the allocation is for source or for a sink */
    GHashTable *busData,            /**< [in] Bus-specific data to be processed by the ABM plugin.  */
    DBusGMethodInvocation *context  /**< [in] the method invocation context. */
);

/**************************************************************************/
/*                                                                        */
/*  Function Name: acm_de_allocate                                        */
/*                                                                        */
/**************************************************************************/
/**
 *  Description goes here.               
 *
 *  \return TRUE if successful.                 
 *                                                     
 **************************************************************************
 */
gboolean acm_de_allocate
(
    AudioConnectionManager *server, /**< [in] the object instance that the method was called on. */
    gchar *deviceID,                /**< [In] Device identifier that is same as the device ID registerd with ACM*/
    gchar *busName,                 /**< [in] The name of the bus to allocate a channel on.  */
    gchar *clientType,              /**< [In] client type to indicate if the allocation is for source or for a sink */
    GHashTable *busData,            /**< [in] Bus-specific data to be processed by the ABM plugin.  */
    DBusGMethodInvocation *context  /**< [in] the method invocation context. */
);

/**************************************************************************/
/*                                                                        */
/*  Function Name: acm_connect                                            */
/*                                                                        */
/**************************************************************************/
/**
 *  Allocates an audio channel on the specified bus and then connects the 
 *  specified audio device to that channel.               
 *
 *  \return TRUE if successful.                 
 *                                                     
 **************************************************************************
 */

gboolean acm_connect
(
    AudioConnectionManager *server, /**< [in] the object instance that the method was called on. */
    gchar *deviceID,                /**< [In] Device identifier that is same as the device ID registerd with ACM*/
    gchar *busName,                 /**< [in] The name of the bus to connect a channel on.  */
    gchar *clientType,              /**< [In] client type to indicate if the connect operation is for source or for a sink */
    GHashTable *busData,            /**< [in] Bus-specific data to be processed by the ABM plugin.  */
    DBusGMethodInvocation *context  /**< [in] the method invocation context. */
);

/**************************************************************************/
/*                                                                        */
/*  Function Name: acm_disconnect                                         */
/*                                                                        */
/**************************************************************************/
/**
 *  Disconnects the specified audio device from its channel and then de-allocates 
 *  the associated audio channel.             
 *
 *  \return TRUE if successful.                 
 *                                                     
 **************************************************************************
 */
gboolean acm_disconnect
(
    AudioConnectionManager *server, /**< [in] the object instance that the method was called on. */
    gchar *deviceID,                /**< [In] Device identifier that is same as the device ID registerd with ACM*/
    gchar *busName,                 /**< [in] The name of the bus to disconnect a channel on.  */
    gchar *clientType,              /**< [In] client type to indicate if the dis connection is for source or for a sink */
    GHashTable *busData,            /**< [in] Bus-specific data to be processed by the ABM plugin.  */
    DBusGMethodInvocation *context  /**< [in] the method invocation context. */
);

/**************************************************************************/
/*           ConnectionControl DBus message handler functions             */
/**************************************************************************/

/**************************************************************************/
/*                                                                        */
/*  Function Name: acm_mute                                               */
/*                                                                        */
/**************************************************************************/
/**
 *  Mutes the audio of a specific device.               
 *
 *  \return TRUE if successful.                 
 *                                                     
 **************************************************************************
 */
gboolean acm_mute
(
    AudioConnectionManager *server, /**< [in] the object instance that the method was called on. */
    gchar *deviceID,                /**< [in] The ID of the audio device to mute. */
    DBusGMethodInvocation *context  /**< [in] the method invocation context. */
);

/**************************************************************************/
/*                                                                        */
/*  Function Name: acm_un_mute                                               */
/*                                                                        */
/**************************************************************************/
/**
 *  UnMutes the audio of a specific device.               
 *
 *  \return TRUE if successful.                 
 *                                                     
 **************************************************************************
 */
gboolean acm_un_mute
(
    AudioConnectionManager *server, /**< [in] the object instance that the method was called on. */
    gchar *deviceID,                /**< [in] The ID of the audio device to unmute. */
    DBusGMethodInvocation *context  /**< [in] the method invocation context. */
);

/**************************************************************************/
/*                                                                        */
/*  Function Name: acm_start                                              */
/*                                                                        */
/**************************************************************************/
/**
 *  Starts the audio of a specific device.               
 *
 *  \return TRUE if successful.                 
 *                                                     
 **************************************************************************
 */
gboolean acm_start
(
    AudioConnectionManager *server, /**< [in] the object instance that the method was called on. */
    gchar *deviceID,                /**< [in] The ID of the audio device to start. */
    DBusGMethodInvocation *context  /**< [in] the method invocation context. */
);

/**************************************************************************/
/*                                                                        */
/*  Function Name: acm_stop                                               */
/*                                                                        */
/**************************************************************************/
/**
 *  Stops the audio of a specific device.               
 *
 *  \return TRUE if successful.                 
 *                                                     
 **************************************************************************
 */
gboolean acm_stop
(
    AudioConnectionManager *server, /**< [in] the object instance that the method was called on. */
    gchar *deviceID,                /**< [in] The ID of the audio device to stop. */
    DBusGMethodInvocation *context  /**< [in] the method invocation context. */
);

/**************************************************************************/
/*                                                                        */
/*  Function Name: acm_pause                                               */
/*                                                                        */
/**************************************************************************/
/**
 *  Pauses the audio of a specific device.               
 *
 *  \return TRUE if successful.                 
 *                                                     
 **************************************************************************
 */
gboolean acm_pause
(
    AudioConnectionManager *server, /**< [in] the object instance that the method was called on. */
    gchar *deviceID,                /**< [in] The ID of the audio device to pause. */
    DBusGMethodInvocation *context  /**< [in] the method invocation context. */
);
#endif                              /*AUDIOCONNECTIONMANAGER_H_*/
/// @endcond
