/*
 * acmd.c 
 *
 * Copyright (c) 2008 Wind River Systems, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Lesser GNU General Public License version 2.1 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the Lesser GNU General Public License for more details.
 *
 * You should have received a copy of the Lesser GNU General Public License version 2.1
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */
/// @cond internal_documentation
/**
 * \file 
 *    Audio Connection Manager Daemon main file.    
 *
 * \version $Revision: 1.7 $
 * \date    $Date: 2008/07/12 01:27:17 $
 */

/**************************************************************************/
/*                              Includes                                  */
/**************************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <signal.h>
#include <unistd.h>
#include <sys/stat.h>
#include <glib.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-bindings.h>
#include <dbus/dbus-glib-lowlevel.h>
#include "acmd.h"
#include "AudioConnectionManager.h"
#include "AudioBusManagerPlugin.h"
#include "AudioBusManager.h"

/**************************************************************************/
/*                               Defines                                  */
/**************************************************************************/

/**************************************************************************/
/*                              Typedefs                                  */
/**************************************************************************/

/**************************************************************************/
/*                              Externs                                   */
/**************************************************************************/

/**************************************************************************/
/*                              Globals                                   */
/**************************************************************************/
/* Connection to system Dbus */

DBusGConnection *dbus_connection = NULL;

/**************************************************************************/
/*                           Local Variables                              */
/**************************************************************************/

/* Manager object to handle DBus calls from applications */
static AudioConnectionManager *manager = NULL;

/* Main loop for ACM daemon */
static GMainLoop *main_loop = NULL;

/* Command line options */
static gboolean opt_no_daemon = FALSE;
static gboolean opt_version = FALSE;


static GOptionEntry opt_entries[] =
{
{ "version", 'v', 0, G_OPTION_ARG_NONE, &opt_version, "Print version", NULL },
{ "no-daemon", 'n', 0, G_OPTION_ARG_NONE, &opt_no_daemon, "Do not daemonize", NULL },
{ NULL }
};

/**************************************************************************/
/*  Private Functions Prototypes-functions used internally in this file   */
/**************************************************************************/

/**************************************************************************/
/*                       Public Functions Bodies                         */
/**************************************************************************/

/**************************************************************************/
/*                       Private Functions Bodies                         */
/**************************************************************************/

static void daemon_log_handler
(
    const gchar *LogDomain, GLogLevelFlags LogLevel, const gchar *Message, gpointer UserData
)
{
/* Do nothing.  This function is used when running in daemon mode to keep
 * the g_message(), g_debug(), g_critical() functions from trying to use
 * the STDERR file descriptor after it has been closed.
 */
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: terminate                                              */
/*                                                                        */
/**************************************************************************/
/**
 *  Cleans up the ACM Daemon.               
 *
 *  \return TRUE if successful.                 
 *                                                     
 **************************************************************************
 */
static void terminate (
)
{
    g_debug ("%s: Shutting down Audio Connection Manager.", __FUNCTION__);

    if (manager)
    {
        acm_finalize (manager);
        manager = NULL;
    }

    error_finalize ();

    if (dbus_connection)
    {
        dbus_g_connection_unref (dbus_connection);
        dbus_connection = NULL;
    }

    ABM_Terminate ();

    return;
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: initialize                                             */
/*                                                                        */
/**************************************************************************/
/**
 *  Initializes the ACM Daemon.               
 *
 *  \return TRUE if successful.                 
 *                                                     
 **************************************************************************
 */
gboolean initialize
(
    void
)
{
    gboolean success = TRUE;
    GError *error = NULL;
    DBusGProxy *proxy;
    guint request_ret;

    g_type_init ();

    if (!g_thread_supported ())
        g_thread_init (NULL);

    dbus_g_thread_init ();

    if (!opt_no_daemon)
    {
        g_message ("Setting log handler to NULL");
        g_log_set_handler (NULL, (G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION), daemon_log_handler,
                           NULL);

        g_message ("DONE Setting log handler to NULL");
    }

    success = error_init ();

    if (success)
    {
        dbus_connection = dbus_g_bus_get (DBUS_BUS_SYSTEM, &error);
        if (!dbus_connection)
        {
            g_critical (
                G_STRLOC " Failed to open connection to bus: %s\n", error->message);
            g_error_free (error);
            success = FALSE;
            error = NULL;
        }
    }

    if (success)
    {
        /* Register a well known service name */
        proxy = dbus_g_proxy_new_for_name (dbus_connection, DBUS_SERVICE_DBUS, DBUS_PATH_DBUS, DBUS_INTERFACE_DBUS);
        if (!org_freedesktop_DBus_request_name (proxy, ACM_DBUS_SERVICE, 0, &request_ret, &error))
        {
            g_critical (
                G_STRLOC " Unable to register Audio Connection Manager service: %s\n", error->message);
            g_error_free (error);
            success = FALSE;
            error = NULL;
        }
        else if (request_ret != DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER)
        {
            g_critical (
                G_STRLOC " Unable to register DBus name %s - "
                             "Audio Connection Manager daemon already running?\n", ACM_DBUS_SERVICE);
            success = FALSE;
        }
        g_object_unref (proxy);
    }

    if (success)
    {
        /* Set up the manager object. */
        manager = (AudioConnectionManager *)g_object_new (acm_get_type (), NULL);
        if (!manager->initialized)
        {
            g_critical (
                G_STRLOC " Failed to initialize Manager server object\n");
            success = FALSE;
        }
        else
        {
            g_object_ref (manager);
        }
    }

    if (success)
    {
        success = ABM_Initialize ();
        if (!success)
        {
            g_critical (
                G_STRLOC " Failed to initialize the Audio Bus Manager!");
        }
    }

    if (!success)
    {
        terminate ();
    }

    return success;
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: sighandler                                             */
/*                                                                        */
/**************************************************************************/
/**
 *  Handles incoming signals by terminating the daemon.               
 *               
 *                                                     
 **************************************************************************
 */
void sighandler
(
    int sig /**< [in] signal number.*/
)
{
    if (main_loop)
    {
        g_main_loop_quit (main_loop);
    }
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: daemonize                                              */
/*                                                                        */
/**************************************************************************/
/**
 *  Daemonize the process.               
 *
 *  \return TRUE if successful.                 
 *                                                     
 **************************************************************************
 */
gboolean daemonize (
)
{
    char pid_buf[256];
    int pfile;

    if (signal (SIGPIPE, SIG_IGN) == SIG_ERR)
    {
        return FALSE;
    }

    pid_t pid = fork ();

    if (pid < 0)
    {
        return FALSE;
    }
    /* parent process */
    if (pid != 0)
    {
        exit (0);
    }
    /* become the session leader */
    setsid ();

    /* second fork to become a real daemon */
    pid = fork ();

    if (pid < 0)
    {
        return FALSE;
    }
    /* parent process */
    if (pid != 0)
    {
        exit (0);
    }
    /* create the pid file (overwrite of old pid file is ok) */
    unlink (ACM_PID_FILE);
    pfile = open (ACM_PID_FILE, O_WRONLY | O_CREAT, 0640);
    snprintf (pid_buf, sizeof (pid_buf), "%d\n", (int)getpid ());
    write (pfile, pid_buf, strlen (pid_buf));
    close (pfile);

    /* housekeeping */
    chdir ("/");
    umask (0);

    /* Close stdin and stdout */
    close (0);
    close (1);

    return TRUE;
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: main                                                   */
/*                                                                        */
/**************************************************************************/
/**
 *  Entry point for the Audio Connection Manager.               
 *
 *  \return 0 if successful, -1 if error.                 
 *                                                     
 **************************************************************************
 */
int main
(
    int argc,    /**< [in] Number of command line args*/
    char * *argv /**< [in] Command line args.*/
)
{
    gboolean success = TRUE;
    GError *error = NULL;
    GOptionContext *opt_context = NULL;

    opt_context = g_option_context_new ("- Audio Connection Manager Daemon");
    g_option_context_add_main_entries (opt_context, opt_entries, NULL);
    g_option_context_parse (opt_context, &argc, &argv, &error);
    g_option_context_free (opt_context);

    if (opt_version)
    {
        g_print (
            ACM_PACKAGE_NAME " version " ACM_PACKAGE_VERSION "\n");
        return 0;
    }

    if (!opt_no_daemon)
    {
        /* Split the process off as a daemon. */
        daemonize ();
    }

    signal (SIGTERM, sighandler);
    signal (SIGINT, sighandler);

    success = initialize ();

    /* Start the event loop. */
    if (success)
    {
        main_loop = g_main_loop_new (NULL, FALSE);
        g_main_loop_run (main_loop);
    }

    terminate ();

    unlink (ACM_PID_FILE);

    return (success ? 0 : -1);
}
/// @endcond
