/*
 * acmd.h 
 *
 * Copyright (c) 2008 Wind River Systems, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Lesser GNU General Public License version 2.1 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the Lesser GNU General Public License for more details.
 *
 * You should have received a copy of the Lesser GNU General Public License version 2.1
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */
/**
 * \defgroup audioconnman_daemon Audio Connection Manager Daemon
 *
 * The acmd is the controller that sits between the Audio Bus Manager
 * and the PulseAudio routing.  In response to commands sent over its
 * dbus interfaces, it requests channels and connections from the
 * Audio Bus Manager, and then sets up the corresponding routes in
 * PulseAudio.
 *
 * \section usage_sec Usage
 * <code>acmd [OPTION...]</code> - Audio Connection Manager Daemon
 *
 * Help Options:<br>
 *  <code>-?, --help</code>          Show help options
 *
 * Application Options:<br>
 *  <code>-v, --version</code>       Print version<br>
 *  <code>-n, --no-daemon</code>     Do not daemonize
 */
/**
 * \ingroup audioconnman_daemon
 * \file 
 *    The Audio Connection Manager
 *
 * \version $Revision: 1.7 $
 * \date    $Date: 2008/07/12 01:27:17 $
 */
#ifndef _ACMD_H_
#define _ACMD_H_

/**************************************************************************/
/*                              Includes                                  */
/**************************************************************************/
#include <glib.h>
#include <dbus/dbus-glib.h>
#include "acmd_errors.h"

/**************************************************************************/
/*                               Defines                                  */
/**************************************************************************/
/** The name of the ACM daemon package.*/
#define ACM_PACKAGE_NAME "acmd"

/** The version of the ACM daemon package.*/
#define ACM_PACKAGE_VERSION "0.0.1"

/** The base DBus service name for the ACM daemon.*/
#define ACM_DBUS_SERVICE "com.windriver.audio"

/** The base DBus object path for the ACM daemon.*/
#define ACM_BASE_OBJECT_PATH "/com/windriver/audio"

/** The Manager DBus interface.*/
#define ACM_MANAGER_INTERFACE "com.windriver.audio.ConnectionManager"

/** The Manager Device Control DBus Interface */
#define ACM_MANAGER_CONTROL_INTERFACE "com.windriver.audio.ConnectionControl"

/** The base error path used by all DBus objects in the ACM Daemon.*/
#define ACM_ERROR_PATH "com.windriver.audio.error"

/** The name of the Device Control Module DBus interface */
#define ACM_DEVICE_CONTROL_MODULE_DBUS_INTERFACE "com.windriver.audio.DeviceControlModule"

/** This is the keyword that must be specified in the environment variable
 * identifying the device. e.g:
 * PULSE_PROP_acmdeviceid=com.windriver.audio.system_alert
 */
#define ACM_DEVICE_CONTROL_ENVIRONMENT_KEYWORD "acmdeviceid"

/**
 * This macro handles the success and error return cases for dbus methods.  
 * The parameters are:
 * \a success gboolean to indicate if the method was successful.
 * \a context DBusGMethodInvocation* for the method.
 * \a retVal the value to be returned on success.
 * \a error a GError* to be returned on the error case.
 * 
 */
#define ACM_DBUS_RETURN(success, context, retVal, error) { \
    if (success){                                          \
        dbus_g_method_return(context, retVal);             \
    } else {                                               \
        dbus_g_method_return_error(context, error);        \
    }                                                      \
}
/**
 * This macro handles the success and error return cases for dbus methods with no
 * return values.  
 * The parameters are:
 * \a success gboolean to indicate if the method was successful.
 * \a context DBusGMethodInvocation* for the method.
 * \a error a GError* to be returned on the error case.
 * 
 */
#define ACM_DBUS_VOID_RETURN(success, context, error) { \
    if (success){                                       \
        dbus_g_method_return(context);                  \
    } else {                                            \
        dbus_g_method_return_error(context, error);     \
    }                                                   \
}
/**************************************************************************/
/*                             Typedefs                                   */
/**************************************************************************/
/** This structure represents an application that has registered as a client of
 * the Audio Connection Manager.  
 */

typedef struct
{
    gchar *deviceID;        /**< A unique identifier for this audio device.*/
    gchar *dbusServiceName; /**< The DBus service name (e.g. com.windriver.audio)*/
    gchar *dbusObjectPath; /**< The DBus object path where this object implements the 
                             DeviceControlModule interface.*/
} AudioDevice_t;

/**************************************************************************/
/*                          Functions Declarations                        */
/**************************************************************************/
/// @cond internal_documentation
/**************************************************************************/
/*                                                                        */
/*  Function Name: acm_get_type                                       */
/*                                                                        */
/**************************************************************************/
/**
 *  This function is provided by the G_DEFINE_TYPE macro.  It is used when 
 *  creating a new object and registering it on the system bus.               
 *
 *  \return the type of the GObject.                 
 *                                                     
 **************************************************************************
 */
GType acm_get_type (
);

/// @endcond

#endif //_ACMD_H_
