/*
 *  Copyright © 2007 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software may be
 * licensed only pursuant to the terms of an applicable Wind River license
 * agreement. No license to Wind River intellectual property rights is granted
 * herein. All rights not licensed by Wind River are reserved by Wind River.
 *
 */
/// @cond internal_documentation
/**
 * \file
 *    Audio manager test application for MOST plugin
 *
 *  
 *
 */

/**************************************************************************/
/*                              Includes                                  */
/**************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <memory.h>
#include <fcntl.h>
#include <string.h>

#include <glib.h>
#include <glib-object.h>
#include <glib/gprintf.h>
#include "dbus-header.h"

#include "AudioBusManagerPlugin.h"

/**************************************************************************/
/*                               Defines                                  */
/**************************************************************************/

/**************************************************************************/
/*  Private Functions Prototypes-functions used internally in this file   */
/**************************************************************************/

/**************************************************************************/
/*                               Defines                                  */
/**************************************************************************/
#define ACM_TEST_APP_VERSION_STRING "0.01"
#define ACM_TEST_APP_DECRIPTION "Audio Connection Manager test application"
#define MAX_KEY_BUFF 100
#define ACM_TEST_PCM_AUDIO_PARAM_TABLE_KEY_CHANNEL_ID  "channel_id"
#define ACM_TEST_LOCAL_AUDIO_PARAM_TABLE_KEY_CHANNEL_ID  "channel_id"

#define ACM_TEST_MOST_AUDIO_PARAM_TABLE_KEY_CHANNEL_COUNT      "channel_count"
#define ACM_TEST_MOST_AUDIO_PARAM_TABLE_KEY_DEVICE_HANDLE      "dev_hnd"
#define ACM_TEST_MOST_AUDIO_PARAM_TABLE_KEY_SOURCE_DELAY       "source_delay"
#define ACM_TEST_MOST_AUDIO_PARAM_TABLE_KEY_CHANNEL_ID_FORMAT  "channel_id-%d"

#define ACM_TEST_MOST_AUDIO_CHANNEL_COUNT 32
#define ACM_TEST_CHANNEL_CLIENT_COUNT 32
#define ACM_TEST_CHANNEL_COUNT 32

#define ACM_TEST_ALLOC_MAX_COUNT 10

#define ACM_TEST_PCM_CHANNEL_ID_LENGTH 64
#define ACM_TEST_LOCAL_CHANNEL_ID_LENGTH 64

#define ACM_TEST_PCM_BUS_NAME "pcm"
#define ACM_TEST_MOST_BUS_NAME "most"
#define ACM_TEST_LOCAL_BUS_NAME "local"
#define ACM_TEST_ALL_BUS_NAME "all"

#define ACM_TEST_SOURCE_CLIENT_TYPE_NAME "source"
#define ACM_TEST_SINK_CLIENT_TYPE_NAME "sink"
#define ACM_TEST_LOCAL_SOURCE_CLIENT_TYPE_NAME "local_source"
#define ACM_TEST_LOCAL_SINK_CLIENT_TYPE_NAME "local_sink"

#define ACM_TEST_AUDIO_PARAM_TEMP_BUF_SIZE 256
#define ACM_TEST_MOST_AUDIO_PARAM_TEMP_BUF_SIZE 10
#define ACM_TEST_MOST_AUDIO_PARAM_CHANNEL_ID_KEY_LENGTH 20
#define ACM_TEST_PARAM_STRING_LENGTH 25

#define ACM_TEST_IS_MOST(x) \
       ((strcmp(x, ACM_TEST_MOST_BUS_NAME) == 0) ? TRUE:FALSE )

/**************************************************************************/
/*                                Types                                   */
/**************************************************************************/

typedef struct
{
    int IsConnected;
    int ConnectionCount;
    char ChannelID[ACM_TEST_LOCAL_CHANNEL_ID_LENGTH];
    int ActiveConnections[ACM_TEST_CHANNEL_CLIENT_COUNT];
    ABM_AudioPluginParams_t Connections[ACM_TEST_CHANNEL_CLIENT_COUNT];
} ChannelParams_t;


typedef struct
{
    int IsAllocated;
    int ChannelCount;
    int SyncHandle;
    int SourceDelay;
    ChannelParams_t ChannelParams[ACM_TEST_CHANNEL_COUNT];
    ABM_AudioPluginParams_t AllocatorParam;
} ACM_TEST_ChannelParams_t;


ACM_TEST_ChannelParams_t sMostChannelParams[ACM_TEST_ALLOC_MAX_COUNT];
ACM_TEST_ChannelParams_t sLocalChannelParams[ACM_TEST_ALLOC_MAX_COUNT];
ACM_TEST_ChannelParams_t sPcmChannelParams[ACM_TEST_ALLOC_MAX_COUNT];

/**************************************************************************/
/*                              Globals                                   */
/**************************************************************************/
static char *gPluginName = "/usr/lib/audiomanagement/abm/plugins/libabm-most.so";
static gboolean sPrintVersion = FALSE;
static volatile gboolean gExit;
static DBusGConnection *pDbusConnection = NULL;
static DBusGProxy *pACM_Proxy = NULL;
static DBusGProxy *pACMCC_Proxy = NULL;


static GOptionEntry entries[] =
{
{ "plugin", 'p', 0, G_OPTION_ARG_STRING, &gPluginName, "Name of the plugin file", "<filename>" },
{ "version", 0, 0, G_OPTION_ARG_NONE, &sPrintVersion, "Print version", NULL},
{ NULL },
};

/**************************************************************************/
/*                                                                        */
/*  Function Name: sSignalCatch                                           */
/*                                                                        */
/**************************************************************************/
/**
 *  \return None
 **************************************************************************
*/
static void sSignalCatch
(
    int sig
)
{
    g_printf ("Received signal %d. Initiating shutdown procedure...\n", sig);
    g_object_unref (pACM_Proxy);
    g_object_unref (pACMCC_Proxy);
    gExit = TRUE;
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: sUsage                                                   */
/*                                                                        */
/**************************************************************************/
/**
 *  Print usage
 *
 *  \return None
 **************************************************************************
*/
static void sUsage
(
    void
)
{
    g_printf ("%s\n", ACM_TEST_APP_DECRIPTION);
    g_printf ("Supported commands:\n");
    g_printf ("\thelp - Print this menu\n");
    g_printf ("\talloc <DeviceID> <BusName> <ClientType> <ChannelID | ChannelCount> Allocate channels.\n");
    g_printf ("\talloc Examples-1: alloc MP3Palyer pcm local_source hw:0,0\n");
    g_printf ("\talloc Examples-2: alloc MostApp most local_source 10 \n");
    g_printf (
        "\tdealloc <BusName> <AllocIndex> - Deallocate previously allocated channels. From the allocation listing,"
            "pcmAllocParam:index=1:, AllocIndex is 1.\n");

    g_printf ("\tdealloc Example: dealloc pcm 1 \n");
    g_printf (
        "\tconnect <DeviceID> <BusName> <ClientType> <ChannelID> <SourceDelay> Connect to sync channels (sourceDelay only for most)\n");
    g_printf (
        "\tconnect command is valid only for sources of type: \"local_source\" and sinks of type: \"local_sink\"\n");
    g_printf ("\tconnect Examples-1: connect MP3Player pcm local_source hw:0,0 \n");
    g_printf ("\tconnect Examples-2: connect mostplayer most local_source 2 5 \n");
    g_printf (
        "\ticonnect <DeviceID> <BusName> <AllocIndex> <SourceDelay> Connect to sync channels (sourceDelay only for most), using alloc index\n");
    g_printf ("\ticonnect Examples-1: iconnect MP3Player pcm 0 \n");
    g_printf ("\ticonnect Examples-2: iconnect mostplayer most 1 5 \n");
    g_printf (
        "\tdisconnect <BusName> <AllocIndex> <ConnectionIndex>- Disconnect from previously connected channels. From the connection listing,"
            "ConId=0-1, AllocIndex is 0 and ConnectionIndex is 1\n");

    g_printf ("\tdisconnect Example: disconnect pcm 0 1 \n");
    g_printf ("\tlist-allocs <all | pcm | most | local> List currently allocated items \n");
    g_printf ("\tlist-allocs Example: list-allocs pcm \n");
    g_printf ("\tstart <DeviceID> Send the \"Start\" command to <DeviceID>\n");
    g_printf ("\tstop <DeviceID> Send the \"Stop\" command to <DeviceID>\n");
    g_printf ("\tmute <DeviceID> Send the \"Mute\" command to <DeviceID>\n");
    g_printf ("\tunmute <DeviceID> Send the \"UnMute\" command to <DeviceID>\n");
    g_printf ("\tpause <DeviceID> Send the \"Pause\" command to <DeviceID>\n");
    g_printf ("Press <Crtl-C> to exit.\n");
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: sProcessStdIn                                          */
/*                                                                        */
/**************************************************************************/
/**
 *  Read char from STDIN. If '\n' is detected, process command line 
 *
 *  \return FALSE - no action
 *          TRUE  - end of line detected. pKeyBuff is valid.
 **************************************************************************
*/
static gboolean sProcessStdIn
(
    char * *pKeyBuff
)
{
    static char KeyBuffer[MAX_KEY_BUFF];
    static int KeyLen;
    int readcount = read (STDIN_FILENO, &KeyBuffer[KeyLen], 1);

    if (readcount <= 0)
    {
        return FALSE;
    }

    if ('\n' == KeyBuffer[KeyLen])
    {
        KeyBuffer[KeyLen] = 0;
        *pKeyBuff = KeyBuffer;
        KeyLen = 0;
        return TRUE;
    }
    else
    {
        KeyLen++;
    }

    return FALSE;
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: setNonblocking                                         */
/*                                                                        */
/*************************************************************************/
/**
 *  \return None
 **************************************************************************
*/
int setNonblocking
(
    int fd
)
{
    int flags;

    /* If they have O_NONBLOCK, use the Posix way to do it */


    #if defined(O_NONBLOCK)

    /* Fixme: O_NONBLOCK is defined but broken on SunOS 4.1.x and AIX 3.2.5. */
    if (-1 == (flags = fcntl (fd, F_GETFL, 0)))
        flags = 0;

    return fcntl (fd, F_SETFL, flags | O_NONBLOCK);
    #else

    /* Otherwise, use the old way of doing it */
    flags = 1;
    return ioctl (fd, FIOBIO, &flags);
    #endif

}


static void InitChannelParams
(
    ACM_TEST_ChannelParams_t pChannelParams[]
)
{
    int i;
    int j;
    int k;

    for (i = 0; i < ACM_TEST_ALLOC_MAX_COUNT; i++)
    {
        for (j = 0; j < ACM_TEST_CHANNEL_COUNT; j++)
        {
            (pChannelParams)[i].ChannelParams[j].IsConnected = FALSE;
            (pChannelParams)[i].ChannelParams[j].ConnectionCount = 0;
            for (k = 0; k < ACM_TEST_CHANNEL_CLIENT_COUNT; k++)
            (pChannelParams)[i].ChannelParams[j].ActiveConnections[k] = 0;
        }
        (pChannelParams)[i].IsAllocated = FALSE;
        (pChannelParams)[i].ChannelCount = 0;
    }
}


static void sPrintPluginParams
(
    ABM_AudioPluginParams_t PluginParam
)
{
    g_printf ("DevID: %s, Bus: %s, Client: %s ", PluginParam.DeviceID, PluginParam.ClientType, PluginParam.PluginType);
}


static void sPopulateChannelAllocateParams
(
    GHashTable *pTable, ABM_AudioPluginParams_t PluginParam, ACM_TEST_ChannelParams_t pChannelParams[]
)
{
    int i = 0;
    int j = 0;
    gchar *pData;
    gchar Tempbuffer[ACM_TEST_MOST_AUDIO_PARAM_TEMP_BUF_SIZE];

    for (i = 0; i < ACM_TEST_ALLOC_MAX_COUNT; i++)
    {
        if(!((pChannelParams)[i].IsAllocated))
        break;
    }

    if ((ACM_TEST_ALLOC_MAX_COUNT - 1) == i)
    {
        g_printf ("Unabled to store allocation data in ChannelParams \n");
        return;
    }

    memcpy (& ((pChannelParams)[i].AllocatorParam), &PluginParam, sizeof (ABM_AudioPluginParams_t));

    if (ACM_TEST_IS_MOST (PluginParam.PluginType))
    {
        pData = g_hash_table_lookup (pTable, ACM_TEST_MOST_AUDIO_PARAM_TABLE_KEY_CHANNEL_COUNT);
        if (NULL == pData)
        {
            g_printf (
                "Error: parameter %s not found in hash table\n", ACM_TEST_MOST_AUDIO_PARAM_TABLE_KEY_CHANNEL_COUNT);
            return;
        }
        (pChannelParams)[i].ChannelCount = atoi (pData);

        pData = g_hash_table_lookup (pTable, ACM_TEST_MOST_AUDIO_PARAM_TABLE_KEY_DEVICE_HANDLE);
        if (NULL == pData)
        {
            g_printf (
                "Error: parameter %s not found in hash table\n", ACM_TEST_MOST_AUDIO_PARAM_TABLE_KEY_DEVICE_HANDLE);
            return;
        }
        (pChannelParams)[i].SyncHandle = atoi (pData);

        pData = g_hash_table_lookup (pTable, ACM_TEST_MOST_AUDIO_PARAM_TABLE_KEY_SOURCE_DELAY);
        if (NULL == pData)
        {
            g_printf (
                "Error: parameter %s not found in hash table\n", ACM_TEST_MOST_AUDIO_PARAM_TABLE_KEY_SOURCE_DELAY);
            return;
        }
        (pChannelParams)[i].SourceDelay = atoi (pData);
        for (j = 0; j < (pChannelParams)[i].ChannelCount; j++)
        {
            sprintf (Tempbuffer, ACM_TEST_MOST_AUDIO_PARAM_TABLE_KEY_CHANNEL_ID_FORMAT, j);
            pData = g_hash_table_lookup (pTable, Tempbuffer);
            strcpy ((pChannelParams)[i].ChannelParams[j].ChannelID, pData);
        }
    }
    else
    {
        g_printf ("Adding %d sPopulateChannelAllocateParams\n", i);
        pData = g_hash_table_lookup (pTable, ACM_TEST_PCM_AUDIO_PARAM_TABLE_KEY_CHANNEL_ID);
        if (NULL == pData)
        {
            g_printf ("Error: parameter %s not found in hash table\n", ACM_TEST_PCM_AUDIO_PARAM_TABLE_KEY_CHANNEL_ID);
            return;
        }
        strcpy ((pChannelParams)[i].ChannelParams[j].ChannelID, pData);
        (pChannelParams)[i].ChannelCount = 1;
    }

    (pChannelParams)[i].IsAllocated = TRUE;
}


static void sPopulateChannelConnectParams
(
    GHashTable *pTable, ABM_AudioPluginParams_t PluginParam, ACM_TEST_ChannelParams_t pChannelParams[]
)
{
    int i;
    int j;
    int k;

    gchar *pData;
    gboolean ChannelFound = FALSE;
    gboolean ConnectionIndexFound = FALSE;
    int ConnectionIndex = 0;

    pData = g_hash_table_lookup (pTable, ACM_TEST_PCM_AUDIO_PARAM_TABLE_KEY_CHANNEL_ID);

    if (NULL == pData)
    {
        g_printf ("Error: parameter %s not found in hash table\n", ACM_TEST_PCM_AUDIO_PARAM_TABLE_KEY_CHANNEL_ID);
        return;
    }
    /* now locate the channel Id in the chane params array*/
    for (i = 0; i < ACM_TEST_ALLOC_MAX_COUNT; i++)
    {
        if (((pChannelParams)[i].IsAllocated))
        {
            for (j = 0; j < ACM_TEST_CHANNEL_COUNT; j++)
            {
                if (strcmp ((pChannelParams)[i].ChannelParams[j].ChannelID, pData) == 0)
                {
                    ChannelFound = TRUE;
                    break;
                }
            }
        }
        if (ChannelFound)
            break;
    }

    if (ChannelFound)
    {
        /* find a free connection index*/
        for (k = 0; k < ACM_TEST_CHANNEL_CLIENT_COUNT; k++)
        {
            if (!(pChannelParams)[i].ChannelParams[j].ActiveConnections[ConnectionIndex])
            {
                ConnectionIndexFound = TRUE;
                break;
            }
        }
        if (ConnectionIndexFound)
        {
            ConnectionIndex = k;
            memcpy (& ((pChannelParams)[i].ChannelParams[j].Connections[ConnectionIndex]), &PluginParam, sizeof (
                                                                                                             ABM_AudioPluginParams_t));
            (pChannelParams)[i].ChannelParams[j].ActiveConnections[ConnectionIndex] = TRUE;
            (pChannelParams)[i].ChannelParams[j].IsConnected = TRUE;
        }
        else
        {
            g_printf ("Error: Connection Index is not available\n");
        }
    }
    else
    {
        g_printf ("Error: Channel ID: %s is not allocated\n", pData);
    }
}


static void sPopulateChannelDisconnectParams
(
    ACM_TEST_ChannelParams_t pChannelParams[], int AllocParamIndex, int ConnectionIndex
)
{
    (pChannelParams)[AllocParamIndex].ChannelParams[0].ConnectionCount--;
    (pChannelParams)[AllocParamIndex].ChannelParams[0].ActiveConnections[ConnectionIndex] = FALSE;

    if (!(pChannelParams)[AllocParamIndex].ChannelParams[0].ConnectionCount)
    {
        (pChannelParams)[AllocParamIndex].ChannelParams[0].IsConnected = FALSE;
    }
}


static void sPopulateChannelDeallocParams
(
    ACM_TEST_ChannelParams_t pChannelParams[], int AllocParamIndex
)
{
    int k;
    (pChannelParams)[AllocParamIndex].ChannelParams[0].ConnectionCount = 0;
    (pChannelParams)[AllocParamIndex].IsAllocated = FALSE;
    (pChannelParams)[AllocParamIndex].ChannelParams[0].IsConnected = FALSE;

    for (k = 0; k < ACM_TEST_CHANNEL_CLIENT_COUNT; k++)
    {
        (pChannelParams)[AllocParamIndex].ChannelParams[0].ActiveConnections[k] = FALSE;
    }
}


static void sPrintChannelAllocParams
(
    ACM_TEST_ChannelParams_t pChannelParams[]
)
{
    int i;
    int j;
    int k;

    for (i = 0; i < ACM_TEST_ALLOC_MAX_COUNT; i++)
    {
        if (((pChannelParams)[i].IsAllocated))
        {
            g_printf ("%sAllocParam:index=%d: ", (pChannelParams)[i].AllocatorParam.PluginType, i);
            sPrintPluginParams ((pChannelParams)[i].AllocatorParam);
            if (ACM_TEST_IS_MOST ((pChannelParams)[i].AllocatorParam.PluginType))
            {
                g_printf (", SyncH: %d, SDelay: %d", (pChannelParams)[i].SyncHandle, (pChannelParams)[i].SourceDelay);
            }
            g_printf (", Channels: \n");
            for (j = 0; j < (pChannelParams)[i].ChannelCount; j++)
            {
                g_printf ("ChId: %s Connections: \n", (pChannelParams)[i].ChannelParams[j].ChannelID);
                for (k = 0; k < ACM_TEST_CHANNEL_CLIENT_COUNT; k++)
                {
                    if ((pChannelParams)[i].ChannelParams[j].ActiveConnections[k])
                    {
                        g_printf ("\tConId=%d-%d: ", i, k);
                        sPrintPluginParams ((pChannelParams)[i].ChannelParams[j].Connections[k]);
                        g_printf ("\n");
                    }
                }
            }
        }
    }
}


static void sPrintAllocs
(
    char *pBusType
)
{
    if (!pBusType)
    {
        return;
    }

    if (strcmp (pBusType, ACM_TEST_PCM_BUS_NAME) == 0)
    {
        sPrintChannelAllocParams (sPcmChannelParams);
    }
    else if (strcmp (pBusType, ACM_TEST_MOST_BUS_NAME) == 0)
    {
        sPrintChannelAllocParams (sMostChannelParams);
    }
    else if (strcmp (pBusType, ACM_TEST_LOCAL_BUS_NAME) == 0)
    {
        sPrintChannelAllocParams (sLocalChannelParams);
    }
    else if (strcmp (pBusType, ACM_TEST_ALL_BUS_NAME) == 0)
    {
        sPrintChannelAllocParams (sPcmChannelParams);
        sPrintChannelAllocParams (sMostChannelParams);
        sPrintChannelAllocParams (sLocalChannelParams);
    }
    else
    {
        g_printf ("Can not list Allocs, Invalid Bus Type!\n");
    }
}


static gboolean sCheckInputsAlloc
(
    int argc,     /**< number of arguments including application name */
    char * *argv, /**< array of string arguments */
    ABM_AudioPluginParams_t *pPluginParams
)
{
    if (argc > 2)
    {
        if ((strcmp (argv[2], ACM_TEST_PCM_BUS_NAME) == 0)
                || (strcmp (argv[2], ACM_TEST_MOST_BUS_NAME) == 0) || (strcmp (argv[2], ACM_TEST_LOCAL_BUS_NAME) == 0))
        {
            strcpy (pPluginParams->PluginType, argv[2]);
        }
        else
        {
            g_printf ("unknown bus name: allowed names: %s, %s, %s !\n", ACM_TEST_PCM_BUS_NAME, ACM_TEST_MOST_BUS_NAME,
                      ACM_TEST_LOCAL_BUS_NAME);
            return FALSE;
        }
        if ((strcmp (argv[3], ACM_TEST_SOURCE_CLIENT_TYPE_NAME) == 0)
                || (strcmp (argv[3], ACM_TEST_SINK_CLIENT_TYPE_NAME) == 0)
                || (strcmp (argv[3], ACM_TEST_LOCAL_SOURCE_CLIENT_TYPE_NAME) == 0)
                || (strcmp (argv[3], ACM_TEST_LOCAL_SINK_CLIENT_TYPE_NAME) == 0))
        {
            strcpy (pPluginParams->ClientType, argv[3]);
        }
        else
        {
            g_printf (
                "unknown client type name: allowed names: %s, %s, %s, %s !\n", ACM_TEST_SOURCE_CLIENT_TYPE_NAME,
                ACM_TEST_SINK_CLIENT_TYPE_NAME,                                ACM_TEST_LOCAL_SOURCE_CLIENT_TYPE_NAME,
                ACM_TEST_LOCAL_SINK_CLIENT_TYPE_NAME);
            return FALSE;
        }
        strcpy (pPluginParams->DeviceID, argv[1]);
        return TRUE;
    }
    else
    {
        return FALSE;
    }
}


void sHashTableInsert
(
    GHashTable *pTable, gchar *pKey, gchar *pValue
)
{
    gchar *pTabKey;
    gchar *pTabValue;

    pTabKey = g_malloc (strlen (pKey) + 1);
    pTabValue = g_malloc (strlen (pValue) + 1);

    strcpy (pTabKey, pKey);
    strcpy (pTabValue, pValue);

    g_hash_table_insert (pTable, pTabKey, pTabValue);
}


static gboolean sCheckInputsConnect
(
    int argc,     /**< number of arguments including application name */
    char * *argv, /**< array of string arguments */
    ABM_AudioPluginParams_t *pPluginParams
)
{
    if (argc == 3)
    {
        if ((strcmp (argv[1], ACM_TEST_PCM_BUS_NAME) == 0)
                || (strcmp (argv[1], ACM_TEST_MOST_BUS_NAME) == 0) || (strcmp (argv[1], ACM_TEST_LOCAL_BUS_NAME) == 0))
        {
            strcpy (pPluginParams->PluginType, argv[1]);
        }
        else
        {
            g_printf ("unknown bus name: allowed names: %s, %s, %s !\n", ACM_TEST_PCM_BUS_NAME, ACM_TEST_MOST_BUS_NAME,
                      ACM_TEST_LOCAL_BUS_NAME);
            return FALSE;
        }
        return TRUE;
    }

    if (argc > 3)
    {
        if ((strcmp (argv[2], ACM_TEST_PCM_BUS_NAME) == 0)
                || (strcmp (argv[2], ACM_TEST_MOST_BUS_NAME) == 0) || (strcmp (argv[2], ACM_TEST_LOCAL_BUS_NAME) == 0))
        {
            strcpy (pPluginParams->PluginType, argv[2]);
        }
        else
        {
            g_printf ("unknown bus name: allowed names: %s, %s, %s !\n", ACM_TEST_PCM_BUS_NAME, ACM_TEST_MOST_BUS_NAME,
                      ACM_TEST_LOCAL_BUS_NAME);
            return FALSE;
        }
        if ((strcmp (argv[3], ACM_TEST_SOURCE_CLIENT_TYPE_NAME) == 0)
                || (strcmp (argv[3], ACM_TEST_SINK_CLIENT_TYPE_NAME) == 0)
                || (strcmp (argv[3], ACM_TEST_LOCAL_SOURCE_CLIENT_TYPE_NAME) == 0)
                || (strcmp (argv[3], ACM_TEST_LOCAL_SINK_CLIENT_TYPE_NAME) == 0))
        {
            strcpy (pPluginParams->ClientType, argv[3]);
        }
        else
        {
            g_printf (
                "unknown client type name: allowed names: %s, %s, %s, %s !\n", ACM_TEST_SOURCE_CLIENT_TYPE_NAME,
                ACM_TEST_SINK_CLIENT_TYPE_NAME,                                ACM_TEST_LOCAL_SOURCE_CLIENT_TYPE_NAME,
                ACM_TEST_LOCAL_SINK_CLIENT_TYPE_NAME);
            return FALSE;
        }
        strcpy (pPluginParams->DeviceID, argv[1]);
    }

    return TRUE;
}


static gboolean sCheckInputsDisonnect
(
    int argc,     /**< number of arguments including application name */
    char * *argv, /**< array of string arguments */
    ABM_AudioPluginParams_t *pPluginParams
)
{
    if (argc == 4)
    {
        if ((strcmp (argv[1], ACM_TEST_PCM_BUS_NAME) == 0)
                || (strcmp (argv[1], ACM_TEST_MOST_BUS_NAME) == 0) || (strcmp (argv[1], ACM_TEST_LOCAL_BUS_NAME) == 0))
        {
            strcpy (pPluginParams->PluginType, argv[1]);
        }
        else
        {
            g_printf ("unknown bus name: allowed names: %s, %s, %s !\n", ACM_TEST_PCM_BUS_NAME, ACM_TEST_MOST_BUS_NAME,
                      ACM_TEST_LOCAL_BUS_NAME);
            return FALSE;
        }
        return TRUE;
    }
    else
    {
        return FALSE;
    }
}


static gboolean sCheckInputsDealloc
(
    int argc,     /**< number of arguments including application name */
    char * *argv, /**< array of string arguments */
    ABM_AudioPluginParams_t *pPluginParams
)
{
    if (argc == 3)
    {
        if ((strcmp (argv[1], ACM_TEST_PCM_BUS_NAME) == 0)
                || (strcmp (argv[1], ACM_TEST_MOST_BUS_NAME) == 0) || (strcmp (argv[1], ACM_TEST_LOCAL_BUS_NAME) == 0))
        {
            strcpy (pPluginParams->PluginType, argv[1]);
        }
        else
        {
            g_printf ("unknown bus name: allowed names: %s, %s, %s !\n", ACM_TEST_PCM_BUS_NAME, ACM_TEST_MOST_BUS_NAME,
                      ACM_TEST_LOCAL_BUS_NAME);
            return FALSE;
        }
        return TRUE;
    }
    else
    {
        return FALSE;
    }
}


static gboolean sAlloc
(
    int argc,    /**< number of arguments including application name */
    char * *argv /**< array of string arguments */
)
{
    GHashTable *allocateReturnTable;
    GHashTable *allocateBusData;
    gchar ClientTypeName[ACM_TEST_PARAM_STRING_LENGTH];
    gchar PluginTypeName[ACM_TEST_PARAM_STRING_LENGTH];

    ABM_AudioPluginParams_t PluginParams;
    GError *error;
    gboolean success = TRUE;

    /* Set up the tables to handle the input and return data. */
    allocateReturnTable = g_hash_table_new (g_str_hash, g_str_equal);
    allocateBusData = g_hash_table_new (g_str_hash, g_str_equal);

    success = sCheckInputsAlloc (argc, argv, &PluginParams);

    if (success)
    {
        if (strcmp (PluginParams.PluginType, ACM_TEST_MOST_BUS_NAME) == 0)
        {
            if (argc < 5)
            {
                g_printf ("Error: Too few arguments for allocate on most bus !\n");
                success = FALSE;
            }
            g_hash_table_insert (allocateBusData, ACM_TEST_MOST_AUDIO_PARAM_TABLE_KEY_CHANNEL_COUNT, argv[4]);
        }
        else if (strcmp (PluginParams.PluginType, ACM_TEST_PCM_BUS_NAME) == 0)
        {
            if (argc < 5)
            {
                g_printf ("Error: Too few arguments for allocate on pcm bus !\n");
                success = FALSE;
            }
            else
            {
                g_printf ("PluginParams.PluginType %s\n", PluginParams.PluginType);
                g_hash_table_insert (allocateBusData, ACM_TEST_PCM_AUDIO_PARAM_TABLE_KEY_CHANNEL_ID, argv[4]);
            }
        }
        else if (strcmp (PluginParams.PluginType, ACM_TEST_LOCAL_BUS_NAME) == 0)
        {}
    }

    if (success)
    {
        g_printf ("PluginParams.PluginType %s\n", PluginParams.PluginType);
        strcpy (ClientTypeName, argv[3]);
        strcpy (PluginTypeName, argv[2]);
        g_printf ("--%s, %s, %s \n", PluginParams.DeviceID, ClientTypeName, PluginTypeName);

        /* Call allocate */
        error = NULL;
        if (!dbus_g_proxy_call (pACM_Proxy,                          "Allocate",            &error,
                                G_TYPE_STRING,                       PluginParams.DeviceID, G_TYPE_STRING,
                                PluginTypeName,                      G_TYPE_STRING,         ClientTypeName,
                                DBUS_TYPE_G_STRING_STRING_HASHTABLE, allocateBusData,       G_TYPE_INVALID,
                                DBUS_TYPE_G_STRING_STRING_HASHTABLE, &allocateReturnTable,  G_TYPE_INVALID))
        {
            /* Just do demonstrate remote exceptions versus regular GError */
            if (error->domain == DBUS_GERROR && error->code == DBUS_GERROR_REMOTE_EXCEPTION)
            {
                g_printerr ("Caught remote method exception %s: %s", dbus_g_error_get_name (error), error->message);
            }
            else
            {
                g_printerr ("Error: %s\n", error->message);
            }
            g_error_free (error);
            success = FALSE;
        }
    }

    if (success)
    {
        if (strcmp (PluginParams.PluginType, ACM_TEST_MOST_BUS_NAME) == 0)
        {
            sPopulateChannelAllocateParams (allocateReturnTable, PluginParams, sMostChannelParams);
        }
        else if (strcmp (PluginParams.PluginType, ACM_TEST_PCM_BUS_NAME) == 0)
        {
            sPopulateChannelAllocateParams (allocateReturnTable, PluginParams, sPcmChannelParams);
        }
        else if (strcmp (PluginParams.PluginType, ACM_TEST_LOCAL_BUS_NAME) == 0)
        {
            sPopulateChannelAllocateParams (allocateReturnTable, PluginParams, sLocalChannelParams);
        }
        g_printf ("Alloc Success\n");
        sPrintAllocs (argv[2]);
    }
    else
    {
        g_printf ("Alloc Failure\n");
    }

    g_hash_table_unref (allocateBusData);
    g_hash_table_unref (allocateReturnTable);

    return success;
}


static gboolean sConnect
(
    int argc,     /**< number of arguments including application name */
    char * *argv, /**< array of string arguments */
    int useIndex  /**< if an alloc index can be used to get connect params*/
)
{
    GHashTable *connectReturnTable;
    GHashTable *connectBusData;
    ABM_AudioPluginParams_t PluginParams;
    ABM_AudioPluginParams_t *pCommonPluginParams = NULL;
    GError *error;
    gboolean success = TRUE;
    int AllocParamIndex;
    gchar Tempbuffer[ACM_TEST_MOST_AUDIO_PARAM_TEMP_BUF_SIZE];
    gchar ChannelIdKey[ACM_TEST_MOST_AUDIO_PARAM_CHANNEL_ID_KEY_LENGTH];
    int i;
    gchar ClientTypeName[ACM_TEST_PARAM_STRING_LENGTH];
    gchar PluginTypeName[ACM_TEST_PARAM_STRING_LENGTH];
    gchar DeviceName[ABM_MAX_DEVICE_ID_LENGTH];

    /* Set up the tables to handle the input and return data. */
    connectBusData = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);
    connectReturnTable = g_hash_table_new (g_str_hash, g_str_equal);

    if (useIndex)
    {
        if (argc < 3)
        {
            g_printf ("Error: Too few arguments for connect !\n");
            success = FALSE;
        }
        if (success)
        {
            AllocParamIndex = atoi (argv[2]);
            if (AllocParamIndex >= ACM_TEST_ALLOC_MAX_COUNT)
            {
                g_printf ("Error: Invalid Allocation Index\n");
                success = FALSE;
            }
        }
    }

    if (success)
    {
        success = sCheckInputsConnect (argc, argv, &PluginParams);
        if (strcmp (PluginParams.PluginType, ACM_TEST_MOST_BUS_NAME) == 0)
        {
            if (useIndex)
            {
                if (argc < 4)
                {
                    g_printf ("Error: Too few arguments for connect on most bus !\n");
                    success = FALSE;
                }
                if (sMostChannelParams[AllocParamIndex].IsAllocated == FALSE)
                {
                    g_printf ("Error: Index not Allocated\n");
                    success = FALSE;
                }
                if (success)
                {
                    g_sprintf (Tempbuffer, "%d", sMostChannelParams[AllocParamIndex].ChannelCount);

                    sHashTableInsert (connectBusData, ACM_TEST_MOST_AUDIO_PARAM_TABLE_KEY_CHANNEL_COUNT, Tempbuffer);

                    g_sprintf (Tempbuffer, "%s", argv[3]);
                    sHashTableInsert (connectBusData, ACM_TEST_MOST_AUDIO_PARAM_TABLE_KEY_SOURCE_DELAY, Tempbuffer);
                    for (i = 0; i < sMostChannelParams[AllocParamIndex].ChannelCount; i++)
                    {
                        g_sprintf (ChannelIdKey, ACM_TEST_MOST_AUDIO_PARAM_TABLE_KEY_CHANNEL_ID_FORMAT, i);
                        sHashTableInsert (connectBusData, ChannelIdKey,
                                          sMostChannelParams[AllocParamIndex].ChannelParams[i].ChannelID);
                    }
                    pCommonPluginParams = &sMostChannelParams[AllocParamIndex].AllocatorParam;
                }
            }
            else
            {
                pCommonPluginParams = &PluginParams;
            }
        }
        else if (strcmp (PluginParams.PluginType, ACM_TEST_PCM_BUS_NAME) == 0)
        {
            if (useIndex)
            {
                if (sPcmChannelParams[AllocParamIndex].IsAllocated == FALSE)
                {
                    g_printf ("Error: Index not Allocated\n");
                    success = FALSE;
                }
                else
                {
                    sHashTableInsert (connectBusData, ACM_TEST_PCM_AUDIO_PARAM_TABLE_KEY_CHANNEL_ID,
                                      sPcmChannelParams[AllocParamIndex].ChannelParams[0].ChannelID);
                    pCommonPluginParams = &sPcmChannelParams[AllocParamIndex].AllocatorParam;
                    g_printf ("param: %s, %s, %s\n", pCommonPluginParams->DeviceID, pCommonPluginParams->PluginType,
                              pCommonPluginParams->ClientType);
                }
            }
            else
            {
                if (argc < 5)
                {
                    g_printf ("Error: Too few arguments for connect on pcm bus !\n");
                    success = FALSE;
                }
                else
                {
                    sHashTableInsert (connectBusData, ACM_TEST_PCM_AUDIO_PARAM_TABLE_KEY_CHANNEL_ID, argv[4]);
                    pCommonPluginParams = &PluginParams;
                }
            }
        }
        else if (strcmp (PluginParams.PluginType, ACM_TEST_LOCAL_BUS_NAME) == 0)
        {
            if (useIndex)
            {
                if (sLocalChannelParams[AllocParamIndex].IsAllocated == FALSE)
                {
                    g_printf ("Error: Index not Allocated\n");
                    success = FALSE;
                }
                else
                {
                    sHashTableInsert (connectBusData, ACM_TEST_PCM_AUDIO_PARAM_TABLE_KEY_CHANNEL_ID,
                                      sLocalChannelParams[AllocParamIndex].ChannelParams[0].ChannelID);
                    pCommonPluginParams = &sLocalChannelParams[AllocParamIndex].AllocatorParam;
                }
            }
            else
            {
                if (argc < 5)
                {
                    g_printf ("Error: Too few arguments for connect on pcm bus !\n");
                    success = FALSE;
                }
                else
                {
                    sHashTableInsert (connectBusData, ACM_TEST_PCM_AUDIO_PARAM_TABLE_KEY_CHANNEL_ID, argv[4]);
                    pCommonPluginParams = &PluginParams;
                }
            }
        }
    }

    if (success)
    {
        if (useIndex)
        {
            if (pCommonPluginParams)
            {
                strcpy (ClientTypeName, pCommonPluginParams->ClientType);
                ClientTypeName[strlen (pCommonPluginParams->ClientType)] = '\0';
                strcpy (PluginTypeName, pCommonPluginParams->PluginType);
                PluginTypeName[strlen (pCommonPluginParams->PluginType)] = '\0';
                strcpy (DeviceName, pCommonPluginParams->DeviceID);
                DeviceName[strlen (pCommonPluginParams->DeviceID)] = '\0';
            }
        }
        else
        {
            strcpy (DeviceName, argv[1]);
            strcpy (PluginTypeName, argv[2]);
            strcpy (ClientTypeName, argv[3]);
        }
        g_printf ("--%s, %s, %s \n", DeviceName, ClientTypeName, PluginTypeName);

        /* Call connect */
        error = NULL;
        if (!dbus_g_proxy_call (pACM_Proxy,                          "Connect",           &error,
                                G_TYPE_STRING,                       DeviceName,          G_TYPE_STRING,
                                PluginTypeName,                      G_TYPE_STRING,       ClientTypeName,
                                DBUS_TYPE_G_STRING_STRING_HASHTABLE, connectBusData,      G_TYPE_INVALID,
                                DBUS_TYPE_G_STRING_STRING_HASHTABLE, &connectReturnTable, G_TYPE_INVALID))
        {
            /* Just do demonstrate remote exceptions versus regular GError */
            if (error->domain == DBUS_GERROR && error->code == DBUS_GERROR_REMOTE_EXCEPTION)
            {
                g_printerr ("Caught remote method exception %s: %s", dbus_g_error_get_name (error), error->message);
            }
            else
            {
                g_printerr ("Error: %s\n", error->message);
            }
            g_error_free (error);
            success = FALSE;
        }
    }

    if (success)
    {
        if (strcmp (PluginParams.PluginType, ACM_TEST_MOST_BUS_NAME) == 0)
        {
            sPopulateChannelConnectParams (connectReturnTable, *pCommonPluginParams, sMostChannelParams);
        }
        else if (strcmp (PluginParams.PluginType, ACM_TEST_PCM_BUS_NAME) == 0)
        {
            sPopulateChannelConnectParams (connectReturnTable, *pCommonPluginParams, sPcmChannelParams);
        }
        else if (strcmp (PluginParams.PluginType, ACM_TEST_LOCAL_BUS_NAME) == 0)
        {
            sPopulateChannelConnectParams (connectReturnTable, *pCommonPluginParams, sLocalChannelParams);
        }
        g_printf ("Connect Success\n");
        sPrintAllocs (PluginTypeName);
    }
    else
    {
        g_printf ("Connect Failure\n");
    }

    g_hash_table_unref (connectBusData);
    g_hash_table_unref (connectReturnTable);
    return success;
}


static gboolean sDisconnect
(
    int argc,    /**< number of arguments including application name */
    char * *argv /**< array of string arguments */
)
{
    GHashTable *disConnectBusData;
    GHashTable *disconnectReturnTable;
    ABM_AudioPluginParams_t PluginParams;
    ABM_AudioPluginParams_t *pCommonPluginParams = NULL;
    GError *error;
    gboolean success = TRUE;
    int AllocParamIndex, ConnectionIndex;
    gchar Tempbuffer[ACM_TEST_MOST_AUDIO_PARAM_TEMP_BUF_SIZE];

    gchar ClientTypeName[ACM_TEST_PARAM_STRING_LENGTH];
    gchar PluginTypeName[ACM_TEST_PARAM_STRING_LENGTH];

    /* Set up the tables to handle the input and return data. */
    disConnectBusData = g_hash_table_new (g_str_hash, g_str_equal);
    disconnectReturnTable = g_hash_table_new (g_str_hash, g_str_equal);
    success = sCheckInputsDisonnect (argc, argv, &PluginParams);

    if (success)
    {
        AllocParamIndex = atoi (argv[2]);
        ConnectionIndex = atoi (argv[3]);
        if (AllocParamIndex >= ACM_TEST_ALLOC_MAX_COUNT)
        {
            g_printf ("Error: Invalid Allocation Index\n");
            success = FALSE;
        }
        if (AllocParamIndex >= ACM_TEST_CHANNEL_CLIENT_COUNT)
        {
            g_printf ("Error: Invalid connection Index\n");
            success = FALSE;
        }
        if (success)
        {
            if (strcmp (PluginParams.PluginType, ACM_TEST_MOST_BUS_NAME) == 0)
            {
                if (sMostChannelParams[AllocParamIndex].IsAllocated == FALSE)
                {
                    g_printf ("Error: Index not Allocated\n");
                    success = FALSE;
                }
                g_sprintf (Tempbuffer, "%d", sMostChannelParams[AllocParamIndex].SyncHandle);
                g_hash_table_insert (disConnectBusData, ACM_TEST_MOST_AUDIO_PARAM_TABLE_KEY_DEVICE_HANDLE, Tempbuffer);
                pCommonPluginParams
                    = &sMostChannelParams[AllocParamIndex].ChannelParams[0].Connections[ConnectionIndex];
            }
            else if (strcmp (PluginParams.PluginType, ACM_TEST_PCM_BUS_NAME) == 0)
            {
                if (sPcmChannelParams[AllocParamIndex].IsAllocated == FALSE)
                {
                    g_printf ("Error: Index not Allocated\n");
                    success = FALSE;
                }
                else
                {
                    g_hash_table_insert (disConnectBusData, ACM_TEST_PCM_AUDIO_PARAM_TABLE_KEY_CHANNEL_ID,
                                         sPcmChannelParams[AllocParamIndex].ChannelParams[0].ChannelID);
                    pCommonPluginParams
                        = &sPcmChannelParams[AllocParamIndex].ChannelParams[0].Connections[ConnectionIndex];
                }
            }
            else if (strcmp (PluginParams.PluginType, ACM_TEST_LOCAL_BUS_NAME) == 0)
            {
                if (sLocalChannelParams[AllocParamIndex].IsAllocated == FALSE)
                {
                    g_printf ("Error: Index not Allocated\n");
                    success = FALSE;
                }
                else
                {
                    g_hash_table_insert (disConnectBusData, ACM_TEST_PCM_AUDIO_PARAM_TABLE_KEY_CHANNEL_ID,
                                         sLocalChannelParams[AllocParamIndex].ChannelParams[0].ChannelID);
                    pCommonPluginParams
                        = &sLocalChannelParams[AllocParamIndex].ChannelParams[0].Connections[ConnectionIndex];
                }
            }
        }
    }

    if (success)
    {
        if (pCommonPluginParams)
        {
            strcpy (ClientTypeName, pCommonPluginParams->ClientType);
            strcpy (PluginTypeName, pCommonPluginParams->PluginType);
        }
        /* Call disconnect */
        error = NULL;
        if (!dbus_g_proxy_call (pACM_Proxy,                          "Disconnect",                  &error,
                                G_TYPE_STRING,                       pCommonPluginParams->DeviceID, G_TYPE_STRING,
                                PluginTypeName,                      G_TYPE_STRING,                 ClientTypeName,
                                DBUS_TYPE_G_STRING_STRING_HASHTABLE, disConnectBusData,             G_TYPE_INVALID,
                                DBUS_TYPE_G_STRING_STRING_HASHTABLE, &disconnectReturnTable,        G_TYPE_INVALID))
        {
            /* Just do demonstrate remote exceptions versus regular GError */
            if (error->domain == DBUS_GERROR && error->code == DBUS_GERROR_REMOTE_EXCEPTION)
            {
                g_printerr ("Caught remote method exception %s: %s", dbus_g_error_get_name (error), error->message);
            }
            else
            {
                g_printerr ("Error: %s\n", error->message);
            }
            g_error_free (error);
            success = FALSE;
        }
    }

    if (success)
    {
        if (strcmp (PluginParams.PluginType, ACM_TEST_MOST_BUS_NAME) == 0)
        {
            sPopulateChannelDisconnectParams (sMostChannelParams, AllocParamIndex, ConnectionIndex);
        }
        else if (strcmp (PluginParams.PluginType, ACM_TEST_PCM_BUS_NAME) == 0)
        {
            sPopulateChannelDisconnectParams (sPcmChannelParams, AllocParamIndex, ConnectionIndex);
        }
        else if (strcmp (PluginParams.PluginType, ACM_TEST_LOCAL_BUS_NAME) == 0)
        {
            sPopulateChannelDisconnectParams (sLocalChannelParams, AllocParamIndex, ConnectionIndex);
        }
        g_printf ("Disconnect Success\n");
        sPrintAllocs (argv[1]);
    }
    else
    {
        g_printf ("Disconnect Failure\n");
    }

    g_hash_table_unref (disConnectBusData);
    g_hash_table_unref (disconnectReturnTable);
    return success;
}


static gboolean sDealloc
(
    int argc,    /**< number of arguments including application name */
    char * *argv /**< array of string arguments */
)
{
    GHashTable *deallocBusData;
    GHashTable *deallocReturnTable;
    ABM_AudioPluginParams_t PluginParams;
    ABM_AudioPluginParams_t *pCommonPluginParams = NULL;
    GError *error;
    gboolean success = TRUE;
    int AllocParamIndex;
    gchar Tempbuffer[ACM_TEST_MOST_AUDIO_PARAM_TEMP_BUF_SIZE];

    gchar ClientTypeName[ACM_TEST_PARAM_STRING_LENGTH];
    gchar PluginTypeName[ACM_TEST_PARAM_STRING_LENGTH];

    /* Set up the tables to handle the input and return data. */
    deallocBusData = g_hash_table_new (g_str_hash, g_str_equal);
    deallocReturnTable = g_hash_table_new (g_str_hash, g_str_equal);
    success = sCheckInputsDealloc (argc, argv, &PluginParams);

    if (success)
    {
        AllocParamIndex = atoi (argv[2]);
        if (AllocParamIndex >= ACM_TEST_ALLOC_MAX_COUNT)
        {
            g_printf ("Error: Invalid Allocation Index\n");
            success = FALSE;
        }
        if (AllocParamIndex >= ACM_TEST_CHANNEL_CLIENT_COUNT)
        {
            g_printf ("Error: Invalid connection Index\n");
            success = FALSE;
        }
        if (success)
        {
            if (strcmp (PluginParams.PluginType, ACM_TEST_MOST_BUS_NAME) == 0)
            {
                if (sMostChannelParams[AllocParamIndex].IsAllocated == FALSE)
                {
                    g_printf ("Error: Index not Allocated\n");
                    success = FALSE;
                }
                else
                {
                    g_sprintf (Tempbuffer, "%d", sMostChannelParams[AllocParamIndex].SyncHandle);
                    g_hash_table_insert (deallocBusData, ACM_TEST_MOST_AUDIO_PARAM_TABLE_KEY_DEVICE_HANDLE, Tempbuffer);
                    pCommonPluginParams = &sMostChannelParams[AllocParamIndex].AllocatorParam;
                }
            }
            else if (strcmp (PluginParams.PluginType, ACM_TEST_PCM_BUS_NAME) == 0)
            {
                if (sPcmChannelParams[AllocParamIndex].IsAllocated == FALSE)
                {
                    g_printf ("Error: Index not Allocated\n");
                    success = FALSE;
                }
                else
                {
                    g_hash_table_insert (deallocBusData, ACM_TEST_PCM_AUDIO_PARAM_TABLE_KEY_CHANNEL_ID,
                                         sPcmChannelParams[AllocParamIndex].ChannelParams[0].ChannelID);
                    pCommonPluginParams = &sPcmChannelParams[AllocParamIndex].AllocatorParam;
                }
            }
            else if (strcmp (PluginParams.PluginType, ACM_TEST_LOCAL_BUS_NAME) == 0)
            {
                if (sLocalChannelParams[AllocParamIndex].IsAllocated == FALSE)
                {
                    g_printf ("Error: Index not Allocated\n");
                    success = FALSE;
                }
                else
                {
                    g_hash_table_insert (deallocBusData, ACM_TEST_PCM_AUDIO_PARAM_TABLE_KEY_CHANNEL_ID,
                                         sLocalChannelParams[AllocParamIndex].ChannelParams[0].ChannelID);
                    pCommonPluginParams = &sLocalChannelParams[AllocParamIndex].AllocatorParam;
                }
            }
        }
    }

    if (success)
    {
        if (pCommonPluginParams)
        {
            strcpy (ClientTypeName, pCommonPluginParams->ClientType);
            strcpy (PluginTypeName, pCommonPluginParams->PluginType);
        }
        /* Call deallocate */
        error = NULL;
        if (!dbus_g_proxy_call (pACM_Proxy,                          "DeAllocate",                  &error,
                                G_TYPE_STRING,                       pCommonPluginParams->DeviceID, G_TYPE_STRING,
                                PluginTypeName,                      G_TYPE_STRING,                 ClientTypeName,
                                DBUS_TYPE_G_STRING_STRING_HASHTABLE, deallocBusData,                G_TYPE_INVALID,
                                DBUS_TYPE_G_STRING_STRING_HASHTABLE, &deallocReturnTable,           G_TYPE_INVALID))
        {
            /* Just do demonstrate remote exceptions versus regular GError */
            if (error->domain == DBUS_GERROR && error->code == DBUS_GERROR_REMOTE_EXCEPTION)
            {
                g_printerr ("Caught remote method exception %s: %s", dbus_g_error_get_name (error), error->message);
            }
            else
            {
                g_printerr ("Error: %s\n", error->message);
            }
            g_error_free (error);
            success = FALSE;
        }
    }

    if (success)
    {
        if (strcmp (PluginParams.PluginType, ACM_TEST_MOST_BUS_NAME) == 0)
        {
            sPopulateChannelDeallocParams (sMostChannelParams, AllocParamIndex);
        }
        else if (strcmp (PluginParams.PluginType, ACM_TEST_PCM_BUS_NAME) == 0)
        {
            sPopulateChannelDeallocParams (sPcmChannelParams, AllocParamIndex);
        }
        else if (strcmp (PluginParams.PluginType, ACM_TEST_LOCAL_BUS_NAME) == 0)
        {
            sPopulateChannelDeallocParams (sLocalChannelParams, AllocParamIndex);
        }
        g_printf ("Dealloc Success\n");
        sPrintAllocs (argv[1]);
    }
    else
    {
        g_printf ("Dealloc Failure\n");
    }

    g_hash_table_unref (deallocBusData);
    g_hash_table_unref (deallocReturnTable);
    return success;
}


static gboolean sACM_dbusConnect (
)
{
    GError *error;
    error = NULL;
    pDbusConnection = dbus_g_bus_get (DBUS_BUS_SYSTEM, &error);

    if (pDbusConnection == NULL)
    {
        g_printerr ("Failed to open connection to bus: %s\n", error->message);
        g_error_free (error);
        return FALSE;
    }
    /* Create a proxy object for the ACM*/
    pACM_Proxy = dbus_g_proxy_new_for_name (pDbusConnection, "com.windriver.audio", "/com/windriver/audio",
                                            "com.windriver.audio.ConnectionManager");

    if (pACM_Proxy == NULL)
    {
        g_printerr ("Failed to create dbus proxy to ACM: %s\n", error->message);
        g_error_free (error);
        return FALSE;
    }
    /* Create a proxy object for the ACM CC*/
    pACMCC_Proxy = dbus_g_proxy_new_for_name (pDbusConnection, "com.windriver.audio", "/com/windriver/audio",
                                              "com.windriver.audio.ConnectionControl");

    if (pACMCC_Proxy == NULL)
    {
        g_printerr ("Failed to create dbus proxy to ACM CC: %s\n", error->message);
        g_error_free (error);
        return FALSE;
    }

    return TRUE;
}


static gboolean sConnectionControl
(
    gchar *DeviceName, gchar *Operation
)
{
    GError *error;
    gboolean success = TRUE;

    error = NULL;
    success = dbus_g_proxy_call (pACMCC_Proxy, Operation, &error, G_TYPE_STRING, DeviceName, G_TYPE_INVALID,
                                 G_TYPE_INVALID);

    if (!success)
    {
        /* Just do demonstrate remote exceptions versus regular GError */
        if (error->domain == DBUS_GERROR && error->code == DBUS_GERROR_REMOTE_EXCEPTION)
        {
            g_printerr ("Caught remote method exception %s: %s", dbus_g_error_get_name (error), error->message);
        }
        else
        {
            g_printerr ("Error: %s\n", error->message);
        }
        g_error_free (error);
        success = FALSE;
    }

    return success;
}


static gboolean sStart
(
    int argc,    /**< number of arguments including application name */
    char * *argv /**< array of string arguments */
)
{
    gchar DeviceName[ABM_MAX_DEVICE_ID_LENGTH];
    gboolean success = TRUE;

    if (argc < 2)
    {
        g_printf ("Error: Too few arguments for start!\n");
        return FALSE;
    }

    strcpy (DeviceName, argv[1]);
    success = sConnectionControl (DeviceName, "Start");
    return success;
}


static gboolean sStop
(
    int argc,    /**< number of arguments including application name */
    char * *argv /**< array of string arguments */
)
{
    gchar DeviceName[ABM_MAX_DEVICE_ID_LENGTH];
    gboolean success = TRUE;

    if (argc < 2)
    {
        g_printf ("Error: Too few arguments for stop!\n");
        return FALSE;
    }

    strcpy (DeviceName, argv[1]);
    success = sConnectionControl (DeviceName, "Stop");
    return success;
}


static gboolean sMute
(
    int argc,    /**< number of arguments including application name */
    char * *argv /**< array of string arguments */
)
{
    gchar DeviceName[ABM_MAX_DEVICE_ID_LENGTH];
    gboolean success = TRUE;

    if (argc < 2)
    {
        g_printf ("Error: Too few arguments for Mute!\n");
        return FALSE;
    }

    strcpy (DeviceName, argv[1]);
    success = sConnectionControl (DeviceName, "Mute");
    return success;
}


static gboolean sUnMute
(
    int argc,    /**< number of arguments including application name */
    char * *argv /**< array of string arguments */
)
{
    gchar DeviceName[ABM_MAX_DEVICE_ID_LENGTH];
    gboolean success = TRUE;

    if (argc < 2)
    {
        g_printf ("Error: Too few arguments for UnMute!\n");
        return FALSE;
    }

    strcpy (DeviceName, argv[1]);
    success = sConnectionControl (DeviceName, "UnMute");
    return success;
}


static gboolean sPause
(
    int argc,    /**< number of arguments including application name */
    char * *argv /**< array of string arguments */
)
{
    gchar DeviceName[ABM_MAX_DEVICE_ID_LENGTH];
    gboolean success = TRUE;

    if (argc < 2)
    {
        g_printf ("Error: Too few arguments for Pause!\n");
        return FALSE;
    }

    strcpy (DeviceName, argv[1]);
    success = sConnectionControl (DeviceName, "Pause");
    return success;
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: main                                                   */
/*                                                                        */
/**************************************************************************/
/**
 *  Main entry
 *
 *  \return None
 **************************************************************************
*/
int main
(
    int argc,    /**< number of arguments including application name */
    char * *argv /**< array of string arguments */
)
{
    GOptionContext *context;
    GError *pGError = NULL;
    char *pCmdLine;
    gint pCmdLineArgc;
    gchar * *pCmdLineArgv;

    g_type_init ();

    if (!sACM_dbusConnect ())
    {
        g_printf ("Dbus Connection to ACM failed exitin\n");
        exit (1);
    }

    InitChannelParams (sPcmChannelParams);
    InitChannelParams (sLocalChannelParams);
    InitChannelParams (sMostChannelParams);

    if (!g_thread_supported ())
    {
        g_thread_init (NULL);
    }

    {
    GLogLevelFlags fatal_mask;

    fatal_mask = g_log_set_always_fatal (G_LOG_FATAL_MASK);
    fatal_mask |= G_LOG_LEVEL_WARNING | G_LOG_LEVEL_CRITICAL;
    g_log_set_always_fatal (fatal_mask);
    }

    context = g_option_context_new ("");
    g_option_context_add_main_entries (context, entries, NULL);
    g_option_context_set_summary (context, ACM_TEST_APP_DECRIPTION);
    g_option_context_set_description (context, "This is a description");

    if (!g_option_context_parse (context, &argc, &argv, &pGError))
    {
        g_printf ("%s\n", pGError->message);
        exit (1);
    }

    if (sPrintVersion)
    {
        g_printf ("Package: '%s'\n", ACM_TEST_APP_VERSION_STRING);
        exit (1);
    }
    /* catch Ctrl-C and Hangup signals for clean shutdown */
    (void)signal (SIGINT, sSignalCatch);
    (void)signal (SIGHUP, sSignalCatch);
    (void)signal (SIGTERM, sSignalCatch);

    setNonblocking (STDIN_FILENO);

    sUsage ();
    g_printf ("\nACM Command:\n");

    gExit = FALSE;

    while (FALSE == gExit)
    {
        if (TRUE == sProcessStdIn (&pCmdLine))
        {
            /* '/n' char detected. Go parsing command: */
            if (FALSE == g_shell_parse_argv (pCmdLine, &pCmdLineArgc, &pCmdLineArgv, &pGError))
            {
                //g_printf("Error parsing '%s'. Err code: %d, %s\n", pCmdLine, (*pGError).code, (*pGError).message);
                g_clear_error (&pGError);
                sUsage ();
                continue;
            }
            if (0 == strcasecmp ("alloc", pCmdLineArgv[0]))
            {
                sAlloc (pCmdLineArgc, pCmdLineArgv);
            }
            else if (0 == strcasecmp ("dealloc", pCmdLineArgv[0]))
            {
                sDealloc (pCmdLineArgc, pCmdLineArgv);
            }
            else if (0 == strcasecmp ("iconnect", pCmdLineArgv[0]))
            {
                sConnect (pCmdLineArgc, pCmdLineArgv, TRUE);
            }
            else if (0 == strcasecmp ("connect", pCmdLineArgv[0]))
            {
                sConnect (pCmdLineArgc, pCmdLineArgv, FALSE);
            }
            else if (0 == strcasecmp ("disconnect", pCmdLineArgv[0]))
            {
                sDisconnect (pCmdLineArgc, pCmdLineArgv);
            }
            else if (0 == strcasecmp ("start", pCmdLineArgv[0]))
            {
                sStart (pCmdLineArgc, pCmdLineArgv);
            }
            else if (0 == strcasecmp ("stop", pCmdLineArgv[0]))
            {
                sStop (pCmdLineArgc, pCmdLineArgv);
            }
            else if (0 == strcasecmp ("mute", pCmdLineArgv[0]))
            {
                sMute (pCmdLineArgc, pCmdLineArgv);
            }
            else if (0 == strcasecmp ("unmute", pCmdLineArgv[0]))
            {
                sUnMute (pCmdLineArgc, pCmdLineArgv);
            }
            else if (0 == strcasecmp ("pause", pCmdLineArgv[0]))
            {
                sPause (pCmdLineArgc, pCmdLineArgv);
            }
            else if (0 == strcasecmp ("list-allocs", pCmdLineArgv[0]))
            {
                if (2 == pCmdLineArgc)
                {
                    sPrintAllocs (pCmdLineArgv[1]);
                }
                if (1 == pCmdLineArgc)
                {
                    sPrintAllocs ("all");
                }
            }
            else if (0 == strcasecmp ("help", pCmdLineArgv[0]))
            {
                sUsage ();
            }
            else
            {
                g_printf ("Error: Unknown command!\n");

            //sUsage();
            }
            g_printf ("\nACM Command:\n");
        }
        usleep (10000);
    }

    exit (0);
}
/// @endcond
