/*
 * main.c
 *
 * Copyright (c) 2008 Wind River Systems, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Lesser GNU General Public License version 2.1 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the Lesser GNU General Public License for more details.
 *
 * You should have received a copy of the Lesser GNU General Public License version 2.1
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */


/// @cond internal_documentation
/**
 * \file 
 *    Main loop for handling DBus events.    
 *
 * \version $Revision: 1.3 $
 * \date    $Date: 2008/07/12 01:27:19 $
 */

/**************************************************************************/
/*                              Includes                                  */
/**************************************************************************/
#include <stdlib.h>
#include <signal.h>
#include <glib.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-bindings.h>
#include "devicecontrol.h"
#include "acmd.h"
#include <string.h>

/**************************************************************************/
/*                               Defines                                  */
/**************************************************************************/

/**************************************************************************/
/*                              Typedefs                                  */
/**************************************************************************/

/**************************************************************************/
/*                              Externs                                   */
/**************************************************************************/

extern size_t strnlen (__const char *__string, size_t __maxlen)
__THROW __attribute_pure__ __nonnull ((1));

/**************************************************************************/
/*                              Globals                                   */
/**************************************************************************/
/** Connection to system Dbus */
DBusGConnection *dbus_connection = NULL;

/** Tracks if the manager has been initialized yet. */
gboolean dc_initialized = FALSE;

/** The base DBus service name for the power state manager.*/
char DC_DBUS_SERVICE[256];

/** The DBus object path for the power state manager. */
char DC_OBJECT_PATH[256];

/**************************************************************************/
/*                           Local Variables                              */
/**************************************************************************/

/** Main loop for handling incoming DBus events. */
static GMainLoop *main_loop = NULL;

/** object to handle DBus calls from applications */
static DeviceControl *dc_object = NULL;

/**************************************************************************/
/*  Private Functions Prototypes-functions used internally in this file   */
/**************************************************************************/

/**************************************************************************/
/*                       Public Functions Bodies                         */
/**************************************************************************/

/**************************************************************************/
/*                       Private Functions Bodies                         */
/**************************************************************************/

/**************************************************************************/
/*                                                                        */
/*  Function Name: terminate                                              */
/*                                                                        */
/**************************************************************************/
/**
 *  Cleans up the dbus connection.               
 *
 *  \return TRUE if successful.                 
 *                                                     
 **************************************************************************
 */
gboolean terminate (
)
{
    g_debug (
        G_STRLOC " Shutting down DC.");

    if (dc_object)
    {
        g_object_unref (dc_object);
    }

    if (dbus_connection)
    {
        dbus_g_connection_unref (dbus_connection);
        dbus_connection = NULL;
    }

    return TRUE;
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: initialize                                             */
/*                                                                        */
/**************************************************************************/
/**
 *  Initializes the Bluetooth Daemon.               
 *
 *  \return TRUE if successful.                 
 *                                                     
 **************************************************************************
 */
gboolean initialize
(
    void
)
{
    gboolean success = TRUE;
    GError *error = NULL;
    DBusGProxy *proxy;
    guint request_ret;

    g_type_init ();

    /* If thread support is needed, this should be uncommented. */
    /*if ( !g_thread_supported() )
        g_thread_init(NULL);

    dbus_g_thread_init();
    */

    if (success == TRUE)
    {
        if (error != NULL)
        {
            g_error_free (error);
            error = NULL;
        }
        dbus_connection = dbus_g_bus_get (DBUS_BUS_SYSTEM, &error);
        if (!dbus_connection)
        {
            g_critical (
                G_STRLOC " Failed to open connection to bus: %s\n", error->message);
            g_error_free (error);
            success = FALSE;
        }
    }

    if (success)
    {
        /* Get a proxy to DBus. */
        proxy = dbus_g_proxy_new_for_name (dbus_connection, DBUS_SERVICE_DBUS, DBUS_PATH_DBUS, DBUS_INTERFACE_DBUS);
        if (error != NULL)
        {
            g_error_free (error);
            error = NULL;
        }
        /* Register a well known service name */
        if (!org_freedesktop_DBus_request_name (proxy, DC_DBUS_SERVICE, 0, &request_ret, &error))
        {
            g_critical (
                G_STRLOC " Unable to register service: %s\n", error->message);
            g_error_free (error);
            success = FALSE;
        }
        else if (request_ret != DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER)
        {
            g_critical (
                G_STRLOC " Unable to register DBus name %s - "
                             "dc already running?\n", DC_DBUS_SERVICE);
            success = FALSE;
        }
        g_object_unref (proxy);
    }

    if (success)
    {
        // Set up the manager object.
        dc_object = (DeviceControl *)g_object_new (devicecontrol_get_type (), NULL);
        if (!dc_initialized)
        {
            g_critical (
                G_STRLOC " Failed to initialise server object\n");
            success = FALSE;
        }
    }

    if (!success)
    {
        terminate ();
    }

    return success;
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: sighandler                                             */
/*                                                                        */
/**************************************************************************/
/**
 *  Handles incoming signals by terminating the main loop.               
 *               
 *                                                     
 **************************************************************************
 */
void sighandler
(
    int sig /**< [in] signal number.*/
)
{
    if (main_loop)
    {
        g_main_loop_quit (main_loop);
    }
}
/**************************************************************************/
/*                                                                        */
/*  Function Name: main                                                   */
/*                                                                        */
/**************************************************************************/
/**
 *  Entry point for the bluetooth daemon.               
 *
 *  \return 0 if successful, -1 if error.                 
 *                                                     
 **************************************************************************
 */
int main
(
    int argc,    /**< [in] Number of command line args*/
    char * *argv /**< [in] Command line args.*/
)
{
    extern char * *environ;
    gboolean success = TRUE;
    char * *current_env_variable;
    char *match_string = "PULSE_PROP_" ACM_DEVICE_CONTROL_ENVIRONMENT_KEYWORD "=";
    int match_string_len = strlen (match_string);
    gboolean match_found = FALSE;
    int i, ServiceLen;

    signal (SIGTERM, sighandler);
    signal (SIGINT, sighandler);

    for (current_env_variable = environ; *current_env_variable; current_env_variable++)
    {
        if (0 == strncmp (*current_env_variable, match_string, match_string_len))
        {
            strncpy (DC_DBUS_SERVICE, *current_env_variable + match_string_len, 250);
            match_found = TRUE;
            break;
        }
    }

    if (!match_found)
    {
        g_message ("Please set the environment variable:\n %syour dbus control module name\n", match_string);
        exit (1);
    }
    /* make a copy of the deviceID into pathName, substituting "/" for "." as we go */
    ServiceLen = strnlen (DC_DBUS_SERVICE, 250);
    i = 0;
    DC_OBJECT_PATH[0] = '/';
    DC_OBJECT_PATH[ServiceLen + 1] = '\0';

    while (i < ServiceLen)
    {
        if ('\0' == DC_DBUS_SERVICE[i])
        {
            DC_OBJECT_PATH[i + 1] = '\0';
            break;
        }
        else if ('.' == DC_DBUS_SERVICE[i])
        {
            DC_OBJECT_PATH[i + 1] = '/';
        }
        else
        {
            DC_OBJECT_PATH[i + 1] = DC_DBUS_SERVICE[i];
        }
        i++;
    }

    g_message ("DBus Service Name = \"%s\"\n", DC_DBUS_SERVICE);
    g_message ("DBus Object Path = \"%s\"\n", DC_OBJECT_PATH);

    success = initialize ();

    // Start the event loop.
    if (success)
    {
        g_debug ("Starting main loop.");
        main_loop = g_main_loop_new (NULL, FALSE);
        g_main_loop_run (main_loop);
    }

    terminate ();

    return (success ? 0 : -1);
}
/// @endcond
