/*
 * paclient.h 
 *
 * Copyright (c) 2008 Wind River Systems, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Lesser GNU General Public License version 2.1 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the Lesser GNU General Public License for more details.
 *
 * You should have received a copy of the Lesser GNU General Public License version 2.1
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */
/**
 * \defgroup pulse_audio_routing PulseAudio Utility Libraries
 *
 * There are two PulseAudio related utility libraries:
 * \li \ref paclient.h  A synchronous wrapper around the PulseClient API. 
 * \li \ref parouting.h  A module to connect to the local PulseAudio server and set up routing policies.
 */
/**
 * \ingroup pulse_audio_routing
 * \file
 *    PulseAudio Client Library.  This library is a wrapper around the
 *    PulseClient API.  It is mainly implemented as a synchronous API,
 *    with some callbacks (such as EventSubscribe callbacks) from PulseAudio 
 *    remaining asynchronous.
 *
 *
 * \version $Revision: 1.13 $
 * \date    $Date: 2008/07/12 00:56:52 $
 */
#ifndef _PACLIENT_H
#define _PACLIENT_H

/**************************************************************************/
/*                              Includes                                  */
/**************************************************************************/
#include <pulse/pulseaudio.h>

/**************************************************************************/
/*                               Defines                                  */
/**************************************************************************/

/**************************************************************************/
/*                              Typedefs                                  */
/**************************************************************************/

/** PA Client Error Codes **/

typedef enum
{
    PA_CLIENT_OK = 0,
    PA_CLIENT_NO_MATCHING_DEVICE,
    PA_CLIENT_NO_MATCHING_STREAM,
    PA_CLIENT_ERR,
    PA_CLIENT_TERMINATED,
    PA_CLIENT_CONNECT_FAILED
} PA_Client_Err_t;

/** PA Client Device Type **/
typedef enum
{
    PA_CLIENT_DEVICE_SINK,
    PA_CLIENT_DEVICE_SOURCE
} PA_CLIENT_DEVICE_TYPE_t;

/** PA Client Application Stream Direction */
typedef enum
{
    PA_CLIENT_SINK_INPUT,
    PA_CLIENT_SOURCE_OUTPUT,
} PA_CLIENT_STREAM_TYPE_t;

/** Client Library Context **/
typedef struct
{
    pa_context *pPACtx;
    pa_threaded_mainloop *pMainLoop;
} PA_Client_Ctx_t;

/**************************************************************************/
/*                          Functions Declarations                        */
/**************************************************************************/

/**************************************************************************/
/*                                                                        */
/*  Function Name: PAClientCtxInit                                        */
/*                                                                        */
/**************************************************************************/
/**
 * Called to initalize a PA_Client_Ctx_t structure provided by caller prior
 * to calling PAClientCtxConnect.
 *
 * \return PA_CLIENT_OK on success, PA_CLIENT_ERR               
 *
 * \see
 * PAClientCtxConnect
 *                                                     
 **************************************************************************
 */
PA_Client_Err_t PAClientCtxInit
(
    PA_Client_Ctx_t *pCtx, /**< [in] Pointer to unintialized Client Context struct */
    const char *clientName  /**< [in] Pointer to string to be copied to PulseAudio  */
);

/**************************************************************************/
/*                                                                        */
/*  Function Name: PAClientCtxConnect                                     */
/*                                                                        */
/**************************************************************************/
/**
 *
 * Connects an already initalize PA_Client_Ctx_t structure to the PulseAudio 
 * server.
 * 
 * \return PA_CLIENT_OK on success, PA_CLIENT_ERR, PA_CLIENT_TERMINATED 
 * or PA_CLIENT_CONNECT_FAILED on failure              
 *                                                     
 **************************************************************************
 */
PA_Client_Err_t PAClientCtxConnect
(
    PA_Client_Ctx_t *pCtx /**< [in] Pointer to intialized Client Context struct */
);

/**************************************************************************/
/*                                                                        */
/*  Function Name: PAClientCtxCheckConnection                             */
/*                                                                        */
/**************************************************************************/
/**
 *
 *  Check the state of the PulseAudio connection
 *
 *  \return PA_CLIENT_OK if server connection is ok, otherwise
 * PulseAudio connection needs to be disconnected.
 *                                                     
 **************************************************************************
 */
PA_Client_Err_t PAClientCheckConnection
(
    PA_Client_Ctx_t *pCtx /**< [in] Pointer to Client Context struct */
);

/**************************************************************************/
/*                                                                        */
/*  Function Name: PAClientCtxCleanup                                     */
/*                                                                        */
/**************************************************************************/
/**
 *
 * Disconnects a PA_Client_Ctx_t structure from the PulseAudio 
 * server and releases PulseAudio resources.  It does not free the PA_Client_Ctx_t
 * struct.
 * 
 * \return PA_CLIENT_OK on success
 *                                                     
 **************************************************************************
 */
PA_Client_Err_t PAClientCtxCleanup
(
    PA_Client_Ctx_t *pCtx /**< [in] Pointer to Client Context struct */
);

/**************************************************************************/
/*                                                                        */
/*  Function Name: PAClientCtxLoadModule                                  */
/*                                                                        */
/**************************************************************************/
/**
 * Load a module in PulseAudio.   This function is a synchronous wrapper 
 * around pa_context_load_module.  Returns when the result of the load module
 * operation is complete.
 * 
 * 
 * \return PA_CLIENT_OK on success
 *                                                     
 **************************************************************************
 */
PA_Client_Err_t PAClientCtxLoadModule
(
    PA_Client_Ctx_t *pCtx, /**< [in] Pointer to Client Context struct */
    const char *moduleName, /**< [in] Name of PulseAudio module to load */
    const char *argument,   /**< [in] Module arguments */
    uint32_t *moduleIndex /**< [out] Pointer to update with loaded module index 
                        on success */
);

/**************************************************************************/
/*                                                                        */
/*  Function Name: PAClientCtxUnloadModule                                  */
/*                                                                        */
/**************************************************************************/
/**
 * Load a module in PulseAudio.   This function is a synchronous wrapper 
 * around pa_context_unload_module.  Returns when the result of the unload module
 * operation is complete.
 * 
 * 
 * \return PA_CLIENT_OK on success
 *                                                     
 **************************************************************************
 */

PA_Client_Err_t PAClientCtxUnloadModule
(
    PA_Client_Ctx_t *pCtx, /**< [in] Pointer to Client Context struct */
    uint32_t ModuleIndex   /**< [in] Index of Module to unload */
);

/**************************************************************************/
/*                                                                        */
/*  Function Name: PAClientGetDevFromModule                               */
/*                                                                        */
/**************************************************************************/
/**
 * Query the PulseAudio server with a module index to get the associated
 * device (sink or source) index.
 * 
 * 
 * \return PA_CLIENT_OK on success
 *                                                     
 **************************************************************************
 */
PA_Client_Err_t PAClientGetDevFromModule
(
    PA_Client_Ctx_t *pCtx,           /**< [in] Pointer to Client Context struct */
    const uint32_t moduleIndex,      /**< [in] Index of Module */
    PA_CLIENT_DEVICE_TYPE_t devType, /**< [in] Type of device that is connected to module */
    uint32_t *deviceIndex            /**< [out] Location that result is copied to on success */
);

/**************************************************************************/
/*                                                                        */
/*  Function Name: PAClientGetStreamIndexByProperty                       */
/*                                                                        */
/**************************************************************************/
/**
 * Query the PulseAudio server with a Pulse Property to find a matching
 * stream index.  For example if searching for PULSE_PROP_xyz="abc", xyz is the
 * property key, and "abc" would be the property value.
 * 
 * 
 * \return PA_CLIENT_OK on success, PA_CLIENT_NO_MATCHING_STREAM if not found
 *                                                     
 **************************************************************************
 */
PA_Client_Err_t PAClientGetStreamIndexByProperty
(
    PA_Client_Ctx_t *pCtx,              /**< [in] Pointer to Client Context struct */
    const char *propKey,                 /**< [in] Pulse Property to look up */
    const char *propValue,               /**< [in] Property Value to look for */
    PA_CLIENT_STREAM_TYPE_t streamType, /**< [in] Stream type to search for */
    uint32_t *streamIndex               /**< [out] Updated with stream index if found */
);

/**************************************************************************/
/*                                                                        */
/*  Function Name: PAClientMoveStream                                     */
/*                                                                        */
/**************************************************************************/
/**
 * Instruct the PulseAudio Server to move a stream to a different device
 * 
 * 
 * \return PA_CLIENT_OK on success
 *                                                     
 **************************************************************************
 */
PA_Client_Err_t PAClientMoveStream
(
    PA_Client_Ctx_t *pCtx,              /**< [in] Pointer to Client Context struct */
    PA_CLIENT_STREAM_TYPE_t streamType, /**< [in] Type of stream to move   */
    const uint32_t streamIndex,         /**< [in] Index of stream to move */
    const uint32_t devIndex             /**< [in] Index of device to move to */
);

/**************************************************************************/
/*                                                                        */
/*  Function Name: PAClientMoveAllStreamsByProperty                       */
/*                                                                        */
/**************************************************************************/
/**
 * Instruct the PulseAudio Server to move all streams with a matching
 * stream id property to a device
 * 
 * 
 * \return PA_CLIENT_OK on success
 *                                                     
 **************************************************************************
 */
PA_Client_Err_t PAClientMoveAllStreamsByProperty
(
    PA_Client_Ctx_t *pCtx,              /**< [in] Pointer to Client Context struct */
    PA_CLIENT_STREAM_TYPE_t streamType, /**< [in] Type of streams to move   */
    const char *pPropertyKey,            /**< [in] PulseProperty to use as the stream id */
    const char *pStreamId,               /**< [in] Id of streams to move */
    const uint32_t devIndex             /**< [in] Index of device to move to */
);

/**************************************************************************/
/*                                                                        */
/*  Function Name: PAClientEventSubscribe                                 */
/*                                                                        */
/**************************************************************************/
/**
 * Subscribe to Events from the PulseAudio Server.  This function is a 
 * synchronous wrapper around pa_context_subscribe.  See PulseAudio
 * documentation for details.
 * 
 * 
 * \return PA_CLIENT_OK on success
 *                                                     
 **************************************************************************
 */
PA_Client_Err_t PAClientEventSubscribe
(
    PA_Client_Ctx_t *pCtx,              /**< [in] Pointer to Client Context struct */
    pa_context_subscribe_cb_t callback, /**< [in] Event callback to be passed to PulseAudio */
    void *cbData,                       /**< [in] Userdata to be passed to the callback */
    pa_subscription_mask_t eventMask    /**< [in] PulseAudio Event Mask                */
);
#endif                                  //_PACLIENT_H
