/*
 * parouting.c 
 *
 * Copyright (c) 2008 Wind River Systems, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Lesser GNU General Public License version 2.1 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the Lesser GNU General Public License for more details.
 *
 * You should have received a copy of the Lesser GNU General Public License version 2.1
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */
/*
 * \file 
 *    PulseAudio Routing Module.    
 *
 * \version $Revision: 1.11 $
 * \date    $Date: 2008/07/12 00:56:52 $
 */

/**************************************************************************/
/*                              Includes                                  */
/**************************************************************************/
#include <signal.h>
#include <string.h>

#include <stdio.h>
#include <stdlib.h>
#include "paclient.h"
#include "parouting.h"

/**************************************************************************/
/*                               Defines                                  */
/**************************************************************************/
#define CONTEXT_OPERATION_CHECK_DEAD(ctx,operation,cbData)                    \
    if(!ctx->pPACtx || pa_context_get_state(ctx->pPACtx) != PA_CONTEXT_READY) \
    {                                                                         \
        fprintf(stderr,"pulseaudio context died\n");                          \
        if(operation)                                                         \
        {                                                                     \
            pa_operation_unref(operation);                                    \
        }                                                                     \
        pa_threaded_mainloop_unlock(ctx->pMainLoop);                          \
        if(cbData)                                                            \
        {                                                                     \
            g_free(cbData);                                                   \
        }                                                                     \
        return PA_ROUTING_ERR;                                                \
    }


#if 1
#define PAROUTING_DEBUG_LOG(fmt,arg...)
#else
#define PAROUTING_DEBUG_LOG printf
#endif
#define PA_DONT_BLOCK 0

/**************************************************************************/
/*                              Typedefs                                  */
/**************************************************************************/
typedef struct
{
    gchar *pStreamProp; /* A route is a stream (property) to a device (sink/src) */
    uint32_t devIndex;
} PA_Route_List_Element_t;


typedef struct ApplyRouteCBData
{
    PA_Routing_Ctx_t *pRouteCtx;
    pa_threaded_mainloop *pMainLoop;
    gchar *pRouteStreamId; /* property used to find streams to apply route to */
    uint32_t devIndex;     /* device to route stream to */
} PARouteApplyRouteCBData_t;

/**************************************************************************/
/*                              Externs                                   */
/**************************************************************************/

/**************************************************************************/
/*                              Globals                                   */
/**************************************************************************/

/**************************************************************************/
/*                           Local Variables                              */
/**************************************************************************/

/**************************************************************************/
/*  Private Functions Prototypes-functions used internally in this file   */
/**************************************************************************/
static void PARoutingSinkInputInfoCallback
(
    pa_context *c, const pa_sink_input_info *i, int eol, void *userdata
);


static void PARoutingApplySinkInputInfoCallback
(
    pa_context *c, const pa_sink_input_info *i, int eol, void *userdata
);


static void PARoutingApplySourceOutputInfoCallback
(
    pa_context *c, const pa_source_output_info *i, int eol, void *userdata
);


static void PARoutingSourceOutputInfoCallback
(
    pa_context *c, const pa_source_output_info *i, int eol, void *userdata
);


static void PARoutingEventCallback
(
    pa_context *c, pa_subscription_event_type_t t, uint32_t index, void *userdata
);


static void PARoutingSuccessCallback
(
    pa_context *c, int success, void *userdata
);


static int PARoutingStringsEqual
(
    const char *str1, const char *str2, size_t length
);

/*************************************************************************/
/*                       Public Functions Bodies                         */
/**************************************************************************/

PA_Routing_Ctx_t *PARoutingModuleNew
(
    const char *pPropertyKey /**< [in] String to use as the Stream Property Key */
)
{
    PA_Routing_Ctx_t *pRoutingCtx;

    if (!pPropertyKey)
    {
        PAROUTING_DEBUG_LOG ("PARoutingModuleNew Fails: NULL pPropertyKey\n");
        return NULL;
    }

    pRoutingCtx = (PA_Routing_Ctx_t *)malloc (sizeof (PA_Routing_Ctx_t));

    if (!pRoutingCtx)
    {
        PAROUTING_DEBUG_LOG ("PARoutingModuleNew malloc fails\n");
        return NULL;
    }

    bzero (pRoutingCtx, sizeof (PA_Routing_Ctx_t));

    pRoutingCtx->pPropertyKey = (char *)malloc (sizeof (char) * (strlen (pPropertyKey) + 1));

    if (!(pRoutingCtx->pPropertyKey))
    {
        free (pRoutingCtx);
        PAROUTING_DEBUG_LOG ("PARoutingModuleNew malloc fails\n");
        return NULL;
    }

    strcpy (pRoutingCtx->pPropertyKey, pPropertyKey);

    PAROUTING_DEBUG_LOG ("stream property key is %s\n", pRoutingCtx->pPropertyKey);

    if (!g_thread_supported ())
    {
        g_thread_init (NULL);
    }

    pRoutingCtx->pRouteListMutex = g_mutex_new ();

    if (pRoutingCtx->pRouteListMutex == NULL)
    {
        PAROUTING_DEBUG_LOG ("g_mutex_new fails\n");
        free (pRoutingCtx->pPropertyKey);
        free (pRoutingCtx);
        return NULL;
    }

    pRoutingCtx->pClientCtx = (PA_Client_Ctx_t *)malloc (sizeof (PA_Client_Ctx_t *));

    if (pRoutingCtx->pClientCtx == NULL)
    {
        PAROUTING_DEBUG_LOG ("malloc fails\n");
        g_mutex_free (pRoutingCtx->pRouteListMutex);
        free (pRoutingCtx->pPropertyKey);
        free (pRoutingCtx);
        return NULL;
    }

    return pRoutingCtx;
}


PA_Routing_Err_t PARoutingPAConnect
(
    PA_Routing_Ctx_t *pCtx /**< [in] Pointer to Routing Context */
)
{
    int success = TRUE;

    PAROUTING_DEBUG_LOG ("PARoutingPAConnect\n");

    if (!pCtx)
    {
        return PA_ROUTING_ERR;
    }

    if (PA_CLIENT_OK != PAClientCtxInit (pCtx->pClientCtx, "Audio Manager Routing Client"))
    {
        return PA_ROUTING_ERR;
    }

    if (PA_CLIENT_OK != PAClientCtxConnect (pCtx->pClientCtx))
    {
        success = FALSE;
    }

    if (success == TRUE)
    {
        if (PA_CLIENT_OK
                != PAClientEventSubscribe (pCtx->pClientCtx, PARoutingEventCallback, pCtx,
                                           PA_SUBSCRIPTION_MASK_SINK_INPUT | PA_SUBSCRIPTION_MASK_SOURCE_OUTPUT))
        {
            success = FALSE;
            PAROUTING_DEBUG_LOG ("PARoutingPAConnect failed to subscribe to events\n");
        }
    }

    if (success == TRUE)
    {
        if (PAClientCtxLoadModule (pCtx->pClientCtx, "module-null-sink", "sink_name=routing_null_sink",
                                   &(pCtx->defaultSinkModule)) != PA_CLIENT_OK)
        {
            PAROUTING_DEBUG_LOG ("PAClientCtxLoadModule Fails\n");
            success = FALSE;
        }
    }

    if (success == TRUE)
    {
        if (PAClientGetDevFromModule (pCtx->pClientCtx, pCtx->defaultSinkModule, PA_CLIENT_DEVICE_SINK,
                                      &(pCtx->defaultSink)) != PA_CLIENT_OK)
        {
            PAROUTING_DEBUG_LOG ("PAClientGetDevFromModule Fails\n");
            success = FALSE;
        }
    }

    if (success == TRUE)
    {
        if (PAClientCtxLoadModule (pCtx->pClientCtx, "module-null-sink", "sink_name=routing_null_source",
                                   &(pCtx->defaultSourceModule)) != PA_CLIENT_OK)
        {
            PAROUTING_DEBUG_LOG ("PAClientCtxLoadModule Fails\n");
            success = FALSE;
        }
    }

    if (success == TRUE)
    {
        if (PAClientGetDevFromModule (pCtx->pClientCtx, pCtx->defaultSourceModule, PA_CLIENT_DEVICE_SOURCE,
                                      &(pCtx->defaultSource)) != PA_CLIENT_OK)
        {
            PAROUTING_DEBUG_LOG ("PAClientGetDevFromModule Fails\n");
            success = FALSE;
        }
    }

    if (success == FALSE)
    {
        PARoutingPADisconnect (pCtx);
        return PA_ROUTING_ERR;
    }

    return PA_ROUTING_OK;
}


PA_Routing_Err_t PARoutingPADisconnect
(
    PA_Routing_Ctx_t *pCtx /**< [in] Pointer to Routing Context */
)
{
    if (pCtx->pClientCtx != NULL)
    {
        PAClientCtxUnloadModule (pCtx->pClientCtx, pCtx->defaultSinkModule);
        PAClientCtxUnloadModule (pCtx->pClientCtx, pCtx->defaultSourceModule);
        PAClientCtxCleanup (pCtx->pClientCtx);
    }

    return PA_ROUTING_OK;
}


PA_Routing_Err_t PARoutingModuleCheckPAConnection
(
    PA_Routing_Ctx_t *pCtx /**< [in] Pointer to Routing Context */
)
{
    if (PA_CLIENT_OK != PAClientCheckConnection (pCtx->pClientCtx))
    {
        return PA_ROUTING_ERR;
    }

    return PA_ROUTING_OK;
}


PA_Routing_Err_t PARoutingModuleFree
(
    PA_Routing_Ctx_t *pCtx /**< [in] Pointer to Routing Context */
)
{
    if (!pCtx)
    {
        return PA_ROUTING_ERR;
    }

    PARoutingPADisconnect (pCtx);

    if (pCtx->pClientCtx)
    {
        free (pCtx->pClientCtx);
    }

    PARoutingModuleRemoveAllRoutes (pCtx);

    if (pCtx->pRouteListMutex)
    {
        g_mutex_free (pCtx->pRouteListMutex);
    }

    if (pCtx->pPropertyKey)
    {
        free (pCtx->pPropertyKey);
    }

    free (pCtx);
    return PA_ROUTING_OK;
}


PA_Routing_Err_t PARoutingModuleSetRoute
(
    PA_Routing_Ctx_t *pCtx, /**< [in] Pointer to Routing Context */
    PA_Route_Type_t direction, /**< [in] PA_OUTPUT_ROUTE to apply route to 
                                                                                    output (playback) streams, or PA_INPUT_ROUTE 
                                                                                    to apply route to input (capture) streams */
    const char *pPropertyValue, /**< [in] String to search for in the streams' properties */
    uint32_t deviceIndex       /**< [in] Index of the source/sink to route the stream to */
)
{
    PA_Route_List_Element_t *pElem;
    GSList * *list, *tmpList;

    PAROUTING_DEBUG_LOG ("SetRoute\n");
    g_mutex_lock (pCtx->pRouteListMutex);

    if (direction == PA_ROUTE_OUTPUT)
    {
        tmpList = pCtx->pOutputRouteList;
        list = &(pCtx->pOutputRouteList);
    }
    else
    {
        tmpList = pCtx->pInputRouteList;
        list = &(pCtx->pInputRouteList);
    }
    /* now search list to see if we already have an entry for this stream */

    while (tmpList != NULL)
    {
        pElem = (PA_Route_List_Element_t *)tmpList->data;
        if (strcmp (pElem->pStreamProp, pPropertyValue) == 0)
        {
            PAROUTING_DEBUG_LOG ("Found entry already for stream, updating index\n");
            pElem->devIndex = deviceIndex;
            g_mutex_unlock (pCtx->pRouteListMutex);
            return PA_ROUTING_OK;
        }
        tmpList = g_slist_next (tmpList);
    }
    /* No matching element found, so add a new one to the list */

    pElem = (PA_Route_List_Element_t *)malloc (sizeof (PA_Route_List_Element_t));

    if (pElem)
    {
        pElem->pStreamProp = g_strdup (pPropertyValue);
        pElem->devIndex = deviceIndex;
        *list = g_slist_prepend (*list, pElem);
    }
    else
    {
        PAROUTING_DEBUG_LOG ("malloc failed\n");
        g_mutex_unlock (pCtx->pRouteListMutex);
        return PA_ROUTING_ERR;
    }

    g_mutex_unlock (pCtx->pRouteListMutex);

    PAROUTING_DEBUG_LOG ("return from set route\n");
    return PA_ROUTING_OK;
}


PA_Routing_Err_t PARoutingModuleSetAndApplyRoute
(
    PA_Routing_Ctx_t *pCtx, /**< [in] Pointer to Routing Context */
    PA_Route_Type_t direction, /**< [in] PA_OUTPUT_ROUTE tgo apply route to 
                                                                                  output (playback) streams, or PA_INPUT_ROUTE 
                                                                                  to apply route to input (capture) streams */
    const char *pPropertyValue, /**< [in] String to search for in the streams' properties */
    uint32_t deviceIndex       /**< [in] Index of the source/sink to route the stream to */
)
{
    PA_Route_List_Element_t *pElem;
    GSList * *list, *tmpList;
    gboolean foundElem;
    pa_operation *o;

    PARouteApplyRouteCBData_t *pCbData, cbData;
    pa_threaded_mainloop *pMainLoop;
    int success;

    pCbData = &cbData;

    pCbData->pRouteCtx = pCtx;

    pMainLoop = pCtx->pClientCtx->pMainLoop;
    pa_threaded_mainloop_lock (pMainLoop);

    PAROUTING_DEBUG_LOG ("SetandApplyRoute\n");
    g_mutex_lock (pCtx->pRouteListMutex);

    if (direction == PA_ROUTE_OUTPUT)
    {
        tmpList = pCtx->pOutputRouteList;
        list = &(pCtx->pOutputRouteList);
    }
    else
    {
        tmpList = pCtx->pInputRouteList;
        list = &(pCtx->pInputRouteList);
    }
    /* now search list to see if we already have an entry for this stream */

    foundElem = FALSE;

    while (tmpList != NULL)
    {
        pElem = (PA_Route_List_Element_t *)tmpList->data;
        if (strcmp (pElem->pStreamProp, pPropertyValue) == 0)
        {
            PAROUTING_DEBUG_LOG ("Found entry already for stream, updating index\n");
            pElem->devIndex = deviceIndex;
            foundElem = TRUE;
        }
        tmpList = g_slist_next (tmpList);
    }

    if (foundElem == FALSE)
    {
        /* No matching element found, so add a new one to the list */
        pElem = (PA_Route_List_Element_t *)malloc (sizeof (PA_Route_List_Element_t));
        if (!pElem)
        {
            PAROUTING_DEBUG_LOG ("malloc failed\n");
            g_mutex_unlock (pCtx->pRouteListMutex);
            pa_threaded_mainloop_unlock (pMainLoop);
            return PA_ROUTING_ERR;
        }
        pElem->pStreamProp = g_strdup (pPropertyValue);
        pElem->devIndex = deviceIndex;
        *list = g_slist_prepend (*list, pElem);
        PAROUTING_DEBUG_LOG ("return from set route\n");
    }
    /* Now apply route */

    g_mutex_unlock (pCtx->pRouteListMutex);

    pCbData->pMainLoop = pMainLoop;
    pCbData->pRouteStreamId = g_strdup (pPropertyValue);
    pCbData->devIndex = deviceIndex;

    success = TRUE;

    if (direction == PA_ROUTE_OUTPUT)
    {
        o = pa_context_get_sink_input_info_list (pCtx->pClientCtx->pPACtx, PARoutingApplySinkInputInfoCallback,
                                                 pCbData);
    }
    else
    {
        o = pa_context_get_source_output_info_list (pCtx->pClientCtx->pPACtx, PARoutingApplySourceOutputInfoCallback,
                                                    pCbData);
    }

    if (!o)
    {
        PAROUTING_DEBUG_LOG ("operation failed\n");
        success = FALSE;
    }

    if (success == TRUE)
    {
        PAROUTING_DEBUG_LOG ("setandapply, entering wait loop\n");
        while (pa_operation_get_state (o) != PA_OPERATION_DONE)
        {
            pa_threaded_mainloop_wait (pMainLoop);
            CONTEXT_OPERATION_CHECK_DEAD (pCtx->pClientCtx, o, pCbData->pRouteStreamId);
        }
        pa_operation_unref (o);
    }

    g_free (pCbData->pRouteStreamId);

    pa_threaded_mainloop_accept (pMainLoop);
    pa_threaded_mainloop_unlock (pMainLoop);

    PAROUTING_DEBUG_LOG ("return from set and apply \n");

    if (success == TRUE)
    {
        return PA_ROUTING_OK;
    }

    return PA_ROUTING_ERR;
}


PA_Routing_Err_t PARoutingModuleGetRoute
(
    PA_Routing_Ctx_t *pCtx,     /**< [in] Pointer to Routing Context */
    PA_Route_Type_t direction,  /**< [in] PA_OUTPUT_ROUTE to get route for 
                                                                                  output (playback) streams, or PA_INPUT_ROUTE 
                                                                                  to get route to input (capture) streams */
    const char *pPropertyValue, /**< [in]  Stream property string to get route for */
    size_t length,              /**< [in] Length of Stream Property String */
    uint32_t *deviceIndex       /**< [out] If succesful, the device index is copied 
                                                                     to *deviceIndex */
)
{
    PA_Route_List_Element_t *pElem;
    GSList * *list, *tmpList;
    PA_Routing_Err_t ret = PA_ROUTING_ERR;

    PAROUTING_DEBUG_LOG ("GetRoute\n");
    g_mutex_lock (pCtx->pRouteListMutex);

    if (direction == PA_ROUTE_OUTPUT)
    {
        tmpList = pCtx->pOutputRouteList;
        list = &(pCtx->pOutputRouteList);
    }
    else
    {
        tmpList = pCtx->pInputRouteList;
        list = &(pCtx->pInputRouteList);
    }
    /* now search list to see if we already have an entry for this stream */

    while (tmpList != NULL)
    {
        pElem = (PA_Route_List_Element_t *)tmpList->data;
        if (PARoutingStringsEqual (pElem->pStreamProp, pPropertyValue, strlen (pElem->pStreamProp)) == TRUE)
        {
            *deviceIndex = pElem->devIndex;
            PAROUTING_DEBUG_LOG ("Found entry already for stream %s, deviceIndex %d\n", pPropertyValue, *deviceIndex);
            ret = PA_ROUTING_OK;
        }
        tmpList = g_slist_next (tmpList);
    }

    g_mutex_unlock (pCtx->pRouteListMutex);
    PAROUTING_DEBUG_LOG ("return from get route\n");
    return ret;
}


PA_Routing_Err_t PARoutingModuleRemoveRoute
(
    PA_Routing_Ctx_t *pCtx, /**< [in] Pointer to Routing Context */
    PA_Route_Type_t direction, /**< [in] PA_OUTPUT_ROUTE to get route for 
                                                                                   output (playback) streams, or PA_INPUT_ROUTE 
                                                                                    to get route to input (capture) streams */
    const char *pPropertyValue /**< [in]  Stream property string to get route for */
)
{
    PA_Route_List_Element_t *pElem;
    GSList * *list, *tmpList;
    PA_CLIENT_STREAM_TYPE_t streamType;
    uint32_t defaultDeviceIndex;

    PAROUTING_DEBUG_LOG ("RemoveRoute\n");
    g_mutex_lock (pCtx->pRouteListMutex);

    if (direction == PA_ROUTE_OUTPUT)
    {
        tmpList = pCtx->pOutputRouteList;
        list = &(pCtx->pOutputRouteList);
        streamType = PA_CLIENT_SINK_INPUT;
        defaultDeviceIndex = pCtx->defaultSink;
    }
    else
    {
        tmpList = pCtx->pInputRouteList;
        list = &(pCtx->pInputRouteList);
        streamType = PA_CLIENT_SOURCE_OUTPUT;
        defaultDeviceIndex = pCtx->defaultSource;
    }
    /* now search list to see if we  have an entry for this stream */

    while (tmpList != NULL)
    {
        pElem = (PA_Route_List_Element_t *)tmpList->data;
        if (strcmp (pElem->pStreamProp, pPropertyValue) == 0)
        {
            PAROUTING_DEBUG_LOG ("Found element %s to remove\n", pElem->pStreamProp);
            g_free (pElem->pStreamProp);
            *list = g_slist_remove (*list, pElem);
            free (pElem);
        }
        tmpList = g_slist_next (tmpList);
    }

    g_mutex_unlock (pCtx->pRouteListMutex);
    PAROUTING_DEBUG_LOG ("return PARoutingModuleRemoveRoute\n");

    PAClientMoveAllStreamsByProperty (pCtx->pClientCtx, streamType, pCtx->pPropertyKey, pPropertyValue,
                                      defaultDeviceIndex);

    return PA_ROUTING_OK;
}


PA_Routing_Err_t PARoutingModuleRemoveAllRoutes
(
    PA_Routing_Ctx_t *pCtx /**< [in] Pointer to Routing Context */
)
{
    PA_Route_List_Element_t *pElem;
    GSList *tmpList;

    PAROUTING_DEBUG_LOG ("RemoveAllRoutes\n");
    g_mutex_lock (pCtx->pRouteListMutex);

    tmpList = pCtx->pOutputRouteList;

    /* delete each list element, then delete the list */

    while (tmpList != NULL)
    {
        pElem = (PA_Route_List_Element_t *)tmpList->data;
        PAROUTING_DEBUG_LOG ("Found element %s to remove\n", pElem->pStreamProp);
        g_free (pElem->pStreamProp);
        free (pElem);
        tmpList = g_slist_next (tmpList);
    }

    g_slist_free (pCtx->pOutputRouteList);
    pCtx->pOutputRouteList = NULL;

    tmpList = pCtx->pInputRouteList;

    /* delete each list element, then delete the list */

    while (tmpList != NULL)
    {
        pElem = (PA_Route_List_Element_t *)tmpList->data;
        PAROUTING_DEBUG_LOG ("Found element %s to remove\n", pElem->pStreamProp);
        g_free (pElem->pStreamProp);
        free (pElem);
        tmpList = g_slist_next (tmpList);
    }

    g_slist_free (pCtx->pInputRouteList);
    pCtx->pInputRouteList = NULL;

    g_mutex_unlock (pCtx->pRouteListMutex);
    return PA_ROUTING_OK;
}


PA_Routing_Err_t PARoutingModuleDumpRoutes
(
    PA_Routing_Ctx_t *pCtx /**< [in] Pointer to Routing Context */
)
{
    GSList *list;
    PA_Route_List_Element_t *pElem;

    if (!pCtx)
    {
        return PA_ROUTING_ERR;
    }

    PAROUTING_DEBUG_LOG ("Dump Routes\n");
    g_mutex_lock (pCtx->pRouteListMutex);
    PAROUTING_DEBUG_LOG ("Input Routes:\n");

    list = pCtx->pInputRouteList;

    while (list != NULL)
    {
        pElem = (PA_Route_List_Element_t *)list->data;
        PAROUTING_DEBUG_LOG ("Stream property: %s routed to sink index %d\n", pElem->pStreamProp, pElem->devIndex);
        list = g_slist_next (list);
    }

    PAROUTING_DEBUG_LOG ("\n");

    PAROUTING_DEBUG_LOG ("Output Routes:\n");

    list = pCtx->pOutputRouteList;

    while (list != NULL)
    {
        pElem = (PA_Route_List_Element_t *)list->data;
        PAROUTING_DEBUG_LOG ("Stream property: %s routed to sink index %d\n", pElem->pStreamProp, pElem->devIndex);
        list = g_slist_next (list);
    }

    PAROUTING_DEBUG_LOG ("\n");

    g_mutex_unlock (pCtx->pRouteListMutex);
    return PA_ROUTING_OK;
}


PA_Routing_Err_t PARoutingModuleApplyRoutes
(
    PA_Routing_Ctx_t *pCtx /**< [in] Pointer to Routing Context */
)
{
    PA_Routing_Err_t ret = PA_ROUTING_ERR;

    if ((pa_context_get_sink_input_info_list (pCtx->pClientCtx->pPACtx, PARoutingSinkInputInfoCallback, pCtx)) && \
            pa_context_get_source_output_info_list (pCtx->pClientCtx->pPACtx, PARoutingSourceOutputInfoCallback, pCtx))
    {
        ret = PA_ROUTING_OK;
    }

    return ret;
}
/**************************************************************************/
/*                       Private Functions Bodies                         */
/**************************************************************************/
static void PARoutingApplySinkInputInfoCallback
(
    pa_context *c, const pa_sink_input_info *i, int eol, void *userdata
)
{
    const void *prop;
    size_t length;

    PARouteApplyRouteCBData_t *pCbData;

    pCbData = (PARouteApplyRouteCBData_t *)userdata;

    if (!eol)
    {
        PAROUTING_DEBUG_LOG ("Routing Apply SinkInput Info callback,stream %i for sink %d\n", i->index, i->sink);
        if (!pa_proplist_get (i->proplist, pCbData->pRouteCtx->pPropertyKey, &prop, &length))
        {
            PAROUTING_DEBUG_LOG ("%s:%s\n", pCbData->pRouteCtx->pPropertyKey, (char *)prop);
            if (PARoutingStringsEqual (pCbData->pRouteStreamId, (char *)prop, strlen (pCbData->pRouteStreamId)) == TRUE)
            {
                if (pCbData->devIndex != i->sink)
                {
                    PAROUTING_DEBUG_LOG ("Stream needs to move to sink %d\n", pCbData->devIndex);
                    if (!pa_context_move_sink_input_by_index (c, i->index, pCbData->devIndex, PARoutingSuccessCallback,
                                                              NULL))
                    {
                        fprintf (stderr, "pa_context_move_sink_input_by_index error\n");
                    }
                }
                else
                {
                    PAROUTING_DEBUG_LOG ("Stream already connected to correct sink\n");
                }
            }
            else
            {
                PAROUTING_DEBUG_LOG ("Did not find route for stream\n");
            }
        }
        else
        {
            PAROUTING_DEBUG_LOG ("%s not found\n", pCbData->pRouteCtx->pPropertyKey);
        }
    }
    else
    {
        pa_threaded_mainloop_signal (pCbData->pMainLoop, PA_DONT_BLOCK);
    }
}


static void PARoutingApplySourceOutputInfoCallback
(
    pa_context *c, const pa_source_output_info *i, int eol, void *userdata
)
{
    const void *prop;
    size_t length;
    PARouteApplyRouteCBData_t *pCbData;

    pCbData = (PARouteApplyRouteCBData_t *)userdata;

    if (!eol)
    {
        PAROUTING_DEBUG_LOG ("Apply SourceOutput Route Info callback,stream %i for sink %d\n", i->index, i->source);
        if (!pa_proplist_get (i->proplist, pCbData->pRouteCtx->pPropertyKey, &prop, &length))
        {
            PAROUTING_DEBUG_LOG ("comparing %s:%s\n", pCbData->pRouteCtx->pPropertyKey, (char *)prop);
            if (PARoutingStringsEqual (pCbData->pRouteStreamId, (char *)prop, strlen (pCbData->pRouteStreamId)) == TRUE)
            {
                if (pCbData->devIndex != i->source)
                {
                    PAROUTING_DEBUG_LOG ("Stream needs to move to source %d\n", pCbData->devIndex);
                    if (!pa_context_move_source_output_by_index (c,                        i->index, pCbData->devIndex,
                                                                 PARoutingSuccessCallback, NULL))
                    {
                        fprintf (stderr, "pa_context_move_source_output_by_index error\n");
                    }
                }
                else
                {
                    PAROUTING_DEBUG_LOG ("Stream already connected to correct source\n");
                }
            }
        }
        else
        {
            PAROUTING_DEBUG_LOG ("%s not found\n", pCbData->pRouteCtx->pPropertyKey);
        }
    }
    else
    {
        pa_threaded_mainloop_signal (pCbData->pMainLoop, PA_DONT_BLOCK);
    }
}


static void PARoutingSinkInputInfoCallback
(
    pa_context *c, const pa_sink_input_info *i, int eol, void *userdata
)
{
    const void *prop;
    size_t length;
    uint32_t deviceIndex;

    PA_Routing_Ctx_t *pCtx;

    pCtx = (PA_Routing_Ctx_t *)userdata;

    if (!eol)
    {
        PAROUTING_DEBUG_LOG ("New SinkInput Info callback,stream %i for sink %d\n", i->index, i->sink);
        if (!pa_proplist_get (i->proplist, pCtx->pPropertyKey, &prop, &length))
        {
            PAROUTING_DEBUG_LOG ("%s:%s\n", pCtx->pPropertyKey, (char *)prop);
            if (PARoutingModuleGetRoute (pCtx, PA_ROUTE_OUTPUT, (char *)prop, length, &deviceIndex) == PA_ROUTING_OK)
            {
                if (deviceIndex != i->sink)
                {
                    PAROUTING_DEBUG_LOG ("Stream needs to move to sink %d\n", deviceIndex);
                    if (!pa_context_move_sink_input_by_index (c, i->index, deviceIndex, PARoutingSuccessCallback, NULL))
                    {
                        fprintf (stderr, "pa_context_move_sink_input_by_index error\n");
                    }
                }
                else
                {
                    PAROUTING_DEBUG_LOG ("Stream already connected to correct sink\n");
                }
            }
            else
            {
                PAROUTING_DEBUG_LOG ("Did not find route for stream\n");
            }
        }
        else
        {
            PAROUTING_DEBUG_LOG ("%s not found\n", pCtx->pPropertyKey);
        }
    }
}


static void PARoutingSourceOutputInfoCallback
(
    pa_context *c, const pa_source_output_info *i, int eol, void *userdata
)
{
    const void *prop;
    size_t length;
    uint32_t deviceIndex;

    PA_Routing_Ctx_t *pCtx;

    pCtx = (PA_Routing_Ctx_t *)userdata;

    if (!eol)
    {
        PAROUTING_DEBUG_LOG ("New SourceOutput Info callback,stream %i for sink %d looking for %s\n", i->index, i->source, pCtx->pPropertyKey);
        if (!pa_proplist_get (i->proplist, pCtx->pPropertyKey, &prop, &length))
        {
            PAROUTING_DEBUG_LOG ("%s:%s\n", pCtx->pPropertyKey, (char *)prop);
            PAROUTING_DEBUG_LOG ("comparing %s:%s\n", pCtx->pPropertyKey, (char *)prop);
            if (PARoutingModuleGetRoute (pCtx, PA_ROUTE_INPUT, (char *)prop, length, &deviceIndex) == PA_ROUTING_OK)
            {
                if (deviceIndex != i->source)
                {
                    PAROUTING_DEBUG_LOG ("Stream needs to move to source %d\n", deviceIndex);
                    if (!pa_context_move_source_output_by_index (c, i->index, deviceIndex, PARoutingSuccessCallback, NULL))
                    {
                        fprintf (stderr, "pa_context_move_source_output_by_index error\n");
                    }
                }
                else
                {
                    PAROUTING_DEBUG_LOG ("Stream already connected to correct source\n");
                }
            }
        }
        else
        {
            PAROUTING_DEBUG_LOG ("%s not found\n", pCtx->pPropertyKey);
        }
    }
}


static void PARoutingEventCallback
(
    pa_context *c, pa_subscription_event_type_t t, uint32_t index, void *userdata
)
{
    PA_Routing_Ctx_t *pCtx;
    int sink = 0;

    pCtx = (PA_Routing_Ctx_t *)userdata;
    PAROUTING_DEBUG_LOG ("PARoutingEventCallback\n");

    switch (t & PA_SUBSCRIPTION_EVENT_FACILITY_MASK)
    {
        case PA_SUBSCRIPTION_EVENT_SINK_INPUT:
            PAROUTING_DEBUG_LOG ("Sink Input changed, index is %d, event is: ", index);
            sink = 1;
            break;
        case PA_SUBSCRIPTION_EVENT_SOURCE_OUTPUT:
            PAROUTING_DEBUG_LOG ("Source Output changed, index is %d event is: ", index);
            sink = 0;
            break;

        default:
            PAROUTING_DEBUG_LOG ("unknown EVENT\n");
    }

    switch (t & PA_SUBSCRIPTION_EVENT_TYPE_MASK)
    {
        case PA_SUBSCRIPTION_EVENT_NEW:
            PAROUTING_DEBUG_LOG ("New object, dispatching new_sink/src info request\n");
        {
            if (sink)
            {
                pa_context_get_sink_input_info (c, index, PARoutingSinkInputInfoCallback, pCtx);
            }
            else
            {
                pa_context_get_source_output_info (c, index, PARoutingSourceOutputInfoCallback, pCtx);
            }
        }
            break;
        case PA_SUBSCRIPTION_EVENT_REMOVE:
            PAROUTING_DEBUG_LOG ("Object Removed\n");
            break;
        case PA_SUBSCRIPTION_EVENT_CHANGE:
            PAROUTING_DEBUG_LOG ("Object Changed\n");
            break;

        default:
            PAROUTING_DEBUG_LOG ("unhandled event\n");
            break;
    }
}


static void PARoutingSuccessCallback
(
    pa_context *c, int success, void *userdata
)
{
    PAROUTING_DEBUG_LOG ("success %d\n", success);
} static int PARoutingStringsEqual
(
    const char *str1, const char *str2, size_t length
)
{
    if (strncmp (str1, str2, length) == 0)
    {
        if ((str1[length] == '\0') && (str2[length] == '\0'))
        {
            return TRUE;
        }
    }

    return FALSE;
}
