/*
 * parouting.h 
 *
 * Copyright (c) 2008 Wind River Systems, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Lesser GNU General Public License version 2.1 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the Lesser GNU General Public License for more details.
 *
 * You should have received a copy of the Lesser GNU General Public License version 2.1
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */
/** 
 * \ingroup pulse_audio_routing
 * \file
 *    PulseAudio Routing Module.  This module is used to connect to the
 *    local PulseAudio server and set up routing policies.  The routing 
 *    module monitors PulseAudio streams (sink inputs/ source outputs) and
 *    if a stream whose property is set in the routing list appears, it
 *    is moved to the device set in the routing list.
 *
 *    The property key that is used is set in the call to PARoutingNew and
 *    cannot be changed. The property value is set by the application either
 *    programmatically using the PulseAudio client API, or by setting the
 *    environment variable PULSE_PROP_xyz=abc  where xyz is the property key
 *    and abc is the property value that is to be set in the route.
 *
 * \version $Revision: 1.9 $
 * \date    $Date: 2008/07/12 00:56:52 $
 */
#ifndef _PAROUTING_H
#define _PAROUTING_H

/**************************************************************************/
/*                              Includes                                  */
/**************************************************************************/
#include <glib.h>
#include <pulse/pulseaudio.h>
#include "paclient.h"

/**************************************************************************/
/*                               Defines                                  */
/**************************************************************************/

/**************************************************************************/
/*                              Typedefs                                  */
/**************************************************************************/

/** Context for Routing Module **/
typedef struct
{
    GSList *pInputRouteList;
    GSList *pOutputRouteList;
    GMutex *pRouteListMutex;
    char *pPropertyKey;
    PA_Client_Ctx_t *pClientCtx;
    uint32_t defaultSink;
    uint32_t defaultSource;
    uint32_t defaultSourceModule;
    uint32_t defaultSinkModule;
} PA_Routing_Ctx_t;

/** Routing Module Error Codes */
typedef enum
{
    PA_ROUTING_OK = 0,
    PA_ROUTING_ERR
} PA_Routing_Err_t;

/** enum used to indicate an input routing (PulseAudio Src to application)
  or output routing (application to PulseAudio Sink )**/
typedef enum
{
    PA_ROUTE_OUTPUT = 0,
    PA_ROUTE_INPUT
} PA_Route_Type_t;

/**************************************************************************/
/*                          Functions Declarations                        */
/**************************************************************************/
/**************************************************************************/
/*                                                                        */
/*  Function Name: PARoutingModuleNew                                     */
/*                                                                        */
/**************************************************************************/
/**
 *  Create new instance of routing module.  The string pPropertyKey will be
 *  used by the module to match PulseAudio Source Outputs and Sink Inputs
 *  against configured routes.  For example setting the env variable
 *  PA_PROP_acmdevid for an Audio application, would require
 *  pPropertyKey to be "acmdevid".   The Routing Module routes based on only
 *  one PulseAudio Stream Property. 
 *
 *
 *  \return Pointer to newly allocated PA_Routing_Ctx_t on success or NULL.
 *
 *  \see PARoutingModuleFree
 *                                                     
 **************************************************************************
 */
PA_Routing_Ctx_t *PARoutingModuleNew
(
    const char *pPropertyKey /**< [in] String to use as the Stream Property Key */
);

/**************************************************************************/
/*                                                                        */
/*  Function Name: PARoutingPAConnect                                     */
/*                                                                        */
/**************************************************************************/
/**
 *  Connect the PARoutingModule to PulseAudio.  Requires that PARoutingModuleNew
 *  have been called.             
 *
 *  \return PA_ROUTING_OK if able to connect or PA_ROUTING_ERR
 *
 *  \see PARoutingPADisconnect
 *                                                     
 **************************************************************************
 */
PA_Routing_Err_t PARoutingPAConnect
(
    PA_Routing_Ctx_t *pCtx /**< [in] Pointer to Routing Context */
);

/**************************************************************************/
/*                                                                        */
/*  Function Name: PARoutingPADisconnect                                  */
/*                                                                        */
/**************************************************************************/
/**
 *  Disconnects the PARoutingModule from PulseAudio.  pCtx can be used
 *  subsequently in a call to PARoutingPAConnect, or freed using 
 *  PARoutingModuleFree               
 *
 *  \return PA_ROUTING_OK                 
 *                                                     
 **************************************************************************
 */
PA_Routing_Err_t PARoutingPADisconnect
(
    PA_Routing_Ctx_t *pCtx /**< [in] Pointer to Routing Context */
);

/**************************************************************************/
/*                                                                        */
/*  Function Name: PARoutingModuleFree                                    */
/*                                                                        */
/**************************************************************************/
/**
 *  Releases all Resources pointed to by pCtx               
 *
 *  \return PA_ROUTING_OK                  
 *                                                     
 **************************************************************************
 */
PA_Routing_Err_t PARoutingModuleFree
(
    PA_Routing_Ctx_t *pCtx /**< [in] Pointer to Routing Context */
);

/**************************************************************************/
/*                                                                        */
/*  Function Name: PARoutingModuleSetRoute                                */
/*                                                                        */
/**************************************************************************/
/**
 *  Adds a Route to the Routing Module.   After adding routes call
 *  PAROutingModuleApplyRoutes to act on current sink inputs/source outputs,
 *  otherwse the Route will be applied when the new sink input/source output
 *  is created.
 *
 *  Only one route per unique pPropertyValue string is allowed.  Each call with
 *  the same pPropertyValue string will overwrite any existing route.
 *
 *  \return PA_ROUTING_OK if successful.                 
 *                                                     
 **************************************************************************
 */
PA_Routing_Err_t PARoutingModuleSetRoute
(
    PA_Routing_Ctx_t *pCtx, /**< [in] Pointer to Routing Context */
    PA_Route_Type_t direction, /**< [in] PA_OUTPUT_ROUTE tgo apply route to 
                                  output (playback) streams, or PA_INPUT_ROUTE 
                                  to apply route to input (capture) streams */

    const char *pPropertyValue, /**< [in] String to search for in the streams' properties */
    uint32_t deviceIndex       /**< [in] Index of the source/sink to route the stream to */
);

/**************************************************************************/
/*                                                                        */
/*  Function Name: PARoutingModuleSetAndApplyRoute                        */
/*                                                                        */
/**************************************************************************/
/**
 *  Adds a Route to the Routing Module and apply route to current connections
 *  as well as future routes.
 *
 *  Only one route per unique pPropertyValue string is allowed.  Each call with
 *  the same pPropertyValue string will overwrite any existing route.
 *
 *  \return PA_ROUTING_OK if successful.                 
 *                                                     
 **************************************************************************
 */
PA_Routing_Err_t PARoutingModuleSetAndApplyRoute
(
    PA_Routing_Ctx_t *pCtx, /**< [in] Pointer to Routing Context */
    PA_Route_Type_t direction, /**< [in] PA_OUTPUT_ROUTE tgo apply route to 
                                  output (playback) streams, or PA_INPUT_ROUTE 
                                  to apply route to input (capture) streams */

    const char *pPropertyValue, /**< [in] String to search for in the streams' properties */
    uint32_t deviceIndex       /**< [in] Index of the source/sink to route the stream to */
);

/**************************************************************************/
/*                                                                        */
/*  Function Name: PARoutingModuleGetRoute                                */
/*                                                                        */
/**************************************************************************/
/**
 *  Retrieves the device index for the route. 
 *
 *  \return PA_ROUTING_OK if successful.                 
 *                                                     
 **************************************************************************
 */
PA_Routing_Err_t PARoutingModuleGetRoute
(
    PA_Routing_Ctx_t *pCtx, /**< [in] Pointer to Routing Context */
    PA_Route_Type_t direction, /**< [in] PA_OUTPUT_ROUTE to get route for 
                                                                                   output (playback) streams, or PA_INPUT_ROUTE 
                                                                                   to get route to input (capture) streams */
    const char *pPropertyValue, /**< [in] Stream property string to get route for */
    size_t length,             /**< [in] Length of Stream Property String */
    uint32_t *deviceIndex /**< [out] If succesful, the device index is copied 
                          to *deviceIndex */
);

/**************************************************************************/
/*                                                                        */
/*  Function Name: PARoutingModuleRemoveRoute                             */
/*                                                                        */
/**************************************************************************/
/**
 *  Removes the route from the routing module.  Does not effect streams that
 *  are already connected.
 *
 *  \return PA_ROUTING_OK                 
 *                                                     
 **************************************************************************
 */
PA_Routing_Err_t PARoutingModuleRemoveRoute
(
    PA_Routing_Ctx_t *pCtx,    /**< [in] Pointer to Routing Context */
    PA_Route_Type_t direction, /**< [in] PA_OUTPUT_ROUTE to get route for 
                                                                                   output (playback) streams, or PA_INPUT_ROUTE 
                                                                                   to get route to input (capture) streams */
    const char *pPropertyValue /**< [in]  Stream property string to get route for */
);

/**************************************************************************/
/*                                                                        */
/*  Function Name: PARoutingModuleRemoveAllRoutes                         */
/*                                                                        */
/**************************************************************************/
/**
 *  Removes all routes from Routing Module.  Does not effect streams already
 *  connected
 *
 *  \return PA_ROUTING_OK if successful.                 
 *                                                     
 **************************************************************************
 */
PA_Routing_Err_t PARoutingModuleRemoveAllRoutes
(
    PA_Routing_Ctx_t *pCtx /**< [in] Pointer to Routing Context */
);

/**************************************************************************/
/*                                                                        */
/*  Function Name: PARoutingModuleDumpRoutes                         */
/*                                                                        */
/**************************************************************************/
/**
 *  Prints out all routes in the Routing Module
 *
 *  \return PA_ROUTING_OK if successful.                 
 *                                                     
 **************************************************************************
 */
PA_Routing_Err_t PARoutingModuleDumpRoutes
(
    PA_Routing_Ctx_t *pCtx /**< [in] Pointer to Routing Context */
);

/**************************************************************************/
/*                                                                        */
/*  Function Name: PARoutingModuleApplyRoutes                             */
/*                                                                        */
/**************************************************************************/
/**
 *  Apply all routes stored in routing module to currently connected streams              
 *
 *  \return PA_ROUTING_OK if successful.                 
 *                                                     
 **************************************************************************
 */
PA_Routing_Err_t PARoutingModuleApplyRoutes
(
    PA_Routing_Ctx_t *pCtx /**< [in] Pointer to Routing Context */
);

/**************************************************************************/
/*                                                                        */
/*  Function Name: PARoutingModuleCheckPAConnection                         */
/*                                                                        */
/**************************************************************************/
/**
 *  Check the state of the PulseAudio connection
 *
 *  \return PA_ROUTING_OK if server connection is ok, otherwise
 * PulseAudio connection needs to be disconnected.
 *                                                     
 **************************************************************************
 */
PA_Routing_Err_t PARoutingModuleCheckPAConnection
(
    PA_Routing_Ctx_t *pCtx /**< [in] Pointer to Routing Context */
);
#endif                     //_PAROUTING_H
